(* ::Package:: *)

(* ::Chapter:: *)
(*SimulateP2P Package*)


(* ::Subchapter::Closed:: *)
(*Begin Package*)


(* ::Subsection::Closed:: *)
(*BeginPackage*)


(* ::Input::Initialization:: *)
BeginPackage["SimulateP2P`"]


(* ::Subsection::Closed:: *)
(*Exported Wrapper Symbols*)


(* ::Input::Initialization:: *)
Investment::usage="Investment[funds, principal, month, months, monthlyRate, monthlyPayment, grading] represents a quantity of cash together with an outstanding loan principal which is in the month out of months to run. The monthly interest rate is monthlyRate and the monthly payment made by the borrower is monthlyPayment. As time progresses the lender receives cash and the outstanding principal is adjusted.";


(* ::Input::Initialization:: *)
Basket::usage="Basket[cash, {investments}, gradings] represents a Basket of investments together with the amount in cash, and a policy to pick new investments fom the specification in gradings. (gradings follows the same specification as RandomChoice)";


(* ::Subsection::Closed:: *)
(*Exported Functions*)


(* ::Input::Initialization:: *)
SimulateMonth::usage="SimulateMonth[Investment[...]] will simulate a month of time passing on the given Investment. SimulateMonth[Basket[...]] will simulate a month of time passing on the given basket of Investments and reinvest any outstanding funds.";


(* ::Input::Initialization:: *)
TotalValue::usage="TotalValue[Basket[...]] will return the current outstanding funds and principals in all of the outstanding loans. (As such it doesn't account for the future beahviour and is just a snapshot of the state of the investments at the given time)"


(* ::Input::Initialization:: *)
APR::usage="APR[y][r] will return the annulized rate of return over y years of rate r."


(* ::Input::Initialization:: *)
$ChargeOffMultiplier::usage="$ChargeOffMultiplier defaults to 1 but can be set to represent easier or harder economic times. It simply multiplies the probability of and account charging off at any one time."


(* ::Subsection::Closed:: *)
(*BeginPrivate*)


(* ::Input::Initialization:: *)
Begin["`Private`"]


(* ::Subchapter::Closed:: *)
(*Import*)


(* ::Input:: *)
(*data=Import["/Users/jason/Desktop/figures.csv"];*)


(* ::Input::Initialization:: *)
data={{"A1",0.0699`,0.0005`,0.3725`},{"A2",0.0799`,0.001`,0.3725`},{"A3",0.0899`,0.0011`,0.3725`},{"A4",0.0999`,0.0012`,0.3725`},{"A5",0.1099`,0.0014`,0.3725`},{"B1",0.1249`,0.0016`,0.2559`},{"B2",0.1349`,0.0018`,0.2559`},{"B3",0.1449`,0.0021`,0.2559`},{"B4",0.1499`,0.0025`,0.2559`},{"B5",0.1549`,0.003`,0.2559`},{"C1",0.1649`,0.0037`,0.1765`},{"C2",0.1749`,0.0046`,0.1765`},{"C3",0.1799`,0.0056`,0.1765`},{"C4",0.1899`,0.007`,0.1765`},{"C5",0.1949`,0.0089`,0.1765`},{"D1",0.2049`,0.0113`,0.1648`},{"D2",0.2149`,0.0142`,0.1648`},{"D3",0.2199`,0.0178`,0.1648`},{"D4",0.2249`,0.0218`,0.1648`},{"D5",0.2299`,0.026`,0.1648`},{"E1",0.2399`,0.0314`,0.0276`},{"E2",0.2492`,0.037`,0.0276`},{"E3",0.256`,0.0438`,0.0276`},{"E4",0.2627`,0.0512`,0.0276`},{"E5",0.2695`,0.0584`,0.0276`},{"F1",0.2763`,0.0662`,0.0027`},{"F2",0.2831`,0.0734`,0.0027`},{"F3",0.2898`,0.0809`,0.0027`},{"F4",0.2966`,0.0878`,0.0027`},{"F5",0.2999`,0.0949`,0.0027`}};


(* ::Subchapter::Closed:: *)
(*Grades*)


(* ::Input::Initialization:: *)
process[grade_,rate_,defaultRate_,rewriteRate_]:=grade->Association["grading"->grade,"annualRate"->rate,"defaultRate"->defaultRate,"rewriteRate"->rewriteRate]


(* ::Input::Initialization:: *)
grades=Association@@process@@@data


(* ::Input::Initialization:: *)
grades[["A1"]][["annualRate"]]


(* ::Subchapter::Closed:: *)
(*Hazard Curves*)


(* ::Text:: *)
(*We could model the hazard curves via a PoissonDistribution but it doesn't fit so well.*)


(* ::Input:: *)
(*DiscretePlot[Table[PDF[PoissonDistribution[\[Lambda]],k],{\[Lambda],{5,10,20}}]//Evaluate,{k,0,36},PlotRange->All,PlotMarkers->Automatic]*)


(* ::Text:: *)
(*I am very roughly modeling the hazard curve on a Poisson distribution. It is likely a little bit flatter than this*)


(* ::Input::Initialization:: *)
chargeOffHazard = Table[PDF[PoissonDistribution[12],k]//Evaluate//N,{k,0,36}]


(* ::Input::Initialization:: *)
rewriteHazard = Table[PDF[PoissonDistribution[12],k]//Evaluate//N,{k,0,60}]


(* ::Input:: *)
(*DiscretePlot[PDF[PoissonDistribution[12],k]//Evaluate,{k,0,36},PlotRange->All,PlotMarkers->Automatic]*)


(* ::Input:: *)
(*Plus@@rewriteHazard*)


(* ::Text:: *)
(*Instead I have directly typed the values in from Harmoney*)


(* ::Input::Initialization:: *)
chargeOffHazard = PadRight[{0.2,1.0,2.0,4.0,4.1,5.0,5.6,5.6,5.4,6.5,4.7,7.4,4.3,3.5,5.8,3.0,3.2,4.9,3.9,5.0,2.5,2.8,1.6,1.8,0.8,2.8,1.2,1.0,0.7}/100,60];


(* ::Input::Initialization:: *)
$ChargeOffMultiplier=1.0


(* ::Input:: *)
(*ListPlot[Take[chargeOffHazard,30],Joined->True]*)


(* ::Subchapter::Closed:: *)
(*P2P investment*)


(* ::Subsection::Closed:: *)
(*Constants*)


(* ::Text:: *)
(*We can break up our loans into $25 dollar chunks.*)


(* ::Input::Initialization:: *)
FractionalizationAmount=25;


(* ::Text:: *)
(*Harmoney charges 17.25% interest if the capital invested is over NZ$10,000*)


(* ::Input::Initialization:: *)
HarmoneyFeeRate=0.175;


(* ::Subsection::Closed:: *)
(*Monthly Amounts*)


(* ::Text:: *)
(*The annual rate is related to the monthly rate via: Subscript[r, annual]==(1+Subscript[r, month])^12-1 ie Subscript[r, month]==(1+Subscript[r, annual])^(1/12)-1 *)


(* ::Input::Initialization:: *)
monthlyInterestRate[anualRate_]:=-1+(1+anualRate)^(1/12)


(* ::Text:: *)
(*The monthly payment is calculated via the Amortization schedule A==P i (1+i)^n/((1+i)^n-1)*)


(* ::Input::Initialization:: *)
monthlyPayment[principal_,monthlyRate_,months_]:=principal monthlyRate (1+monthlyRate)^months/((1+monthlyRate)^months-1)


(* ::Input:: *)
(*monthlyPayment[5000,0.005001750284811202`,12]*)


(* ::Subsection::Closed:: *)
(*CreateInvestment*)


(* ::Input::Initialization:: *)
CreateInvestment[grade_Association,loan_, months_]:=
With[{
funds=0,
principal = loan,
month = 0,
monthlyRate=monthlyInterestRate[grade[["annualRate"]]],
grading=grade[["grading"]]},
With[{monthlyRepayment =monthlyPayment[loan,monthlyRate,months]},
Investment[funds,principal,month,months,monthlyRate,monthlyRepayment,grading]]]


(* ::Input:: *)
(*grades["F3"]*)


(* ::Input:: *)
(*aLone=CreateInvestment[grades["F3"],5000,36]*)


(* ::Subsection::Closed:: *)
(*AdvanceMonth*)


(* ::Text:: *)
(*When we advance an Investment by a month then we add to the funds by the payment less the harmoney fees, and the borrowers principal goes up by interest on the outstanding amount but down by his payment.*)


(* ::Input::Initialization:: *)
AdvanceMonth [inv: Investment[funds_,principal_,month_,months_,monthlyRate_,monthlyPayment_,grading_]]:=Investment[funds+monthlyPayment-principal*monthlyRate*HarmoneyFeeRate,
principal*(1+monthlyRate)-monthlyPayment,
month+1,months,monthlyRate,monthlyPayment,grading]


(* ::Subsection::Closed:: *)
(*Reinvest*)


(* ::Text:: *)
(*If the borrower pays off the loan early then all of the outstanding principal is paid back.*)


(* ::Input::Initialization:: *)
Reinvest@Investment[funds_,principal_,rest__]:=Investment[funds+principal,0,rest]


(* ::Subsection::Closed:: *)
(*SimulateMonth*)


(* ::Text:: *)
(*We simulate what happens in a month, which is one of three things, (i) the borrower defaults on the loan and we loose  the remaining principal, (ii) the borrower pays off the entire loan early (rewrites the loan), or (iii) the borrower pays the monthly payment and we advance on to the next month*)


(* ::Input::Initialization:: *)
SimulateMonth[inv: Investment[funds_,principal_,month_,months_,monthlyRate_,monthlyPayment_,grading_]]:=
With[{
rnd=RandomReal[],
chargeOffProb=$ChargeOffMultiplier*chargeOffHazard[[month+1]]*grades[grading]["defaultRate"],
rewriteProb=rewriteHazard[[month+1]]*grades[grading]["rewriteRate"]
},
Which[
chargeOffProb+rewriteProb<=rnd,AdvanceMonth@inv,
chargeOffProb<=rnd<chargeOffProb+rewriteProb,Reinvest@AdvanceMonth@inv,
True,Investment[funds,0,month,months,monthlyRate,monthlyPayment,grading]
]]


(* ::Text:: *)
(*SimulateMonth on a basket of investments, will just simulate the month on each investment, followed by retiring loans which are finished, followed by pooling all payments received, followed by reinvesting all money we have left.*)


(* ::Input::Initialization:: *)
SimulateMonth@Basket[funds_,investments:{___Investment},choice_]:=EndOfMonth@Harvest@RetireFinishedLoans@Basket[funds,SimulateMonth/@investments,choice]


(* ::Text:: *)
(*If a loan has run it's complete course then all payments by the borrower have been made (We might want to put a check in to ensure that*)


(* ::Subsection::Closed:: *)
(*RetireFinishedLoans*)


(* ::Input::Initialization:: *)
RetireFinishedLoans@Basket[funds_,investments:{___Investment}, choice_]:=
Basket[funds,investments/.Investment[f_,p_,m_,m_,rest__]:>Investment[f+p,0,m,m,rest],choice]


(* ::Text:: *)
(*Harvest will transfer all the accumulated funds from each investment into the funds of the basket of investments. (Basically it is just collecting all payments to the lender into a single place)*)


(* ::Subsection::Closed:: *)
(*Harvest*)


(* ::Input::Initialization:: *)
Harvest@Basket[funds_,investments:{___Investment}, choice_]:=
Basket[funds+Plus@@Cases[investments,Investment[f_,rest__]:>f],
investments/.Investment[f_,rest__]:>Investment[0,rest],
choice]


(* ::Subsection::Closed:: *)
(*EndOfMonth*)


(* ::Input::Initialization:: *)
EndOfMonth@Basket[funds_,investments:{___Investment},choice_]:=Module[{
cleanedInvestments=investments/.Investment[0,0,rest__]:>Sequence[],
numNewLoans = Quotient[funds,FractionalizationAmount],
newLoanGrades,
newLoans,
newFunds
},
newFunds = funds - FractionalizationAmount*numNewLoans;
newLoanGrades=RandomChoice[choice,numNewLoans];
newLoans=CreateInvestment[grades@#,FractionalizationAmount,36]&/@newLoanGrades;
Basket[newFunds,Join[cleanedInvestments,newLoans],choice]
]


(* ::Subsection::Closed:: *)
(*TotalValue*)


(* ::Text:: *)
(*TotalValue, will give us a snapshot of the total value of funds and principal outstanding*)


(* ::Input::Initialization:: *)
TotalValue@Basket[funds_,investments:{___Investment},choice_]:=
funds+Plus@@Cases[investments,Investment[f_,p_,rest__]:>f+p]


(* ::Subsection::Closed:: *)
(*TotalValue*)


(* ::Input::Initialization:: *)
APR[years_][rate_]:=100(rate^(1/years)-1)


(* ::Subchapter::Closed:: *)
(*End Package*)


(* ::Input::Initialization:: *)
End[ ]

EndPackage[ ]


(* ::Chapter:: *)
(*Simulations*)


(* ::Subchapter::Closed:: *)
(*Simulations*)


(* ::Input:: *)
(*NestList[SimulateMonth,Basket[25,{},{"B1","B2"}],37]*)


(* ::Input:: *)
(*TotalValue@Nest[SimulateMonth,Basket[2000,{},{"C1"}],37]/2000*)


(* ::Subchapter::Closed:: *)
(*B gradeA*)


(* ::Input:: *)
(*Bdata5000A=#^(1/3)&/@Table[TotalValue@Nest[SimulateMonth,Basket[5000,{},{"B1","B2","B3","B4","B5"}],37]/5000,10000];*)


(* ::Input:: *)
(*Histogram[Bdata5000A,50,"PDF"]*)


(* ::Input:: *)
(*Histogram[Bdata5000A,100,"PDF"]*)


(* ::Subchapter::Closed:: *)
(*C grade*)


(* ::Subsection::Closed:: *)
(*C Grade 5000*)


(* ::Input:: *)
(*Cdata5000=#^(1/3)&/@Table[TotalValue@Nest[SimulateMonth,Basket[5000,{},{"C1","C2","C3","C4","C5"}],37]/5000,10000];*)


(* ::Input:: *)
(*Histogram[Cdata5000,50,"ProbabilityDensity"]*)


(* ::Subsection::Closed:: *)
(*C GradeA 2500*)


(* ::Input:: *)
(*Cdata2500=#^(1/3)&/@Table[TotalValue@Nest[SimulateMonth,Basket[2500,{},{"C1","C2","C3","C4","C5"}],37]/2500,10000];*)


(* ::Input:: *)
(*Histogram[Cdata2500,100,"ProbabilityDensity"]*)


(* ::Subsection::Closed:: *)
(*C Grade 1500*)


(* ::Input:: *)
(*Cdata1500=#^(1/3)&/@Table[TotalValue@Nest[SimulateMonth,Basket[1500,{},{"C1","C2","C3","C4","C5"}],37]/1500,10000];*)


(* ::Input:: *)
(*Histogram[Cdata1500,100,"ProbabilityDensity"]*)


(* ::Subchapter::Closed:: *)
(*D grade*)


(* ::Input:: *)
(*Ddata5000=#^(1/3)&/@Table[TotalValue@Nest[SimulateMonth,Basket[5000,{},{"D1","D2","D3","D4","D5"}],37]/5000,10000];*)


(* ::Input:: *)
(*Histogram[Ddata5000,50,"PDF"]*)


(* ::Subchapter::Closed:: *)
(*E grade*)


(* ::Input:: *)
(*Edata5000=#^(1/3)&/@Table[TotalValue@Nest[SimulateMonth,Basket[5000,{},{"E1","E2","E3","E4","E5"}],37]/5000,10000];*)


(* ::Input:: *)
(*Histogram[Edata5000,50,"PDF"]*)


(* ::Subchapter::Closed:: *)
(*F grade*)


(* ::Input:: *)
(*Fdata5000=#^(1/3)&/@Table[TotalValue@Nest[SimulateMonth,Basket[5000,{},{"F1","F2","F3","F4","F5"}],37]/5000,10000];*)


(* ::Input:: *)
(*Histogram[Fdata5000,50,"PDF"]*)
