(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    278245,       6431]
NotebookOptionsPosition[    249248,       5754]
NotebookOutlinePosition[    272774,       6283]
CellTagsIndexPosition[    271916,       6263]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell[BoxData[
 FormBox[
  RowBox[{"14.2", "Line", " ", "Integrals"}], SectionTitleFont]], "Section",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.479166409989386*^9, 3.479166416577236*^9}},
 CellTags->"14.2 Line Integrals"],

Cell[TextData[{
 "With integrals of a single variable, we integrate over intervals in ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[DoubleStruckCapitalR]", "1"], TraditionalForm]]],
 "(the real line). With double and triple integrals, we integrate over \
regions in ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[DoubleStruckCapitalR]", "2"], TraditionalForm]]],
 " or ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[DoubleStruckCapitalR]", "3"], TraditionalForm]]],
 ". ",
 StyleBox["Line integrals",
  FontSlant->"Italic"],
 " (which really should be called ",
 StyleBox["curve integrals",
  FontSlant->"Italic"],
 ") are another class of integrals that play an important role in vector \
calculus. They are used to integrate either scalar-valued functions or vector \
fields along curves."
}], "Text",
 CellChangeTimes->{{3.479166434358184*^9, 3.4791665371474123`*^9}}],

Cell[TextData[{
 "\tSuppose a thin, circular plate has a known temperature distribution and \
you must compute the average temperature along the edge of the plate. The \
required calculation involves integrating the temperature function over the ",
 StyleBox["curved",
  FontSlant->"Italic"],
 " boundary of the plate. Similarly, to calculate the amount of work needed \
to put a satellite into orbit, we integrate the gravitational force (a vector \
field) along the curved path of the satellite. Both these calculations \
require line integrals. As you will see, line integrals take several \
different forms. It is the goal of this section to distinguish these various \
forms and show how and when each form should be used."
}], "Text",
 CellChangeTimes->{{3.479166540586193*^9, 3.479166670186523*^9}}],

Cell[CellGroupData[{

Cell["Scalar Line Integrals in the Plane", "Subsection",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.479166677125717*^9, 3.479166683590638*^9}},
 CellTags->"Scalar Line Integrals in the Plane"],

Cell[TextData[{
 "We first consider line integrals of scalar-valued functions over curves in \
the plane. ",
 StyleBox["Figure 14.16", "FigureFontText"],
 " shows a surface ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", "=", 
    RowBox[{"f", "(", 
     RowBox[{"x", ",", "y"}], ")"}]}], TraditionalForm]]],
 " and a parameterized curve ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " in the ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "\[VeryThinSpace]", "y"}], TraditionalForm]]],
 "-plane; for the moment we assume that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"f", "(", 
     RowBox[{"x", ",", "y"}], ")"}], "\[GreaterEqual]", "0"}], 
   TraditionalForm]]],
 " for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{"x", ",", "y"}], ")"}], TraditionalForm]]],
 " on ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". Now visualize the curtain-like surface formed by the vertical line \
segments joining the surface ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", "=", 
    RowBox[{"f", "(", 
     RowBox[{"x", ",", "y"}], ")"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". The goal is to find the area of one side of this curtain in terms of a \
line integral. As with other integrals we have studied, we begin with Riemann \
sums."
}], "Text",
 CellChangeTimes->{{3.4791667615118513`*^9, 3.479166901812745*^9}, 
   3.479166932277637*^9, 3.479777469894005*^9}],

Cell[BoxData[
 TagBox[
  StyleBox[
   DynamicModuleBox[{$CellContext`aValue$$ = 0., $CellContext`bValue$$ = 
    2, $CellContext`showArea$$ = True, $CellContext`showCurve$$ = 
    True, $CellContext`showGrids$$ = False, $CellContext`showLabels$$ = 
    True, $CellContext`showSurface$$ = True, $CellContext`xMax$$ = 
    5, $CellContext`xMin$$ = -1, $CellContext`yMax$$ = 
    5, $CellContext`yMin$$ = -1, $CellContext`zMax$$ = 
    5, $CellContext`zMin$$ = -1, Typeset`show$$ = True, 
    Typeset`bookmarkList$$ = {}, Typeset`bookmarkMode$$ = "Menu", 
    Typeset`animator$$, Typeset`animvar$$ = 1, Typeset`name$$ = 
    "\"untitled\"", Typeset`specs$$ = {{{
       Hold[$CellContext`showSurface$$], True, "show surface"}, {
      True, False}}, {{
       Hold[$CellContext`showCurve$$], True, 
       "show curve \!\(\*\nStyleBox[\"C\",\nFontSlant->\"Italic\"]\)"}, {
      True, False}}, {{
       Hold[$CellContext`showArea$$], True, "show curtain"}, {True, False}}, {{
       Hold[$CellContext`aValue$$], 0, 
       "\!\(\*\nStyleBox[\"a\",\nFontSlant->\"Italic\"]\)"}, 0, 1.99}, {
      Hold[
       Row[{
         Manipulate`Place[1]}]], Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`bValue$$], 2, 
       "\!\(\*\nStyleBox[\"b\",\nFontSlant->\"Italic\"]\)"}, 0.1, 2}, {
      Hold[
       Row[{
         Manipulate`Place[2]}]], Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`showLabels$$], True, "show labels"}, {
      True, False}}, {{
       Hold[$CellContext`showGrids$$], False, "show grids"}, {True, False}}, {
      Hold[
       Grid[{{
          Manipulate`Place[3]}, {
          Manipulate`Place[4]}}, Alignment -> Right]], 
      Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`xMin$$], -1}, 0}, {{
       Hold[$CellContext`xMax$$], 5}, 0}, {{
       Hold[$CellContext`yMin$$], -1}, 0}, {{
       Hold[$CellContext`yMax$$], 5}, 0}, {{
       Hold[$CellContext`zMin$$], -1}, 0}, {{
       Hold[$CellContext`zMax$$], 5}, 0}}, Typeset`size$$ = {
    360., {177., 183.}}, Typeset`update$$ = 0, Typeset`initDone$$, 
    Typeset`skipInitDone$$ = False, $CellContext`showSurface$388915$$ = 
    False, $CellContext`showCurve$388916$$ = 
    False, $CellContext`showArea$388917$$ = 
    False, $CellContext`aValue$388918$$ = 0, $CellContext`bValue$388919$$ = 
    0, $CellContext`showLabels$388920$$ = 
    False, $CellContext`showGrids$388921$$ = False}, 
    DynamicBox[Manipulate`ManipulateBoxes[
     2, StandardForm, 
      "Variables" :> {$CellContext`aValue$$ = 0, $CellContext`bValue$$ = 
        2, $CellContext`showArea$$ = True, $CellContext`showCurve$$ = 
        True, $CellContext`showGrids$$ = False, $CellContext`showLabels$$ = 
        True, $CellContext`showSurface$$ = True, $CellContext`xMax$$ = 
        5, $CellContext`xMin$$ = -1, $CellContext`yMax$$ = 
        5, $CellContext`yMin$$ = -1, $CellContext`zMax$$ = 
        5, $CellContext`zMin$$ = -1}, "ControllerVariables" :> {
        Hold[$CellContext`showSurface$$, $CellContext`showSurface$388915$$, 
         False], 
        Hold[$CellContext`showCurve$$, $CellContext`showCurve$388916$$, 
         False], 
        Hold[$CellContext`showArea$$, $CellContext`showArea$388917$$, False], 
        Hold[$CellContext`aValue$$, $CellContext`aValue$388918$$, 0], 
        Hold[$CellContext`bValue$$, $CellContext`bValue$388919$$, 0], 
        Hold[$CellContext`showLabels$$, $CellContext`showLabels$388920$$, 
         False], 
        Hold[$CellContext`showGrids$$, $CellContext`showGrids$388921$$, 
         False]}, 
      "OtherVariables" :> {
       Typeset`show$$, Typeset`bookmarkList$$, Typeset`bookmarkMode$$, 
        Typeset`animator$$, Typeset`animvar$$, Typeset`name$$, 
        Typeset`specs$$, Typeset`size$$, Typeset`update$$, Typeset`initDone$$,
         Typeset`skipInitDone$$}, "Body" :> Show[{
         If[$CellContext`showGrids$$, 
          $CellContext`CoordinatePlane[
          0, {$CellContext`y, $CellContext`yMin$$, $CellContext`yMax$$}, \
{$CellContext`z, $CellContext`zMin$$, $CellContext`zMax$$}, \
$CellContext`PlaneColor -> GrayLevel[0.9], $CellContext`PlaneOpacity -> 
           Opacity[0.3], $CellContext`PlaneGrids -> True], 
          Graphics3D[{}]], 
         If[$CellContext`showGrids$$, 
          $CellContext`CoordinatePlane[{$CellContext`x, $CellContext`xMin$$, \
$CellContext`xMax$$}, 
           0, {$CellContext`z, $CellContext`zMin$$, $CellContext`zMax$$}, \
$CellContext`PlaneColor -> GrayLevel[0.9], $CellContext`PlaneOpacity -> 
           Opacity[0.3], $CellContext`PlaneGrids -> True], 
          Graphics3D[{}]], 
         If[$CellContext`showGrids$$, 
          $CellContext`CoordinatePlane[{$CellContext`x, $CellContext`xMin$$, \
$CellContext`xMax$$}, {$CellContext`y, $CellContext`yMin$$, \
$CellContext`yMax$$}, 0, $CellContext`PlaneColor -> 
           GrayLevel[0.9], $CellContext`PlaneOpacity -> 
           Opacity[0.3], $CellContext`PlaneGrids -> True], 
          Graphics3D[{}]], 
         Graphics3D[{
           Arrowheads[{0, 0.02}], 
           AbsoluteThickness[0.5], 
           Arrow[{{$CellContext`xMin$$, 0, 0}, {$CellContext`xMax$$, 0, 0}}], 
           Arrow[{{0, $CellContext`yMin$$, 0}, {0, $CellContext`yMax$$, 0}}], 
           Arrow[{{0, 0, $CellContext`zMin$$}, {0, 0, $CellContext`zMax$$}}], 
           Text[
           "\!\(\*\nStyleBox[\"x\",\nFontSlant->\"Italic\"]\)", \
{$CellContext`xMax$$, 0, 0}, {2, 0}], 
           Text[
           "\!\(\*\nStyleBox[\"y\",\nFontSlant->\"Italic\"]\)", {
            0, $CellContext`yMax$$, 0}, {-2, 0}], 
           Text[
           "\!\(\*\nStyleBox[\"z\",\nFontSlant->\"Italic\"]\)", {
            0, 0, $CellContext`zMax$$}, {0, -2}]}], 
         If[$CellContext`showSurface$$, $CellContext`plotC14F16, 
          Graphics3D[{}]], 
         If[$CellContext`showCurve$$, 
          ParametricPlot3D[
           $CellContext`funcC14F16Curve[$CellContext`s, 
            0], {$CellContext`s, $CellContext`aValue$$, \
$CellContext`bValue$$}, PlotStyle -> {{Thick, Black}}], 
          Graphics3D[{}]], 
         If[$CellContext`showArea$$, {
           ParametricPlot3D[$CellContext`funcC14F16Curve[$CellContext`s, 
              $CellContext`funcC14F16[
               Part[
                $CellContext`funcC14F16Curve[$CellContext`s, 0], 1], 
               Part[
                $CellContext`funcC14F16Curve[$CellContext`s, 0], 2]]] + 
            0.01, {$CellContext`s, $CellContext`aValue$$, \
$CellContext`bValue$$}, PlotStyle -> {{Thick, $CellContext`bcR}}], 
           ParametricPlot3D[
            $CellContext`funcC14F16Curve[$CellContext`s, $CellContext`u \
$CellContext`funcC14F16[
               Part[
                $CellContext`funcC14F16Curve[$CellContext`s, 0], 1], 
               Part[
                $CellContext`funcC14F16Curve[$CellContext`s, 0], 
                2]]], {$CellContext`u, 0, 
             1}, {$CellContext`s, $CellContext`aValue$$, \
$CellContext`bValue$$}, Mesh -> None, PlotStyle -> {
              Opacity[0.8], $CellContext`bcAreaPositiveColor}]}, 
          Graphics3D[{}]], 
         If[
          And[$CellContext`showLabels$$, $CellContext`showArea$$], 
          Graphics3D[{
            Text[
             Framed[
              Pane[
              "Portion of surface \!\(\*FormBox[\n RowBox[{\"z\", \"=\", \n  \
RowBox[{\"f\", \"(\", \n   RowBox[{\"x\", \",\", \"y\"}], \")\"}]}],\n \
TraditionalForm]\) over \!\(\*FormBox[\"C\",\n TraditionalForm]\)", \
$CellContext`bcPBS], $CellContext`bcFO, Background -> White], {
             2.5, 2.5, $CellContext`zMax$$ + 0.5}, {0, 1}]}], 
          Graphics3D[{}]]}, 
        PlotRange -> {{$CellContext`xMin$$, $CellContext`xMax$$}, \
{$CellContext`yMin$$, $CellContext`yMax$$}, {$CellContext`zMin$$, \
$CellContext`zMax$$ + 0.5}}, ImageSize -> 5 {72, 72}, 
        BaseStyle -> {"Text", 13}, Lighting -> "Neutral", Boxed -> False, 
        ViewPoint -> {4, 3, 2}, Epilog -> {
          If[
           And[$CellContext`showLabels$$, $CellContext`showArea$$], 
           Inset[
            Framed[
             Pane[
             "Area of curtain\n\!\(\*FormBox[\n RowBox[{\"\
\[InvisiblePostfixScriptBase]\", \"=\", \n  \
RowBox[{SubscriptBox[\"\[Integral]\", \n    RowBox[{\" \", \"C\"}]], \n   \
RowBox[{\"f\", \" \", \"d\", \"\[VeryThinSpace]\", \"s\"}]}]}],\n \
TraditionalForm]\) ", $CellContext`bcPBS], $CellContext`bcFO], 
            ImageScaled[{0.99, 0.01}], 
            ImageScaled[{1, 0}]], Black], 
          If[
           And[$CellContext`showLabels$$, $CellContext`showCurve$$], 
           Inset[
            Framed[
             Pane[
             " \!\(\*FormBox[\n RowBox[{\n  RowBox[{\n   RowBox[{\"C\", \
\":\", \"x\"}], \"=\", \n   RowBox[{\"x\", \"(\", \"s\", \")\"}]}], \",\", \n \
 RowBox[{\"y\", \"=\", \n   RowBox[{\"y\", \"(\", \"s\", \")\"}]}], \",\", \n \
 RowBox[{\"z\", \"=\", \"0\"}]}],\n TraditionalForm]\)\n\!\(\*FormBox[\n \
RowBox[{\"a\", \"\[LessEqual]\", \"s\", \"\[LessEqual]\", \"b\"}],\n \
TraditionalForm]\)", $CellContext`bcPBS], $CellContext`bcFO], 
            ImageScaled[{0.01, 0.01}], 
            ImageScaled[{0, 0}]], Black]}], 
      "Specifications" :> {{{$CellContext`showSurface$$, True, 
          "show surface"}, {True, False}, ControlType -> 
         Checkbox}, {{$CellContext`showCurve$$, True, 
          "show curve \!\(\*\nStyleBox[\"C\",\nFontSlant->\"Italic\"]\)"}, {
         True, False}, ControlType -> 
         Checkbox}, {{$CellContext`showArea$$, True, "show curtain"}, {
         True, False}, ControlType -> Checkbox}, 
        Delimiter, {{$CellContext`aValue$$, 0, 
          "\!\(\*\nStyleBox[\"a\",\nFontSlant->\"Italic\"]\)"}, 0, 1.99, 
         ControlType -> Slider, ImageSize -> Small, ControlPlacement -> 1}, 
        Row[{
          Manipulate`Place[1]}], {{$CellContext`bValue$$, 2, 
          "\!\(\*\nStyleBox[\"b\",\nFontSlant->\"Italic\"]\)"}, 0.1, 2, 
         ControlType -> Slider, ImageSize -> Small, ControlPlacement -> 2}, 
        Row[{
          Manipulate`Place[2]}], 
        Delimiter, {{$CellContext`showLabels$$, True, "show labels"}, {
         True, False}, ControlType -> Checkbox, ControlPlacement -> 
         3}, {{$CellContext`showGrids$$, False, "show grids"}, {True, False}, 
         ControlType -> Checkbox, ControlPlacement -> 4}, 
        Grid[{{
           Manipulate`Place[3]}, {
           Manipulate`Place[4]}}, Alignment -> 
         Right], {{$CellContext`xMin$$, -1}, 0, ControlType -> 
         None}, {{$CellContext`xMax$$, 5}, 0, ControlType -> 
         None}, {{$CellContext`yMin$$, -1}, 0, ControlType -> 
         None}, {{$CellContext`yMax$$, 5}, 0, ControlType -> 
         None}, {{$CellContext`zMin$$, -1}, 0, ControlType -> 
         None}, {{$CellContext`zMax$$, 5}, 0, ControlType -> None}}, 
      "Options" :> {
       ControlPlacement -> Right, LabelStyle -> 11, AppearanceElements -> 
        "ResetButton", FrameLabel -> 
        Style["FIGURE 14.16", "FigureFont", Smaller]}, "DefaultOptions" :> {}],
     ImageSizeCache->{521., {213., 219.}},
     SingleEvaluation->True],
    Deinitialization:>None,
    DynamicModuleValues:>{},
    Initialization:>({{$CellContext`CoordinatePlane[
           Pattern[$CellContext`x, 
            Blank[]], 
           Pattern[$CellContext`y, 
            Blank[]], 
           Pattern[$CellContext`z, 
            Blank[]]] := Graphics3D[{
            Lighter[$CellContext`bcSurfaceBottom], 
            Opacity[0.3], 
            Switch[
             If[
              ListQ[$CellContext`x], 
              If[
               ListQ[$CellContext`y], "z", "y"], "x"], "x", 
             Polygon[{{$CellContext`x, 
                Part[$CellContext`y, 2], 
                Part[$CellContext`z, 2]}, {$CellContext`x, 
                Part[$CellContext`y, 2], 
                Part[$CellContext`z, 3]}, {$CellContext`x, 
                Part[$CellContext`y, 3], 
                Part[$CellContext`z, 3]}, {$CellContext`x, 
                Part[$CellContext`y, 3], 
                Part[$CellContext`z, 2]}, {$CellContext`x, 
                Part[$CellContext`y, 2], 
                Part[$CellContext`z, 2]}}], "y", 
             Polygon[{{
                Part[$CellContext`x, 2], $CellContext`y, 
                Part[$CellContext`z, 2]}, {
                Part[$CellContext`x, 2], $CellContext`y, 
                Part[$CellContext`z, 3]}, {
                Part[$CellContext`x, 3], $CellContext`y, 
                Part[$CellContext`z, 3]}, {
                Part[$CellContext`x, 3], $CellContext`y, 
                Part[$CellContext`z, 2]}, {
                Part[$CellContext`x, 2], $CellContext`y, 
                Part[$CellContext`z, 2]}}], "z", 
             Polygon[{{
                Part[$CellContext`x, 2], 
                Part[$CellContext`y, 2], $CellContext`z}, {
                Part[$CellContext`x, 2], 
                Part[$CellContext`y, 3], $CellContext`z}, {
                Part[$CellContext`x, 3], 
                Part[$CellContext`y, 3], $CellContext`z}, {
                Part[$CellContext`x, 3], 
                Part[$CellContext`y, 2], $CellContext`z}, {
                Part[$CellContext`x, 2], 
                
                Part[$CellContext`y, 
                 2], $CellContext`z}}]]}], $CellContext`CoordinatePlane[
           Pattern[$CellContext`x, 
            Blank[]], 
           Pattern[$CellContext`y, 
            Blank[]], 
           Pattern[$CellContext`z, 
            Blank[]], 
           Pattern[$CellContext`opts, 
            BlankNullSequence[]]] := 
         Module[{$CellContext`color, $CellContext`opacity, \
$CellContext`grids}, {$CellContext`color, $CellContext`opacity, \
$CellContext`grids} = ReplaceAll[
              
              ReplaceAll[{$CellContext`PlaneColor, $CellContext`PlaneOpacity, \
$CellContext`PlaneGrids}, {$CellContext`opts}], 
              Options[$CellContext`CoordinatePlane]]; Graphics3D[{
              If[$CellContext`grids, 
               Switch[
                If[
                 ListQ[$CellContext`x], 
                 If[
                  ListQ[$CellContext`y], "z", "y"], "x"], "x", {Gray, 
                 Opacity[0.5], Thin, 
                 Table[
                  Line[{{$CellContext`x, 
                    Part[$CellContext`y, 2], $CellContext`i}, {$CellContext`x, 
                    
                    Part[$CellContext`y, 
                    3], $CellContext`i}}], {$CellContext`i, 
                   Part[$CellContext`z, 2], 
                   Part[$CellContext`z, 3], 
                   If[Length[$CellContext`z] == 4, 
                    Part[$CellContext`z, 4], 1]}], 
                 Table[
                  Line[{{$CellContext`x, $CellContext`i, 
                    Part[$CellContext`z, 2]}, {$CellContext`x, $CellContext`i, 
                    Part[$CellContext`z, 3]}}], {$CellContext`i, 
                   Part[$CellContext`y, 2], 
                   Part[$CellContext`y, 3], 
                   If[Length[$CellContext`y] == 4, 
                    Part[$CellContext`y, 4], 1]}]}, "y", {Gray, 
                 Opacity[0.5], Thin, 
                 Table[
                  Line[{{
                    Part[$CellContext`x, 2], $CellContext`y, $CellContext`i}, {
                    
                    Part[$CellContext`x, 
                    3], $CellContext`y, $CellContext`i}}], {$CellContext`i, 
                   Part[$CellContext`z, 2], 
                   Part[$CellContext`z, 3], 
                   If[Length[$CellContext`z] == 4, 
                    Part[$CellContext`z, 4], 1]}], 
                 Table[
                  Line[{{$CellContext`i, $CellContext`y, 
                    Part[$CellContext`z, 2]}, {$CellContext`i, $CellContext`y, 
                    Part[$CellContext`z, 3]}}], {$CellContext`i, 
                   Part[$CellContext`x, 2], 
                   Part[$CellContext`x, 3], 
                   If[Length[$CellContext`x] == 4, 
                    Part[$CellContext`x, 4], 1]}]}, "z", {Gray, 
                 Opacity[0.5], Thin, 
                 Table[
                  Line[{{
                    Part[$CellContext`x, 2], $CellContext`i, $CellContext`z}, {
                    
                    Part[$CellContext`x, 
                    3], $CellContext`i, $CellContext`z}}], {$CellContext`i, 
                   Part[$CellContext`y, 2], 
                   Part[$CellContext`y, 3], 
                   If[Length[$CellContext`y] == 4, 
                    Part[$CellContext`y, 4], 1]}], 
                 Table[
                  Line[{{$CellContext`i, 
                    Part[$CellContext`y, 2], $CellContext`z}, {$CellContext`i, 
                    
                    Part[$CellContext`y, 
                    3], $CellContext`z}}], {$CellContext`i, 
                   Part[$CellContext`x, 2], 
                   Part[$CellContext`x, 3], 
                   If[Length[$CellContext`x] == 4, 
                    Part[$CellContext`x, 4], 1]}]}], 
               Black], $CellContext`color, $CellContext`opacity, 
              Switch[
               If[
                ListQ[$CellContext`x], 
                If[
                 ListQ[$CellContext`y], "z", "y"], "x"], "x", 
               Polygon[{{$CellContext`x, 
                  Part[$CellContext`y, 2], 
                  Part[$CellContext`z, 2]}, {$CellContext`x, 
                  Part[$CellContext`y, 2], 
                  Part[$CellContext`z, 3]}, {$CellContext`x, 
                  Part[$CellContext`y, 3], 
                  Part[$CellContext`z, 3]}, {$CellContext`x, 
                  Part[$CellContext`y, 3], 
                  Part[$CellContext`z, 2]}, {$CellContext`x, 
                  Part[$CellContext`y, 2], 
                  Part[$CellContext`z, 2]}}], "y", 
               Polygon[{{
                  Part[$CellContext`x, 2], $CellContext`y, 
                  Part[$CellContext`z, 2]}, {
                  Part[$CellContext`x, 2], $CellContext`y, 
                  Part[$CellContext`z, 3]}, {
                  Part[$CellContext`x, 3], $CellContext`y, 
                  Part[$CellContext`z, 3]}, {
                  Part[$CellContext`x, 3], $CellContext`y, 
                  Part[$CellContext`z, 2]}, {
                  Part[$CellContext`x, 2], $CellContext`y, 
                  Part[$CellContext`z, 2]}}], "z", 
               Polygon[{{
                  Part[$CellContext`x, 2], 
                  Part[$CellContext`y, 2], $CellContext`z}, {
                  Part[$CellContext`x, 2], 
                  Part[$CellContext`y, 3], $CellContext`z}, {
                  Part[$CellContext`x, 3], 
                  Part[$CellContext`y, 3], $CellContext`z}, {
                  Part[$CellContext`x, 3], 
                  Part[$CellContext`y, 2], $CellContext`z}, {
                  Part[$CellContext`x, 2], 
                  Part[$CellContext`y, 2], $CellContext`z}}]]}]], 
         Options[$CellContext`CoordinatePlane] = {$CellContext`PlaneColor -> 
           RGBColor[
            0.6871111111111111, 0.7377777777777778, 
             0.7133333333333334], $CellContext`PlaneOpacity -> 
           Opacity[0.7], $CellContext`PlaneGrids -> 
           False}, $CellContext`bcSurfaceBottom = 
         RGBColor[
          0.5306666666666667, 0.6066666666666667, 
           0.5700000000000001], $CellContext`i = 1, $CellContext`plotC14F16 = 
         Graphics3D[
           GraphicsComplex[CompressedData["
1:eJxdmnlcjtn//+/2tKlbIjIq9bFVSCnrfUVm3KPBhCiV7EuFYSZJacGIYj4l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            "], {{{
               EdgeForm[], 
               Opacity[0.8], 
               RGBColor[0.85, 0.6, 0.2], 
               Specularity[
                GrayLevel[1], 100], 
               GraphicsGroup[{
                 Polygon[CompressedData["
1:eJxNmgn41lMWx39Xsm+hkWwl7SilRUKStimSiIS/UqEFoSJUhEJKiii0EEmS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                  "]]}]}, {}, {}, {}, {}}, {{
               GrayLevel[0], 
               Line[CompressedData["
1:eJwl0Lsug2EYwPHvc2od6nADFYlNmEzSXgFdGkkpCxMGlg7EZClDE1tFJBiQ
NHSrK2CUSNSxWq076a/p8Pv+z/O+7/RNbe6ld8MgCNa4MnTmgk+EcfMkcyRZ
Yjrsvt3SHT3ghDPueCTl7lkzek/VfKmHvJs/+OSLb36o8UudBn+8eN/UV23p
qj7wbz7XFfK82Yu6TIkne04retu545gFZ/ta1nmO2Laf6gw3ZO0buqgJZokz
4eyaMfOFrjNqjjHCMEMMEiXCAP300UsP2bD7n9uRYisq
                "]]}}}, VertexNormals -> CompressedData["
1:eJwd2Hk4VG0YBnAlIZItRVFaaZVCKZ6EhKikooVIJC2Soo1IsiS7KJUlW7Ra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             "]], {
           Axes -> True, BoxRatios -> {1, 1, 0.4}, 
            Method -> {"RotationControl" -> "Globe"}, 
            PlotRange -> {All, All, All}, 
            PlotRangePadding -> {Automatic, Automatic, Automatic}}], 
         Attributes[PlotRange] = {ReadProtected}, $CellContext`funcC14F16Curve[
           Pattern[$CellContext`s, 
            Blank[]], 
           Pattern[$CellContext`z, 
            Blank[]]] := {
          4.5 - $CellContext`s^2, 2 $CellContext`s + 
           0.5, $CellContext`z}, $CellContext`funcC14F16[
           Pattern[$CellContext`x, 
            Blank[]], 
           Pattern[$CellContext`y, 
            Blank[]]] := (0.2 ($CellContext`x - 1)^2) 
           Sin[((2 Pi) $CellContext`y)/6] + 2, $CellContext`bcR = 
         RGBColor[0.8, 0, 0.4], $CellContext`bcAreaPositiveColor = 
         RGBColor[0.796, 0.91, 0.855], $CellContext`bcPBS = 
         BaseStyle -> {
           LineIndent -> 0, LinebreakAdjustments -> {1, 1, 0, 0, 0}, 
            LineBreakWithin -> Automatic, Hyphenation -> True, TextAlignment -> 
            Left, FontSize -> 13}, $CellContext`bcFO = {
          RoundingRadius -> 5, FrameStyle -> GrayLevel[0.85]}}; {Null, Null, 
         Graphics3D[
          GraphicsComplex[CompressedData["
1:eJxdmnlcjtn//+/2tKlbIjIq9bFVSCnrfUVm3KPBhCiV7EuFYSZJacGIYj4l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           "], {{{
              EdgeForm[], 
              Opacity[0.8], 
              RGBColor[0.85, 0.6, 0.2], 
              Specularity[
               GrayLevel[1], 100], 
              GraphicsGroup[{
                Polygon[CompressedData["
1:eJxNmgn41lMWx39Xsm+hkWwl7SilRUKStimSiIS/UqEFoSJUhEJKiii0EEmS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                 "]]}]}, {}, {}, {}, {}}, {{
              GrayLevel[0], 
              Line[CompressedData["
1:eJwl0Lsug2EYwPHvc2od6nADFYlNmEzSXgFdGkkpCxMGlg7EZClDE1tFJBiQ
NHSrK2CUSNSxWq076a/p8Pv+z/O+7/RNbe6ld8MgCNa4MnTmgk+EcfMkcyRZ
Yjrsvt3SHT3ghDPueCTl7lkzek/VfKmHvJs/+OSLb36o8UudBn+8eN/UV23p
qj7wbz7XFfK82Yu6TIkne04retu545gFZ/ta1nmO2Laf6gw3ZO0buqgJZokz
4eyaMfOFrjNqjjHCMEMMEiXCAP300UsP2bD7n9uRYisq
               "]]}}}, VertexNormals -> CompressedData["
1:eJwd2Hk4VG0YBnAlIZItRVFaaZVCKZ6EhKikooVIJC2Soo1IsiS7KJUlW7Ra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            "]], {
          Axes -> True, BoxRatios -> {1, 1, 0.4}, 
           Method -> {"RotationControl" -> "Globe"}, 
           PlotRange -> {All, All, All}, 
           PlotRangePadding -> {Automatic, Automatic, Automatic}}]}}; 
     Typeset`initDone$$ = True),
    SynchronousInitialization->True,
    UnsavedVariables:>{Typeset`initDone$$},
    UntrackedVariables:>{Typeset`size$$}], "Manipulate",
   Deployed->True,
   StripOnInput->False],
  Manipulate`InterpretManipulate[1]]], "Output",
 CellChangeTimes->{3.5181824476968412`*^9},
 CellTags->"FIGURE 14.16  \[LightBulb]"],

Cell[TextData[{
 "\tAssume that ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " is a smooth curve of finite length, parameterized in terms of arc length \
as ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     StyleBox["r",
      FontWeight->Bold,
      FontSlant->Plain], "(", "s", ")"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"x", "(", "s", ")"}], ",", 
      RowBox[{"y", "(", "s", ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[LessEqual]", "s", "\[LessEqual]", "b"}], TraditionalForm]]],
 ", and let ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " be defined on ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". We subdivide ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " into ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " small arcs by forming a partition of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}], TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.4791669515376*^9, 3.479167023596951*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{"a", "=", 
         RowBox[{
          RowBox[{
           SubscriptBox["s", "0"], "<", 
           SubscriptBox["s", "1"], "<", "\[CenterEllipsis]", "<", 
           SubscriptBox["s", 
            RowBox[{"n", "-", "1"}]], "<", 
           SubscriptBox["s", "n"]}], "=", 
          RowBox[{"b", "."}]}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479167029007524*^9, 3.479167053051861*^9}}],

Cell[TextData[{
 "Let ",
 Cell[BoxData[
  FormBox[
   SubscriptBox[
    OverscriptBox["s", "_"], "k"], TraditionalForm]]],
 " be a point in the ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[VeryThinSpace]", "th"}], TraditionalForm]]],
 " subinterval ",
 Cell[BoxData[
  FormBox[
   RowBox[{"[", 
    RowBox[{
     SubscriptBox["s", 
      RowBox[{"k", "-", "1"}]], ",", 
     SubscriptBox["s", "k"]}], "]"}], TraditionalForm]]],
 ", which corresponds to a point ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    RowBox[{
     RowBox[{"x", "(", 
      SubscriptBox[
       OverscriptBox["s", "_"], "k"], ")"}], ",", 
     RowBox[{"y", "(", 
      SubscriptBox[
       OverscriptBox["s", "_"], "k"], ")"}]}], ")"}], TraditionalForm]]],
 " on the ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[VeryThinSpace]", "th"}], TraditionalForm]]],
 " arc of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", for ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", "2", ",", "\[Ellipsis]", ",", "n"}], 
   TraditionalForm]]],
 ". The length of the ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[VeryThinSpace]", "th"}], TraditionalForm]]],
 " arc is denoted ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[CapitalDelta]", "\[VeryThinSpace]", 
    SubscriptBox["s", "k"]}], TraditionalForm]]],
 ". This partition also divides the curtain into ",
 Cell[BoxData[
  FormBox["n", TraditionalForm]]],
 " panels. The ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[VeryThinSpace]", "th"}], TraditionalForm]]],
 " panel has an approximate height of ",
 Cell[BoxData[
  FormBox[
   RowBox[{"f", "(", 
    RowBox[{
     RowBox[{"x", "(", 
      SubscriptBox[
       OverscriptBox["s", "_"], "k"], ")"}], ",", 
     RowBox[{"y", "(", 
      SubscriptBox[
       OverscriptBox["s", "_"], "k"], ")"}]}], ")"}], TraditionalForm]]],
 " and a base of length ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[CapitalDelta]", "\[VeryThinSpace]", 
    SubscriptBox["s", "k"]}], TraditionalForm]]],
 "; therefore, the approximate area of the ",
 Cell[BoxData[
  FormBox[
   RowBox[{"k", "\[VeryThinSpace]", "th"}], TraditionalForm]]],
 " panel is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"f", "(", 
     RowBox[{
      RowBox[{"x", "(", 
       SubscriptBox[
        OverscriptBox["s", "_"], "k"], ")"}], ",", 
      RowBox[{"y", "(", 
       SubscriptBox[
        OverscriptBox["s", "_"], "k"], ")"}]}], ")"}], "\[CapitalDelta]", 
    "\[VeryThinSpace]", 
    SubscriptBox["s", "k"]}], TraditionalForm]]],
 " (",
 StyleBox["Figure 14.17", "FigureFontText"],
 "). Summing the areas of the panels, the approximate area of the curtain is \
given by the Riemann sum"
}], "Text",
 CellChangeTimes->{{3.4791670560511703`*^9, 3.479167347921083*^9}, {
  3.479167516030064*^9, 3.479167516348899*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{"area", "\[TildeTilde]", 
         RowBox[{
          UnderoverscriptBox["\[Sum]", 
           RowBox[{"k", "=", "1"}], "n"], 
          RowBox[{
           RowBox[{"f", "(", 
            RowBox[{
             RowBox[{"x", "(", 
              SubscriptBox[
               OverscriptBox["s", "_"], "k"], ")"}], ",", 
             RowBox[{"y", "(", 
              SubscriptBox[
               OverscriptBox["s", "_"], "k"], ")"}]}], ")"}], 
           "\[CapitalDelta]", "\[VeryThinSpace]", 
           RowBox[{
            SubscriptBox["s", "k"], "."}]}]}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479167418927464*^9, 3.479167448678924*^9}}],

Cell[BoxData[
 TagBox[
  StyleBox[
   DynamicModuleBox[{$CellContext`aValue$$ = 0., $CellContext`bValue$$ = 
    2., $CellContext`kValue$$ = 3, $CellContext`nValue$$ = 
    6, $CellContext`showArea$$ = False, $CellContext`showAreaApproximation$$ =
     True, $CellContext`showCurve$$ = True, $CellContext`showGrids$$ = 
    False, $CellContext`showLabels$$ = True, $CellContext`showPartition$$ = 
    True, $CellContext`showSurface$$ = False, $CellContext`xMax$$ = 
    5, $CellContext`xMin$$ = -1, $CellContext`yMax$$ = 
    5, $CellContext`yMin$$ = -1, $CellContext`zMax$$ = 
    5, $CellContext`zMin$$ = -1, Typeset`show$$ = True, 
    Typeset`bookmarkList$$ = {}, Typeset`bookmarkMode$$ = "Menu", 
    Typeset`animator$$, Typeset`animvar$$ = 1, Typeset`name$$ = 
    "\"untitled\"", Typeset`specs$$ = {{{
       Hold[$CellContext`showSurface$$], False, "show surface"}, {
      True, False}}, {{
       Hold[$CellContext`showCurve$$], True, 
       "show curve \!\(\*\nStyleBox[\"C\",\nFontSlant->\"Italic\"]\)"}, {
      True, False}}, {{
       Hold[$CellContext`showAreaApproximation$$], True, "show panels"}, {
      True, False}}, {{
       Hold[$CellContext`showPartition$$], True, "show partition"}, {
      True, False}}, {{
       Hold[$CellContext`nValue$$], 5, 
       "\!\(\*\nStyleBox[\"n\",\nFontSlant->\"Italic\"]\)"}, 2, 20, 1}, {
      Hold[
       Row[{
         Manipulate`Place[1]}]], Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`kValue$$], 3, 
       "\!\(\*\nStyleBox[\"k\",\nFontSlant->\"Italic\"]\)"}, 1, 
      Dynamic[$CellContext`nValue$$], 1}, {
      Hold[
       Row[{
         Manipulate`Place[2]}]], Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`showArea$$], False, 
       "\!\(\*FormBox[\n RowBox[{\"\[CapitalDelta]\", \"\[Rule]\", \"0\"}],\n \
TraditionalForm]\)"}, {True, False}}, {{
       Hold[$CellContext`aValue$$], 0, 
       "\!\(\*\nStyleBox[\"a\",\nFontSlant->\"Italic\"]\)"}, 0, 1.99}, {
      Hold[
       Row[{
         Manipulate`Place[3]}]], Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`bValue$$], 2, 
       "\!\(\*\nStyleBox[\"b\",\nFontSlant->\"Italic\"]\)"}, 0.1, 2}, {
      Hold[
       Row[{
         Manipulate`Place[4]}]], Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`showLabels$$], True, "show labels"}, {
      True, False}}, {{
       Hold[$CellContext`showGrids$$], False, "show grids"}, {True, False}}, {
      Hold[
       Grid[{{
          Manipulate`Place[5]}, {
          Manipulate`Place[6]}}, Alignment -> Right]], 
      Manipulate`Dump`ThisIsNotAControl}, {{
       Hold[$CellContext`xMin$$], -1}, 0}, {{
       Hold[$CellContext`xMax$$], 5}, 0}, {{
       Hold[$CellContext`yMin$$], -1}, 0}, {{
       Hold[$CellContext`yMax$$], 5}, 0}, {{
       Hold[$CellContext`zMin$$], -1}, 0}, {{
       Hold[$CellContext`zMax$$], 5}, 0}, {{
       Hold[$CellContext`aValue$$], 0}, 0}, {{
       Hold[$CellContext`bValue$$], 2}, 0}}, Typeset`size$$ = {
    396., {195., 201.}}, Typeset`update$$ = 0, Typeset`initDone$$, 
    Typeset`skipInitDone$$ = False, $CellContext`showSurface$388984$$ = 
    False, $CellContext`showCurve$388985$$ = 
    False, $CellContext`showAreaApproximation$388986$$ = 
    False, $CellContext`showPartition$388987$$ = 
    False, $CellContext`nValue$388988$$ = 0, $CellContext`kValue$388989$$ = 
    0, $CellContext`showArea$388990$$ = False, $CellContext`aValue$388991$$ = 
    0, $CellContext`bValue$388992$$ = 0, $CellContext`showLabels$388993$$ = 
    False, $CellContext`showGrids$388994$$ = False}, 
    DynamicBox[Manipulate`ManipulateBoxes[
     2, StandardForm, 
      "Variables" :> {$CellContext`aValue$$ = 0, $CellContext`bValue$$ = 
        2, $CellContext`kValue$$ = 3, $CellContext`nValue$$ = 
        5, $CellContext`showArea$$ = 
        False, $CellContext`showAreaApproximation$$ = 
        True, $CellContext`showCurve$$ = True, $CellContext`showGrids$$ = 
        False, $CellContext`showLabels$$ = True, $CellContext`showPartition$$ = 
        True, $CellContext`showSurface$$ = False, $CellContext`xMax$$ = 
        5, $CellContext`xMin$$ = -1, $CellContext`yMax$$ = 
        5, $CellContext`yMin$$ = -1, $CellContext`zMax$$ = 
        5, $CellContext`zMin$$ = -1}, "ControllerVariables" :> {
        Hold[$CellContext`showSurface$$, $CellContext`showSurface$388984$$, 
         False], 
        Hold[$CellContext`showCurve$$, $CellContext`showCurve$388985$$, 
         False], 
        Hold[$CellContext`showAreaApproximation$$, \
$CellContext`showAreaApproximation$388986$$, False], 
        Hold[$CellContext`showPartition$$, \
$CellContext`showPartition$388987$$, False], 
        Hold[$CellContext`nValue$$, $CellContext`nValue$388988$$, 0], 
        Hold[$CellContext`kValue$$, $CellContext`kValue$388989$$, 0], 
        Hold[$CellContext`showArea$$, $CellContext`showArea$388990$$, False], 
        Hold[$CellContext`aValue$$, $CellContext`aValue$388991$$, 0], 
        Hold[$CellContext`bValue$$, $CellContext`bValue$388992$$, 0], 
        Hold[$CellContext`showLabels$$, $CellContext`showLabels$388993$$, 
         False], 
        Hold[$CellContext`showGrids$$, $CellContext`showGrids$388994$$, 
         False]}, 
      "OtherVariables" :> {
       Typeset`show$$, Typeset`bookmarkList$$, Typeset`bookmarkMode$$, 
        Typeset`animator$$, Typeset`animvar$$, Typeset`name$$, 
        Typeset`specs$$, Typeset`size$$, Typeset`update$$, Typeset`initDone$$,
         Typeset`skipInitDone$$}, 
      "Body" :> (
       If[$CellContext`showArea$$, {$CellContext`showPartition$$, \
$CellContext`showAreaApproximation$$} = {False, False}]; 
       If[$CellContext`kValue$$ >= $CellContext`nValue$$, \
$CellContext`kValue$$ = $CellContext`nValue$$]; 
       If[$CellContext`aValue$$ >= $CellContext`bValue$$ - 
          0.1, $CellContext`aValue$$ = $CellContext`bValue$$ - 0.1]; Show[{
          If[$CellContext`showGrids$$, 
           $CellContext`CoordinatePlane[
           
           0, {$CellContext`y, $CellContext`yMin$$, $CellContext`yMax$$}, \
{$CellContext`z, $CellContext`zMin$$, $CellContext`zMax$$}, \
$CellContext`PlaneColor -> GrayLevel[0.9], $CellContext`PlaneOpacity -> 
            Opacity[0.3], $CellContext`PlaneGrids -> True], 
           Graphics3D[{}]], 
          If[$CellContext`showGrids$$, 
           $CellContext`CoordinatePlane[{$CellContext`x, $CellContext`xMin$$, \
$CellContext`xMax$$}, 
            0, {$CellContext`z, $CellContext`zMin$$, $CellContext`zMax$$}, \
$CellContext`PlaneColor -> GrayLevel[0.9], $CellContext`PlaneOpacity -> 
            Opacity[0.3], $CellContext`PlaneGrids -> True], 
           Graphics3D[{}]], 
          If[$CellContext`showGrids$$, 
           $CellContext`CoordinatePlane[{$CellContext`x, $CellContext`xMin$$, \
$CellContext`xMax$$}, {$CellContext`y, $CellContext`yMin$$, \
$CellContext`yMax$$}, 0, $CellContext`PlaneColor -> 
            GrayLevel[0.9], $CellContext`PlaneOpacity -> 
            Opacity[0.3], $CellContext`PlaneGrids -> True], 
           Graphics3D[{}]], 
          Graphics3D[{
            Arrowheads[{0, 0.02}], 
            AbsoluteThickness[0.5], 
            Arrow[{{$CellContext`xMin$$, 0, 0}, {$CellContext`xMax$$, 0, 0}}], 
            Arrow[{{0, $CellContext`yMin$$, 0}, {0, $CellContext`yMax$$, 0}}], 
            Arrow[{{0, 0, $CellContext`zMin$$}, {0, 0, $CellContext`zMax$$}}], 
            Text[
            "\!\(\*\nStyleBox[\"x\",\nFontSlant->\"Italic\"]\)", \
{$CellContext`xMax$$, 0, 0}, {2, 0}], 
            Text[
            "\!\(\*\nStyleBox[\"y\",\nFontSlant->\"Italic\"]\)", {
             0, $CellContext`yMax$$, 0}, {-2, 0}], 
            Text[
            "\!\(\*\nStyleBox[\"z\",\nFontSlant->\"Italic\"]\)", {
             0, 0, $CellContext`zMax$$}, {0, -2}]}], 
          If[$CellContext`showSurface$$, $CellContext`plotC14F17, 
           Graphics3D[{}]], 
          If[$CellContext`showCurve$$, 
           ParametricPlot3D[
            $CellContext`funcC14F17Curve[$CellContext`s, 
             
             0], {$CellContext`s, $CellContext`aValue$$, \
$CellContext`bValue$$}, PlotStyle -> {{
               AbsoluteThickness[1.5], Black}}], 
           Graphics3D[{}]], 
          If[$CellContext`showPartition$$, 
           Graphics3D[{Black, 
             AbsolutePointSize[6], 
             Table[
              Point[
               $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
$CellContext`i (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0]], {$CellContext`i, 
               0, $CellContext`nValue$$}]}], 
           Graphics3D[{}]], 
          If[$CellContext`showAreaApproximation$$, {
            ParametricPlot3D[$CellContext`funcC14F17Curve[$CellContext`s, 
               $CellContext`funcC14F17[
                Part[
                 $CellContext`funcC14F17Curve[$CellContext`s, 0], 1], 
                Part[
                 $CellContext`funcC14F17Curve[$CellContext`s, 0], 2]]] + 
             0.01, {$CellContext`s, $CellContext`aValue$$, \
$CellContext`bValue$$}, PlotStyle -> {{Thick, $CellContext`bcR}}], 
            Graphics3D[{
              Opacity[0.9], $CellContext`bcAreaPositiveColor, 
              Table[
               Polygon[{
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
$CellContext`i (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
$CellContext`i (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 
                  $CellContext`funcC14F17[
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 1], 
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 2]]], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 
                  $CellContext`funcC14F17[
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 1], 
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 2]]]}], {$CellContext`i, 
                1, $CellContext`nValue$$}], 
              Darker[$CellContext`bcAreaPositiveColor], 
              Table[
               Polygon[{
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
$CellContext`i (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
$CellContext`i (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 
                  $CellContext`funcC14F17[
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 1], 
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 2]]], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 
                  $CellContext`funcC14F17[
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 1], 
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 
                    2]]]}], {$CellContext`i, {$CellContext`kValue$$}}], 
              Dashed, $CellContext`bcAreaPositiveBorderColor, 
              Table[
               Line[{
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 
                  $CellContext`funcC14F17[
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 1], 
                   Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`i - 
                    1/2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 
                    2]]]}], {$CellContext`i, {$CellContext`kValue$$}}], Black, 
              If[$CellContext`showLabels$$, {
                Text[
                 Framed[
                  Pane[
                  "\!\(\*FormBox[FormBox[GridBox[{\n    {\n     RowBox[{\n    \
  RowBox[{\"height\", \"\[InvisiblePostfixScriptBase]\"}], \"=\", \n      \
RowBox[{\"f\", \"(\", \n       RowBox[{\n        RowBox[{\"x\", \"(\", \
SubscriptBox[OverscriptBox[\"s\", \"_\"], \"k\"], \")\"}], \",\", \n        \
RowBox[{\"y\", \n         RowBox[{\"(\", SubscriptBox[OverscriptBox[\"s\", \
\"_\"], \"k\"], \")\"}]}]}], \")\"}]}]},\n    {\n     RowBox[{\n      \
RowBox[{\"area\", \"\[InvisiblePostfixScriptBase]\"}], \"=\", \n      \
RowBox[{\n       RowBox[{\"f\", \"(\", \n        RowBox[{\n         \
RowBox[{\"x\", \"(\", SubscriptBox[OverscriptBox[\"s\", \"_\"], \"k\"], \
\")\"}], \",\", \n         RowBox[{\"y\", \"(\", \
SubscriptBox[OverscriptBox[\"s\", \"_\"], \"k\"], \")\"}]}], \")\"}], \" \", \
\"\[CapitalDelta]\", \"\[VeryThinSpace]\", SubscriptBox[\"s\", \"k\"]}]}]}\n  \
 },\nGridBoxAlignment->{\"Columns\" -> {{\"=\"}}}],\n  TraditionalForm],\n \
TraditionalForm]\)", $CellContext`bcPBS], $CellContext`bcFO, Background -> 
                  White], 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 4.5], {0, -1}], 
                Dashing[{}], Gray, 
                Arrowheads[0.025], 
                Arrow[{
                  $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 4.5], 
                  $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 
                   $CellContext`funcC14F17[
                    Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 1], 
                    Part[
                    $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], 2]]]}], Black, 
                Text[
                "\!\(\*FormBox[\n RowBox[{\"(\", \n  RowBox[{\n   RowBox[{\"x\
\", \"(\", SubscriptBox[OverscriptBox[\"s\", \"_\"], \"k\"], \")\"}], \",\", \
\n   RowBox[{\"y\", \"(\", SubscriptBox[OverscriptBox[\"s\", \"_\"], \"k\"], \
\")\"}]}], \")\"}],\n TraditionalForm]\)", 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], {0, 2}], 
                Text[
                "\!\(\*FormBox[\n RowBox[{\"width\", \"=\", \n  RowBox[{\"\
\[CapitalDelta]\", \"\[VeryThinSpace]\", SubscriptBox[\"s\", \"k\"]}]}],\n \
TraditionalForm]\)", 
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0], {0, 4}], 
                AbsolutePointSize[6], $CellContext`bcR, 
                Point[
                 $CellContext`funcC14F17Curve[$CellContext`aValue$$ + \
($CellContext`kValue$$ - 1/
                    2) (($CellContext`bValue$$ - \
$CellContext`aValue$$)/$CellContext`nValue$$), 0]]}, Black]}]}, 
           Graphics3D[{}]], 
          If[$CellContext`showArea$$, {
            ParametricPlot3D[$CellContext`funcC14F17Curve[$CellContext`s, 
               $CellContext`funcC14F17[
                Part[
                 $CellContext`funcC14F17Curve[$CellContext`s, 0], 1], 
                Part[
                 $CellContext`funcC14F17Curve[$CellContext`s, 0], 2]]] + 
             0.01, {$CellContext`s, $CellContext`aValue$$, \
$CellContext`bValue$$}, PlotStyle -> {{Thick, $CellContext`bcR}}], 
            ParametricPlot3D[
             $CellContext`funcC14F17Curve[$CellContext`s, $CellContext`u \
$CellContext`funcC14F17[
                Part[
                 $CellContext`funcC14F17Curve[$CellContext`s, 0], 1], 
                Part[
                 $CellContext`funcC14F17Curve[$CellContext`s, 0], 
                 2]]], {$CellContext`u, 0, 
              1}, {$CellContext`s, $CellContext`aValue$$, \
$CellContext`bValue$$}, Mesh -> None, PlotStyle -> {
               Opacity[0.9], $CellContext`bcAreaPositiveColor}]}, 
           Graphics3D[{}]]}, 
         PlotRange -> {{$CellContext`xMin$$, $CellContext`xMax$$}, \
{$CellContext`yMin$$, $CellContext`yMax$$}, {$CellContext`zMin$$, \
$CellContext`zMax$$ + 0.5}}, ImageSize -> 5.5 {72, 72}, 
         BaseStyle -> {"Text", 13}, Lighting -> "Neutral", Boxed -> False, 
         ViewPoint -> {4, 3, 2}, Epilog -> {}]), 
      "Specifications" :> {{{$CellContext`showSurface$$, False, 
          "show surface"}, {True, False}, ControlType -> 
         Checkbox}, {{$CellContext`showCurve$$, True, 
          "show curve \!\(\*\nStyleBox[\"C\",\nFontSlant->\"Italic\"]\)"}, {
         True, False}, ControlType -> Checkbox}, 
        Delimiter, {{$CellContext`showAreaApproximation$$, True, 
          "show panels"}, {True, False}, ControlType -> Checkbox, Enabled :> 
         Dynamic[
           Not[$CellContext`showArea$$]]}, {{$CellContext`showPartition$$, 
          True, "show partition"}, {True, False}, ControlType -> Checkbox, 
         Enabled :> Dynamic[
           Not[$CellContext`showArea$$]]}, {{$CellContext`nValue$$, 5, 
          "\!\(\*\nStyleBox[\"n\",\nFontSlant->\"Italic\"]\)"}, 2, 20, 1, 
         ControlType -> Slider, ImageSize -> Small, Enabled :> Dynamic[
           Not[$CellContext`showArea$$]], ControlPlacement -> 1}, 
        Row[{
          Manipulate`Place[1]}], {{$CellContext`kValue$$, 3, 
          "\!\(\*\nStyleBox[\"k\",\nFontSlant->\"Italic\"]\)"}, 1, 
         Dynamic[$CellContext`nValue$$], 1, ControlType -> Slider, ImageSize -> 
         Small, Enabled :> Dynamic[
           Not[$CellContext`showArea$$]], ControlPlacement -> 2}, 
        Row[{
          Manipulate`Place[2]}], 
        Delimiter, {{$CellContext`showArea$$, False, 
          "\!\(\*FormBox[\n RowBox[{\"\[CapitalDelta]\", \"\[Rule]\", \
\"0\"}],\n TraditionalForm]\)"}, {True, False}, ControlType -> Checkbox}, 
        Delimiter, {{$CellContext`aValue$$, 0, 
          "\!\(\*\nStyleBox[\"a\",\nFontSlant->\"Italic\"]\)"}, 0, 1.99, 
         ControlType -> Slider, ImageSize -> Small, ControlPlacement -> 3}, 
        Row[{
          Manipulate`Place[3]}], {{$CellContext`bValue$$, 2, 
          "\!\(\*\nStyleBox[\"b\",\nFontSlant->\"Italic\"]\)"}, 0.1, 2, 
         ControlType -> Slider, ImageSize -> Small, ControlPlacement -> 4}, 
        Row[{
          Manipulate`Place[4]}], 
        Delimiter, {{$CellContext`showLabels$$, True, "show labels"}, {
         True, False}, ControlType -> Checkbox, ControlPlacement -> 
         5}, {{$CellContext`showGrids$$, False, "show grids"}, {True, False}, 
         ControlType -> Checkbox, ControlPlacement -> 6}, 
        Grid[{{
           Manipulate`Place[5]}, {
           Manipulate`Place[6]}}, Alignment -> 
         Right], {{$CellContext`xMin$$, -1}, 0, ControlType -> 
         None}, {{$CellContext`xMax$$, 5}, 0, ControlType -> 
         None}, {{$CellContext`yMin$$, -1}, 0, ControlType -> 
         None}, {{$CellContext`yMax$$, 5}, 0, ControlType -> 
         None}, {{$CellContext`zMin$$, -1}, 0, ControlType -> 
         None}, {{$CellContext`zMax$$, 5}, 0, ControlType -> 
         None}, {{$CellContext`aValue$$, 0}, 0, ControlType -> 
         None}, {{$CellContext`bValue$$, 2}, 0, ControlType -> None}}, 
      "Options" :> {
       ControlPlacement -> Right, LabelStyle -> 11, AppearanceElements -> 
        "ResetButton", FrameLabel -> 
        Style["FIGURE 14.17", "FigureFont", Smaller]}, "DefaultOptions" :> {}],
     ImageSizeCache->{562., {234., 239.}},
     SingleEvaluation->True],
    Deinitialization:>None,
    DynamicModuleValues:>{},
    Initialization:>({{$CellContext`CoordinatePlane[
           Pattern[$CellContext`x, 
            Blank[]], 
           Pattern[$CellContext`y, 
            Blank[]], 
           Pattern[$CellContext`z, 
            Blank[]]] := Graphics3D[{
            Lighter[$CellContext`bcSurfaceBottom], 
            Opacity[0.3], 
            Switch[
             If[
              ListQ[$CellContext`x], 
              If[
               ListQ[$CellContext`y], "z", "y"], "x"], "x", 
             Polygon[{{$CellContext`x, 
                Part[$CellContext`y, 2], 
                Part[$CellContext`z, 2]}, {$CellContext`x, 
                Part[$CellContext`y, 2], 
                Part[$CellContext`z, 3]}, {$CellContext`x, 
                Part[$CellContext`y, 3], 
                Part[$CellContext`z, 3]}, {$CellContext`x, 
                Part[$CellContext`y, 3], 
                Part[$CellContext`z, 2]}, {$CellContext`x, 
                Part[$CellContext`y, 2], 
                Part[$CellContext`z, 2]}}], "y", 
             Polygon[{{
                Part[$CellContext`x, 2], $CellContext`y, 
                Part[$CellContext`z, 2]}, {
                Part[$CellContext`x, 2], $CellContext`y, 
                Part[$CellContext`z, 3]}, {
                Part[$CellContext`x, 3], $CellContext`y, 
                Part[$CellContext`z, 3]}, {
                Part[$CellContext`x, 3], $CellContext`y, 
                Part[$CellContext`z, 2]}, {
                Part[$CellContext`x, 2], $CellContext`y, 
                Part[$CellContext`z, 2]}}], "z", 
             Polygon[{{
                Part[$CellContext`x, 2], 
                Part[$CellContext`y, 2], $CellContext`z}, {
                Part[$CellContext`x, 2], 
                Part[$CellContext`y, 3], $CellContext`z}, {
                Part[$CellContext`x, 3], 
                Part[$CellContext`y, 3], $CellContext`z}, {
                Part[$CellContext`x, 3], 
                Part[$CellContext`y, 2], $CellContext`z}, {
                Part[$CellContext`x, 2], 
                
                Part[$CellContext`y, 
                 2], $CellContext`z}}]]}], $CellContext`CoordinatePlane[
           Pattern[$CellContext`x, 
            Blank[]], 
           Pattern[$CellContext`y, 
            Blank[]], 
           Pattern[$CellContext`z, 
            Blank[]], 
           Pattern[$CellContext`opts, 
            BlankNullSequence[]]] := 
         Module[{$CellContext`color, $CellContext`opacity, \
$CellContext`grids}, {$CellContext`color, $CellContext`opacity, \
$CellContext`grids} = ReplaceAll[
              
              ReplaceAll[{$CellContext`PlaneColor, $CellContext`PlaneOpacity, \
$CellContext`PlaneGrids}, {$CellContext`opts}], 
              Options[$CellContext`CoordinatePlane]]; Graphics3D[{
              If[$CellContext`grids, 
               Switch[
                If[
                 ListQ[$CellContext`x], 
                 If[
                  ListQ[$CellContext`y], "z", "y"], "x"], "x", {Gray, 
                 Opacity[0.5], Thin, 
                 Table[
                  Line[{{$CellContext`x, 
                    Part[$CellContext`y, 2], $CellContext`i}, {$CellContext`x, 
                    
                    Part[$CellContext`y, 
                    3], $CellContext`i}}], {$CellContext`i, 
                   Part[$CellContext`z, 2], 
                   Part[$CellContext`z, 3], 
                   If[Length[$CellContext`z] == 4, 
                    Part[$CellContext`z, 4], 1]}], 
                 Table[
                  Line[{{$CellContext`x, $CellContext`i, 
                    Part[$CellContext`z, 2]}, {$CellContext`x, $CellContext`i, 
                    Part[$CellContext`z, 3]}}], {$CellContext`i, 
                   Part[$CellContext`y, 2], 
                   Part[$CellContext`y, 3], 
                   If[Length[$CellContext`y] == 4, 
                    Part[$CellContext`y, 4], 1]}]}, "y", {Gray, 
                 Opacity[0.5], Thin, 
                 Table[
                  Line[{{
                    Part[$CellContext`x, 2], $CellContext`y, $CellContext`i}, {
                    
                    Part[$CellContext`x, 
                    3], $CellContext`y, $CellContext`i}}], {$CellContext`i, 
                   Part[$CellContext`z, 2], 
                   Part[$CellContext`z, 3], 
                   If[Length[$CellContext`z] == 4, 
                    Part[$CellContext`z, 4], 1]}], 
                 Table[
                  Line[{{$CellContext`i, $CellContext`y, 
                    Part[$CellContext`z, 2]}, {$CellContext`i, $CellContext`y, 
                    Part[$CellContext`z, 3]}}], {$CellContext`i, 
                   Part[$CellContext`x, 2], 
                   Part[$CellContext`x, 3], 
                   If[Length[$CellContext`x] == 4, 
                    Part[$CellContext`x, 4], 1]}]}, "z", {Gray, 
                 Opacity[0.5], Thin, 
                 Table[
                  Line[{{
                    Part[$CellContext`x, 2], $CellContext`i, $CellContext`z}, {
                    
                    Part[$CellContext`x, 
                    3], $CellContext`i, $CellContext`z}}], {$CellContext`i, 
                   Part[$CellContext`y, 2], 
                   Part[$CellContext`y, 3], 
                   If[Length[$CellContext`y] == 4, 
                    Part[$CellContext`y, 4], 1]}], 
                 Table[
                  Line[{{$CellContext`i, 
                    Part[$CellContext`y, 2], $CellContext`z}, {$CellContext`i, 
                    
                    Part[$CellContext`y, 
                    3], $CellContext`z}}], {$CellContext`i, 
                   Part[$CellContext`x, 2], 
                   Part[$CellContext`x, 3], 
                   If[Length[$CellContext`x] == 4, 
                    Part[$CellContext`x, 4], 1]}]}], 
               Black], $CellContext`color, $CellContext`opacity, 
              Switch[
               If[
                ListQ[$CellContext`x], 
                If[
                 ListQ[$CellContext`y], "z", "y"], "x"], "x", 
               Polygon[{{$CellContext`x, 
                  Part[$CellContext`y, 2], 
                  Part[$CellContext`z, 2]}, {$CellContext`x, 
                  Part[$CellContext`y, 2], 
                  Part[$CellContext`z, 3]}, {$CellContext`x, 
                  Part[$CellContext`y, 3], 
                  Part[$CellContext`z, 3]}, {$CellContext`x, 
                  Part[$CellContext`y, 3], 
                  Part[$CellContext`z, 2]}, {$CellContext`x, 
                  Part[$CellContext`y, 2], 
                  Part[$CellContext`z, 2]}}], "y", 
               Polygon[{{
                  Part[$CellContext`x, 2], $CellContext`y, 
                  Part[$CellContext`z, 2]}, {
                  Part[$CellContext`x, 2], $CellContext`y, 
                  Part[$CellContext`z, 3]}, {
                  Part[$CellContext`x, 3], $CellContext`y, 
                  Part[$CellContext`z, 3]}, {
                  Part[$CellContext`x, 3], $CellContext`y, 
                  Part[$CellContext`z, 2]}, {
                  Part[$CellContext`x, 2], $CellContext`y, 
                  Part[$CellContext`z, 2]}}], "z", 
               Polygon[{{
                  Part[$CellContext`x, 2], 
                  Part[$CellContext`y, 2], $CellContext`z}, {
                  Part[$CellContext`x, 2], 
                  Part[$CellContext`y, 3], $CellContext`z}, {
                  Part[$CellContext`x, 3], 
                  Part[$CellContext`y, 3], $CellContext`z}, {
                  Part[$CellContext`x, 3], 
                  Part[$CellContext`y, 2], $CellContext`z}, {
                  Part[$CellContext`x, 2], 
                  Part[$CellContext`y, 2], $CellContext`z}}]]}]], 
         Options[$CellContext`CoordinatePlane] = {$CellContext`PlaneColor -> 
           RGBColor[
            0.6871111111111111, 0.7377777777777778, 
             0.7133333333333334], $CellContext`PlaneOpacity -> 
           Opacity[0.7], $CellContext`PlaneGrids -> 
           False}, $CellContext`bcSurfaceBottom = 
         RGBColor[
          0.5306666666666667, 0.6066666666666667, 
           0.5700000000000001], $CellContext`i = 1, $CellContext`plotC14F17 = 
         Graphics3D[
           GraphicsComplex[CompressedData["
1:eJxdmnlcjtn//+/2tKlbIjIq9bFVSCnrfUVm3KPBhCiV7EuFYSZJacGIYj4l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            "], {{{
               EdgeForm[], 
               Opacity[0.8], 
               RGBColor[0.85, 0.6, 0.2], 
               Specularity[
                GrayLevel[1], 100], 
               GraphicsGroup[{
                 Polygon[CompressedData["
1:eJxNmgn41lMWx39Xsm+hkWwl7SilRUKStimSiIS/UqEFoSJUhEJKiii0EEmS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                  "]]}]}, {}, {}, {}, {}}, {{
               GrayLevel[0], 
               Line[CompressedData["
1:eJwl0Lsug2EYwPHvc2od6nADFYlNmEzSXgFdGkkpCxMGlg7EZClDE1tFJBiQ
NHSrK2CUSNSxWq076a/p8Pv+z/O+7/RNbe6ld8MgCNa4MnTmgk+EcfMkcyRZ
Yjrsvt3SHT3ghDPueCTl7lkzek/VfKmHvJs/+OSLb36o8UudBn+8eN/UV23p
qj7wbz7XFfK82Yu6TIkne04retu545gFZ/ta1nmO2Laf6gw3ZO0buqgJZokz
4eyaMfOFrjNqjjHCMEMMEiXCAP300UsP2bD7n9uRYisq
                "]]}}}, VertexNormals -> CompressedData["
1:eJwd2Hk4VG0YBnAlIZItRVFaaZVCKZ6EhKikooVIJC2Soo1IsiS7KJUlW7Ra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             "]], {
           Axes -> True, BoxRatios -> {1, 1, 0.4}, 
            Method -> {"RotationControl" -> "Globe"}, 
            PlotRange -> {All, All, All}, 
            PlotRangePadding -> {Automatic, Automatic, Automatic}}], 
         Attributes[PlotRange] = {ReadProtected}, $CellContext`funcC14F17Curve[
           Pattern[$CellContext`s, 
            Blank[]], 
           Pattern[$CellContext`z, 
            Blank[]]] := {
          4.5 - $CellContext`s^2, 2 $CellContext`s + 
           0.5, $CellContext`z}, $CellContext`funcC14F17[
           Pattern[$CellContext`x, 
            Blank[]], 
           Pattern[$CellContext`y, 
            Blank[]]] := (0.2 ($CellContext`x - 1)^2) 
           Sin[((2 Pi) $CellContext`y)/6] + 2, $CellContext`bcR = 
         RGBColor[0.8, 0, 0.4], $CellContext`bcAreaPositiveColor = 
         RGBColor[0.796, 0.91, 0.855], $CellContext`bcAreaPositiveBorderColor = 
         RGBColor[0.467, 0.753, 0.651], $CellContext`bcPBS = 
         BaseStyle -> {
           LineIndent -> 0, LinebreakAdjustments -> {1, 1, 0, 0, 0}, 
            LineBreakWithin -> Automatic, Hyphenation -> True, TextAlignment -> 
            Left, FontSize -> 13}, $CellContext`bcFO = {
          RoundingRadius -> 5, FrameStyle -> GrayLevel[0.85]}}; {Null, Null, 
         Graphics3D[
          GraphicsComplex[CompressedData["
1:eJxdmnlcjtn//+/2tKlbIjIq9bFVSCnrfUVm3KPBhCiV7EuFYSZJacGIYj4l
yZYkyhI+g0Rxt9BUtiRZIi2iVXf73v3t93vcr/dpmn/Ow/N+X+ec63nOeZ9z
ncZg7Xb7DbI8Hu+NDI8n11euP1mlK5HUi1DecD6WalndLDB30Xo+RlhGfLlN
oZPuuA5Bb8gXnqEwg3jYt73i1Wu6BRvN5pwyFuYLwNcu9vcbFNkr2Pjrbt1t
RV+J53qXyOhkSQTPnoy9bSRsJP7PuvN+Bxskgoqy447hRh3EHTVTxSuGSwS9
kv//H/GgfdNWHZ3RIwjVmPY6yVOOAxfc0HhovLJToDvK47JnkSLx8PO2Oga/
tQouT4zdYyRUIe5j/26j/8EGwdQZ7+yKktSJi48edwmc8E3gZqSkKQzXJG7S
vC75eGC+tB4+8SVOee+js3NFteazyyWSIcS141ydlt6oJF/wmlz0xGlUbjt5
RqnycEnOztAu8gz+vsacP+9SD3kG/89mD8tj2hLyDC6e0mO+7APzDG4trFA5
0co8g/ckGj1csoV5BufcUoSHzHvJMzjP7cota/tu8gw+J/F7w/rsDvIM3v5T
iLraiVbyDD55sr+cWXIjeQZ/9jQi/sS6WvIMPi3/ffdjuTIBPINvfmE5MmFe
Ds0P+FuR9vZ09LBeeh/wH7fPbfHP7yXPKDP33F57yE1CnsGff55QfiedeQbf
cCHBS7mVeQYPfWQyKqiHeQbnrO4VGrxjnsGDVBdc/HZQQp7B7Wd+DnqtKCHP
4LeyfP+ocOolz+BnE/X39jVMnsGHinOP7/XtJs/gesf2iiTCLvIM/nffoF+p
7iDP4IUf6xf7uLaTZ3j6aHJ2WV2ZDI0HuPqno5MMJ8jQ8+D58++k3zrH4/Bv
lOM3DVY9NbuPSz2Da9ucVy0axOPgGTz89bpML1keB8/gcT+4mHuM5XHwDC40
DFmZ4s3j4Bk8sLjOcn0rj4NncLulB3M3xshw8Ax+LeBXnb6GOXgGP7fRXnfP
HjkOnsENFYJfByfIc/AMPsO50bZRTpGDZ/AK9zDPuyFKlEfgIzVv5+ANTvLM
s5TfFGt+aLspyzxL+eTCCwvL3GWYZykXBBqvNzrWz7O0/NRwZswjnX6epVx5
nMKRmwr9PEt5+pg1J9ud+nmW8kEVNxee0pFhnqW82L3hcpi1LPMs5XOTja+U
pskxz1JukWtnf/imAvMs5XcubIwNlldmnqU8Y86uU59FKsyzlDtf2Gl1rFSd
eZbyzPXmc3LvsPyO9/ae/vmUUYsCy0dSfntjWsGRRLY+wZcJR/nJNDLP4I6l
LeHBGcwzeKbZzzr6o5lnlGFTRihlKzHP4E+qNmwI8GWewZ1XmE4R+jHP1K7P
jq2GKnLkGfyB/SQttTEK5Bk8sGSLsWK2EnkGvz5GP16zQ4U8g8/Ss4+cmKRB
nsGHV/jvueavRZ7BH0dGuAqr2b6I99N1kNn8ag3bP8BXxaysuSBmnsErt/Hf
+A2VZetEytPTzYw35DHP4F+iEotWGDDP4Eubj7UvV2aeUaoX5O9dfZB5Budb
79nwexTzDL5WP+huhA3zTPk+tM49LYB5pnb8UoLbliuTZ/BhTTWqM9JVyTN4
RO2+lkMZg8kzuJnlg1ixCZ88g1tN+9oTm6lNnvEehzRfTa84wuYz+KdXv69p
cmeewY0mxgxe8pLNZ/BVekttu+4wz+AHLh2VNOsxz+DnshbbzO43n8Hj/U8N
KtjFPKOMzdu68v5a5hn82P1M07rXLG+Au8+IO7z9H3nyDD59qZoHN5vNZ/B2
ue6C1ZZsPoMn2O16/PymOnkGv6gzfdT8Qk3yDF5r73smdyibz+hvWLOnT8Np
dk4Cn2paeztrOsvP4DWvv+53NWWewe9XulQn/8E8g59x//XbGz7zDH7cOd3n
rjzzDB6Tkhi3YiHzDJ4ZrLMh9SvbB1F2ZXallBfKkGfwBZ4uV56PZPkZ/OqB
maZ9DZNncEOVyJ+qLiqSZ/BbPR7tI74pk2dw+9VpgsH+quQZfNCk/47Kctcg
z+jXX/kzyiY388gzeODZHMO0SB55Bj8VZOmaupidN8ALD+yLMB7NPINPvxj3
pEueeQZP/xTfMEeGeQbfYRE8vH4I8ww+L37+DLXZzDP47GklDpG+7LyB0rl0
xbaj+TzyDH4h4XpAX8PkGVwttPhoRroMeQY/G1J3grdKljyD/xr/Pupa374A
z+Cmn2MjM/PYdxTaj/Au+dt4Rjedn8GLHUXvU6710PkVvFXm6unox7103gIv
2RHT8vwAOx+An7ocWy2sZfsZ+JiLt/bwu1j+BffbmBs79n8sX4DHi+s3HNZg
8xs81sYg02pIL40HuIej6wNzUTc9j1LROn6hD7+Lzs/gv33q8tJQ76DzM/h1
ofOs+hutdH4GT/bNPqPX3kTnZ/LpxZ2IrBTT+RntLBCnjZ22sI08g995sXe3
ZkwneQaXM37hP3Mc+x4Et2y6x90rZ9/d4AunT3+wP595Bp/X9mttTDPzDD7a
VPatxs/MM3jxB6HP83z23Q3uzxtf+PFYF3kGl48+Xz1rXzt5Bve4nXivLqaZ
6iUeeNmnuK6ePINnhei3BfEqyDP4R9XjOWFNeeQZPIbzaZhzKou+u1Gfxend
9UlD2f0GeHjvTI/quHbyDO6aa6vVM5Ldb4Cf6gkr/u7DPIPbRI0pEKUzz+Ar
T3c0bBczz+Af5bStJYrMM/jLgp2XPRWYZ/BJ6jqzU2pY3gHvud7T9e1+C3kG
t74zvrxpu5g8gz+eb1HxcOQXag9l9j+77YXnHpNn8Hbjz4V3O9/S/Qa4trD5
g9Ce3R/huZADOeNcZjUNyBt8Tt/no1HqD+0D8gafe8IPXCr/pmtA3uBzvovu
3bFc3Tsgb/C5eab+C5Y8HJg3+NzIm+80ltUPzBt8Tr4gU8mmpXdA3uBzkjPc
1GEvuwfkDT6norQ4rNC3Y0De4HPGQ+sN/HubB+QNPhcbapNrkPB9QN7gc3qW
2XeHdZUMyBt8jrfb9gUvNXpA3uBzB+LT5h3cVCr6d97gc0t0NmuHajSRZ/x+
fZ2pc//7OvCH3g6j+9/Xgd/wkHvV/74OPFBg8Vv/+zpwi4ZKydB+93XgOYEG
/7qvA5/T8Kmy/30deITgh3n97+vAcz3KQvrf14GXek/I1O93Xwf+bl1r5b5+
93Xga1eFdvr3u68DT3DcxPtvv/s68LzBf/ye0O++DmW310YXhxuVooU+frvC
jKpE5gE//+JRVCjqejZbL25+3z4o/TfKwbMs/5P5rkeQP/TVrHCjvvO1lPNj
uNYZ73Dfokx8QUrxjryuTqkXVeLy18Pu//6/dgHafZF3csp44d+C3gNu686s
4VEc+N/1LUlu1jzuR6UzW4qSZDjwOjkbE0NTHof+gP/Sorv8nD2PQ3/A/X4w
UFmxn8ehP+B7/R0WLhbxOB+bXUlJnoOJ/zz8wfA/FWU49NPOQnUXj1ci2Lqv
7crgo7LULri5+Te7tHGy1C74teqUTRnb5Khd8L6TF5/HU6B2wW3NV9vt+ajI
ySUUPE7y1CK+Md1Ux0x+EPUH7Y95d3Go2ScZ8gaeNP5oYPQQGeonSjVbo4tl
x2Won+BG8262DQuSpX6C1+5eVDU/Xo7axXMPRnw6mDNLQvMEPMRm+aM5f0kE
GC/wg0drclMuSWj+gCsGXIp84iSh+YMy5q9Ce/uOXpo/4EtT90nWBPYK4A1c
XzH6bHNjD80rPPdzz+nvJRWd1C548sT90f43uqhd8KfDtI7sNeygdlGeHtHu
wuO1UrvghhbWIzR+ahRgvMBzs2P9hPU11B+UET2uk4cdZPeK4NZRoQric63U
T/C5V+c6WV/rYOsO9956Ml7+hqz/iD8UJaNpV9/C1qM0vkZzp1HVvHZ6L8Rr
PXLPyXJoYOtUGr/Bfkbi85Rmel/EJ1vKqv+a+Y04xn/uTse8sR0tAuHN7Wbh
RkMoPm3VuPn1Ffcp/9A8SmzIkDEopHt/1FNQ5fVPgx4bR+TvMk+zkO4A9r2G
dVqTMcw1fAv7jgCXVZ7mWHOeR+ck8MoDQSVBpTzqh/+2clFRUp1A51qwyUQr
ecozWHfROinpHabsOx38mabziZYLrD/g+T7OlY4ZstQf8NRAhacdpuz7ETzY
eMXsN1/YeK8scPmzKKlFkJRQGxWsyvID+vnfiVVrgh1UiSO+Vf3Olbth6pQ3
EB/5zE087asGccSbpaRobazW5DBeiD94esnDkXu0iCPezdJDtNyB3QOA1x8J
Pqy7TJHWIfioUoVV5TmM72zIMRGGdwuOeXY/3P6XIq0TxB+Zckxrc/kg4ojX
9YrStj02iN4X3Kv7w4wSPfa+4Et4a4fHnmP7IupP7g7jpiZo03sh3i59fNe9
Zj75B+fPeWw+vESe8iB4Xpmi1Tp9xinPbe2uXDdBjt4X3Nxi3EgnZZZ3wC26
Al6cXaZE7wW+SaQoZzJYhd4LPCfgfrKSNTsHoD9u733XWc9k74X4G+eftZ3c
pEn9xPwtVvm0Om0b26fArWp8o1SL2T4FHjf91CbdWDYPUd/GMhlvDzkF6g/a
fdqb6qK5VovyCOKtOlusDv3G9lOsj9U7NKYq9XYPOOcrc5ONbkYWmUgGnPOV
uRUZI5+OT+sdcM5X5iL9JPWds3opb6C/b29Oy78S1kb7Ed7P1WHRGbW4gec0
Vc4i+pNEO7V7wDlNlavpnbPbbm/ngHOaKufs1VGoFtc24Jymyh2SVx/tMK2Z
+oPxzLM6/f57ZQvlT5TWEpM1ny+VEUd8hGW46dKdr2l/we/BEx+mnzibThyl
+ExiU5rdcxHmA+I3r2nITQ57Sxzxcdd2RiT+XU3nbfDHM+26nq1oon0E/EO1
s+N0j0biqO9xzdEN+qWNtI8gPun91a+rd1YSR3xc3Ntl9VlV9L7gqdv2W/YM
f0vvBV6SdjJ0/p1M2kfwe/C1/OxnMY30XijH2X+OS7v0jfZX8Nr6V/tfClpp
/wX3yBpy/Z/H7B4DPGDZCUlprZg9j+/Cq08aphQXU/2o70VklknK+Da2P0rX
gVOpo6bpETHtj4jfcSYw1KRfPfgukQ8u91/nzc5d+C7Z4hhz5Y1InsYVfMGV
QVp6WWrkBdz2ke+Z8+tZPkSeV3LSEbgf59N7Iv5kkPmG6df5tK7o7xEWwUdP
v2PrAbx50yGXt+0sP4MvGl11fPRNlp/BvRMdou73sH0HfJt6tW7VNuYF+e2H
z+kjz45SpLxD9+iK5UfWP2T5FrxgqdIEjUaWb8FPW99erHhFmfoJbn/3a5dd
GZuv4J05F8zKzrF5Cb5gZmNTWR1bh8gfY1unPK96oETjhXwQtCPEevrgDsG/
z8+a3O+CV9nLAloHnGM1uXa5W5f0X76n+Q+uv/TltZKx7BxFeeOw848rz3yh
+Yn4L0rzTHPXVNC5HN+FK+9bfspdxNYpfb97u+2K1auk/oDbHmowlDuUT/0B
v3XVLHHqvZfULvi7/FpZ4e4a0b/P4X378M6Gzw/61sW74Ns/LgivFaF80iHr
tyWrVTDit0JlYXjfd6+UWxcHB/nwP4rA+c8z9Hi8ZsGdv56+/HB3CAceXdRe
n+TZJrj6XlS9fZQ2cZNqs0zPok6B2vOUCQFVbN2Cvwi3W3L9OIs3/+pZLZFo
csuqHNPf3Kijdgu2zDzvWaTFRVuvPCf6Q0zcu055WVESnxv3NPaTjF09vQ9+
73SbZJbt10L5BPGXBcXB/1vUSHEoV3r8bDu1opLuibBOxU1yex22aVI//nwU
KhSGiwULlKuclf3YPQvGjbek6/vxEykixB9bG2MZbtQ3f0JXZWvMFVGeQn8z
HJLa9t1uoXsDrDux99PQyWla1C68FZiObpLdzv7uCf+uPs2yQbnMJ85dRbHe
zg3d7P2T29dHGgl7BUp5V28Pr9WiePAv85RtL+sxL+iP621tzfsn2DzDvMt+
rWmlIveR3hfteEd0K08WF9L8o99vbepRmtBE/jGvZ7nXeR35Vi/C8/h9eKKl
7dH0YqoHnLvupmBQxOYJ7V/bv67ISWTzAL7bI55ItvaWU/3wqSOJGip/kc/6
LS3jfpL7j8HwChontK9id0ZxFL+V1hHq/zuk3vn/cbo/wjq7FLh10Hp2L4lx
OTTax0lrGJ/Nb+k4NhyOWWEUxcYd+dnm2/f67nA27liP+joB441d2LgjPu3k
ouI6HW0aL3huvmXhfSS5hMYLZXZW9qL5Qe+J47ngP12Xv5z0nt4f61Svx3pu
xO8N5Bnxzc/2R9xbJSYv+F3FK3DQWdtmyjOoj/dF7YNn37jg35iHz/4y6FyX
wu7Z4Y3b5pneEq/KFT68JCMM1yBv81foZNs/UKN4zNujVY9NT81j8Rj3GLvF
uzP2q1G+Rv1XNdfmxscMpvekcZkqLLz5G7u/pu+RGx9m6o5g8ahf6Lgo0/cX
TZpPqGdy+k7X9a/YPMN4zTmZfCviKft7A8Yr2t5i0ni3UgH6j/FZP7unaIvp
Bzr3Yv4vqvix5HRDGcWjnc1a5Uk1pUV0b4n6OaPjhVqF8QL0H/XzZFU7m91T
ab9E/by3nGFBairFo/6SqOjsqatv0/pCPY+HWBQZhnwlTv8/41v/pEvOX+m8
C2/tF0t9xUWsf3gupk1nsta+pyLkU8SXyf6iVu6rQeOL+TNFdC435Dobd/qu
FOX/8WiiGo0X4mXFwfJb/dg4Iv6rVtXM4UM0KY9Tf3aLxc+nZNN+jf6cqByj
cGevOu0HiNdfYDF7tWHBgPurIZyrmsjmqVcJjRfWx41ug1l5v7Bxp/uA5Pnb
VWOLB9x39eXDpxdd1x1KoHFBPXnuB/n37W8Rp3E8q3p10fDj5BPxm610D2uW
vaZ+gifMXPnUfk2RAOMKHmD101W1Uzk0juBuzS6tWRW3RP8Hr/ed0w==
           "], {{{
              EdgeForm[], 
              Opacity[0.8], 
              RGBColor[0.85, 0.6, 0.2], 
              Specularity[
               GrayLevel[1], 100], 
              GraphicsGroup[{
                Polygon[CompressedData["
1:eJxNmgn41lMWx39Xsm+hkWwl7SilRUKStimSiIS/UqEFoSJUhEJKiii0EEmS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                 "]]}]}, {}, {}, {}, {}}, {{
              GrayLevel[0], 
              Line[CompressedData["
1:eJwl0Lsug2EYwPHvc2od6nADFYlNmEzSXgFdGkkpCxMGlg7EZClDE1tFJBiQ
NHSrK2CUSNSxWq076a/p8Pv+z/O+7/RNbe6ld8MgCNa4MnTmgk+EcfMkcyRZ
Yjrsvt3SHT3ghDPueCTl7lkzek/VfKmHvJs/+OSLb36o8UudBn+8eN/UV23p
qj7wbz7XFfK82Yu6TIkne04retu545gFZ/ta1nmO2Laf6gw3ZO0buqgJZokz
4eyaMfOFrjNqjjHCMEMMEiXCAP300UsP2bD7n9uRYisq
               "]]}}}, VertexNormals -> CompressedData["
1:eJwd2Hk4VG0YBnAlIZItRVFaaZVCKZ6EhKikooVIJC2Soo1IsiS7KJUlW7Ra
ovDYs2fft5kxzIxlSkhJfe/5/nJd5zrXed9n7rnf8xvy1pdNzs7m4eFpnsXD
w0v+Ri0u8Fq1k4ZOxksqx5ub8N6XD9bfeljwtVxc52tLN7pGlUx4m1fjiW2r
pY6dG4LUkd5nRRateKf3h8DQ6gJUnH127vi9EWCKz+0SzqjFwBeH5rV6vcK6
M6s3ePJxYSSn+vkT1Rw0/DzPZczBA3+MTEb2tnLhc8dl2yvGKdCd2sp/8bYV
pqpnzvyc4ELk6ng41loOC/6lmy5fF4u+10YHAyy5oBQpF9PZ1gScgTO7j6jk
4MGx3LJFi0bB9UJnyyJaJ/wTZFXaS5bjTBz6iqwfhvpWyzn0oD5YL7sxrCKq
AYWdokqEC9jQ1sOP97XoIJtaf9tfvR0tzaeqr4gMgtOP2ZGNSxigxa7xLz3R
i6ILNBKSTPrB5cKUofYBBqw9ez3T24yOlhIRZ51yadAiMGqquIABecqTUBja
jw6xTRmBX3vglqzf7aUP6GAX+euppd0A7v6mJrw4vwMSbMaDy3a1oEabyelR
CRqGRqvayKSxoMLLyyusrhEn7EZHMtmt2JVSFMv6PQRFOfdb8zd9Re2Ar/tP
W1ShVviy35m6o/BkY+YxPe0SrPtmuIV/ZxYaioRbnLzBBeGJhX38Tu+wylTM
f2vPQzz6arun9igXNsz+uGzSNQZmy9zoPs25gIuHtu7NnOaC9l1UETMpALpD
t450YhpmDCw9ofyOC1p7Beyv5VfBHk3faV/jUhR8+l1gZfkoBK34JBER0QCX
/vS+z7BpwoL1Gw/c6B2G67sd5VxeNkMc7O4R6+nBxJ3uO6c62WBy4+BCzdYW
SLHk0xUT7sdXi8WEHm4YgLklR06wolpgc3G0lt2jQYxdKOu35CYdpqLQlfaw
GeQZ086Dszh4wmmu/P5LPVBbbj5Uu74Jdn3pfhCVMoTtv6KbskXbwWGbraR/
bgNEflYq0v8yjEnxmS25FU1gJ3f1xsFzH9FuEuKWmvZj78DCIDk/FpgnV7+z
0MtE1oSMrNK8bnxi/PXPQZth+DBbJ+Jv9TsMrom725NWh5OSm/YUNoyCcbXE
shcpCZg251QV9H3C2hgF3+NVXLh+KHce63Iw/ozJ5QsyDsb4t+WmL2e40Oes
s+KfowecSVL/nHjgCuYsyVCL/ccFZ3Hdb042sTA9R+v1Fu33eH3ztohnLC60
Rnzl+9jyGkKHXi4bZ1eg7Ez+w1X7uDB77qIZO74M4NPcUzzY0ob5RSmiUdYj
IHpk3umIqSw44cjYZ+BNx+96F6vUuGzQr91W8zj0I0y0Rardsx3E40zhuoAY
Jigd+XrzwvMs8Oy5PK80mYNC5w+q/XKlwa6YpyHP/2bA6FPlW4OJw6g9pv1D
k7cLzuTam2/XTAc/UxHNT60jGJCtd3trZAtMm8UIBUa+B5ue3abSQaN4X2G3
gb1TA8iLL+i02vAFhOJqzvKkMtBHtYnpFcyCb6kSt+R/lUBfh2v68FQnnulv
W9ewfxi2yk4lj/sXwINkvZEK4TrsPyz5Vil9FHSe04acLDNh9svqFTPsHHxR
z79nQw4Xmr9LZ2vsigWBw8eLv3wMQq9f/GXav7mQKIqCrzWDUfrQTd41nxxR
3axbpvQvFy5vPpvAu+E96hSIbBdKfof697y5a7q4EHfr5LNrtnnoxNN/fuxH
OW5fLHE6cD0Xdjet2Hg+uxgXbZ4TJ056p566/ouO7gjcfcB7VS29DJlCm2R7
GDTM0LM79YfDhsW7mw0POH7Bp0XHt2zMHUBxm/GuY/lM2BraK2MYW4ZeB9dO
zhLjYIPgwsKc1zTQmC+4sVi2FJNuu24uVxxG9sek1+DWBVFPmwMuXyrGtJ/R
XLWrI6hx4OvBUvFW8HzWl1VUX4gNzloXL8Aoij0f1bLub4B74o9Zi4c7QHSP
0JY71j04T1l72bw8FnQ8KUjsLGyFVOn9jzeKNWPj3Lthr8qHYLXuZsmWlAbQ
2OiZ9lK8HCVcjjb+mT0KG26M5sWYlcPme2oPLsWk44Nfkw2MQ1yo5LfkT25J
h9gA9EuM8MUEg106OQNcMA/xenaPHo8hWfa3xhPsMdBAwO3LLy4ICoZW5MUX
Y3htW+a0aQryN03MbQvjwopHoiZVz+uwe4H/T56iIryQfnLLYf9R2Nfh9+GC
VgtKOEwm/P5Wjz/EfdAqdhikm7sbNRd3YIIRKnWbdWIZ/478gAY2PHY3mR4Y
7cQlQyUBRbx03JSi7+F2bQBcvdal98l1YQjfJdo2Xyb+Zpz2VpzHAHlHB+F2
rU7k8XSRzFrPwrAq96l7fb2wtSdf9mRPO2aaBbF1SC4mr0KV7V50gHJ6w0yA
YRty196PuOowhKo+cxuFD7fAkNfztbWBdHhUqSDWfa8WGavsjrcOsyA/YUT0
zEAveJaceg9PSjBs5y0jqz1DMH8g/dT58TYQXjSjaHXlIy7fID61zWQErn2Y
v1HRsQ7Ks3XVvYzi0avBmLFmkJwnZ+dNN0h+hrJn61/WON7APIN5Q731XHCp
32qnoZyKs0IumTs9NEcDW5m6C+NcMBbRNTsbW4l203tuWc6JQhfcZq1pwAXW
eokzrntbcPH5P04WBz7gy3DFm+fZI7D16dvDSmXdeHpLqrLWrULMMA+uevNn
CD6y9gnpV9DwmGV0bcOaKlzF6t3xOJsNpgzbR8atDFQfZKqFxzfg66vCh+Yd
HIT+6YnAvauZqGF4tqEuohXfDq1sVxzuB8dyxZJTLUzcU5p0dYFKF3Y28ixd
kkIn7+lDy+4sGMAJcyW7aqU+FMxeMVQbQd67/m/25n5nopSZ3suPgnT8+Gbv
TKxLN/iq3bmXNMqA+3p/Vfn5U6C0qct02x8WmBhPLHP90weTt50y5xXFgHZj
pZmD9BCUFLdadft1wMmH4WomQqGgPbLCO2fpCLTzqo1o+daDIP8y42fO7qBi
Gf5gZ9kocLJ0wlr+5ELyZEC8c7UpdM+xsTYm5za99qZb1fLXuE4/eof2wDa4
ucP4xOcxLmxZv6dyrlM1Rgtcf991zhk2dy4JEd7Khf3rty99mEnex79tFkqV
BoBSccJBlewRcLiRvdslrRelT2uaGZyLhj+em8/VVA7BSY/UgHf8DNx09czQ
IqVEWJpwY+jSO+KNfcy4HfeZGFwl8A5pb6CVR/ll9JNBsLipoXj38iD607jN
JT/Twcn2AN08gQkqnvJHHg6wcH4LQ3AgKBuiRBmttC4GeMiZD+/j4+C3gcv7
uevzYG+RU/AvbTqkPhdRxT4Oqn54u7SxtwB+ei1P9m7rAx/vwnriOph/iZ37
o7kJ5FQ+naVcl5P3Jpe4Dminzty6b14Njo+/zKdcd2Z52W/iOlj/3EyJs7oA
Yl3sp38Q132x7T8+P6MWfngUv2r2egUiba6rKdeJbrMQeqqaA619sqHfHTzA
uCfxMeW66/GbmojrMM4teuWF21awo8Hif9eZBA0nEdeh5EkDp2XrYqF9jRaD
cp2UYckP4jrcengL97BKDqwdS8mlXFfqPKy7mNaJybxyPuckyyHP8J475brF
/UcNievwUJR9THlUA5iZ/CyiXFfhECjtrUXHgf3q2n7q7eBcJd5Jue7uX33B
piUM3Ghov6rkRC/Uxp0qpVwXd1euibgOn4fU/LxvRofGJN9wynVpihu+E9ch
/X2zZ0FoPwQPuqwPIq6Ta7uvIPuAjnqnTw5Y2A1AptDWIMp1Aqrn/L7saoFV
Fh3ziOvg4H4bO8p1jqKhZuF1jVBd+/c8cR3Un9kSSbnuVO6XFQWbvgJvb1WG
pUUVLJpzZoRyXfSKV8r7tEuguoaWM3dnFnTxLTpGuS7I+1eYgNM72FcVPqHc
8xCCnavvUq4zLXZqJq7Dq5sy9hHXgVJbiy7lupuJJoPEdagNS18vTkyDC+Wj
RyjX6Rr+CbqeX4Wib+W9ievg4bD5zAriunfJ1u3EdSgdXLebuA5kaYe1Kdfl
xv8Ucn3ZjKHfX2kQ10GWaPYOynWGb52XQWsLKpzZzxEV7gfamoDFlOsW2b11
ZkeRc0m1lG37aJD0TfQl5Trnhy1h9IfNmHB08VriOnDJWWL9v+umLWbVrW/C
P/6xusR1YPBIVzqHuC6kxWrZw9wGvBGW7kdcB22qZtJ5xHWnjD8LHTn3EVpr
//kT1wHXXiL4f9clXGi21suEpkXsqc3zumHFnIRflOv4NTU6Z9W8g2VXzzkQ
1wFHY1iDcl09hh2LT0mAt9ZVr4jroCgoy5ty3fnCkCtDl4NBLqb9e6BxMIRm
mv7vutFbd31nX/FArtL9JOI60IyM+N91Z9O2mV63iUXH7MJI4jpgOj8Io1wX
c7SgKbflNX7nE/r3g10Biwv6fSjXKd8PojvwZaBmTE48cR0U7skUplz3z1u8
88lUFg4+vLaVuA7G+jsrKddVlRs7RId+RPkHh+WJ66BLnLeJcl0gD444Ps/C
kpTyyZJkDkgf09akXHeX4fwx7m8Gupxb5kBcB5d2HeMF4jqa+16FXZrpeG3f
9i3EdaBmuO0h5bq/Epk+IZHv0SLxrz5xHbh5/zWjXLd7XlS85YYvyAycDCau
AwmeUQblupKJAPdlv0pw8+2Tv4nrQNCzRYlyXZuajfQP/wJkehVJVQrXQf1e
uY+U69qdDJiOlpm48sIXa+I6kD6UtpdyneOfNoldu2KxUTGOQ1wHF8yZ/7uO
H9k2qZrB4G97ZR1xHez/vHAJ5br4Z994Z294Dxeyzp0mrgOUzxqjXCf5I7DF
yTYPQnVfvyKug8ymE3aU6862VBw8l10MsouTeonrgE/xZS3lun3v2vRV0stA
LeCmHnEd5D65bUG5jnPJht/I8Qu48ctaENdB8Ps9zZTr4h919u2LLYPch/Zy
s8U4MMvvcDrlOu7X52KFsqVgNTV8kLgO0pYeDqBc11SuGnnxUjFYPk2Yt/3q
CBzxsVWiXLdA4fhgQX0h8F4UvkdcBxG6r1ZSrit9ZjWPuA6leJbOcrPugQD3
piWU6zxX+492FLbinA2LlDeJNcPUP5/nlOsOl/zzak5pQEm5E5YJ4uUwfzi2
j3KdRYbn3xdm5eilZLjwckw67I/t7aBcZ9Z42C6pJR0LzuVpJkX4Au/pQl3K
dWarzv7xpMeDnwln00SCPaiPOPzvurzQZxuJ62CkeqfDH9MUOCznLEy5jqNY
VV/5vA7kHwh5zioqgsmvbhqU664qBfIT18HPs6Lrp7/Vg4gRq4pyXZLz3NnE
dTCVWPeduA5KA+QKKNeFSf/mI66DNwkR/4jrwGu7nhPlOm+deXG9cl3Qd/LW
ERVfJvB5TthRrqvMSue0anXC9Ongu8R1ML/pRDLlOp1LSpPHe9rhmoSNvi7J
5f01pw5b4rptT53bHhq2wb9WdzZxHXTQnt2kXCebNT1aE0jHV9kpx3rv1ULp
8tlmlOsCj+74ZD3Qi7QH97S0npTAzKKxw5Tr5r5cu4S4DlsEPMasr3wEtzkP
eFSI6y6E7/yn4FiHBq3Nb+8bxUPG6TssynVDPIxU4jpc/k884KvjDQhNFBum
XGe/cMEi4jp4MlGl7fzQHEp2WPzvOpcnacLEdTB2Qd7Cak4U3H3ZfpZy3f49
gmuJ60B4wdxWywMfQMOI60G5Ttdcag5xHcTo/vbec6sQPg5ebaRc1xT+tHVf
BQ12rOxpa1pTBdGb1+6kXPdNZK4BcR18fnkkLCK+AV6qR+lSrvO8kKBCXAcM
FtulIaIVfgc8zaNcp7jS6CBxHTw2vGcsptIFd1es+iZDXHfgSEnQ7QUDsLWV
/2yNUh8YyUy+o1y3UCDw+efvTPiX75SSLUgH70+RTpTres8qBBLXoYoozWZm
bgoayNscoVzXl5aqQFyHJncals8pisHWANtjlOtq/NddJq5Dk4mhXAOhUFSK
MfGiXLcwyHWSuA4PDrmMBzm74wHj696U6zT8HF4S1+HkjaQbF6pN0fDquv9d
t6yn6xFxHVwse8e3c2AbSlnC/67rtYloJ64D+2N6b5rOOeMKlUPBlOsaWrQV
iOsghRlmM680AFPlfhtTrnuk8OogcR2oHdVVgnPRaGPpZku5zsh9ayRxHcgv
qFNZoJSIJsrK/7tustT+PXEdXP0UGpZOe4M5b6oTKdddcPqgQVwH6zwb+PJ+
puOqVzMsynXVFllXiOtAZtXY1a6gbLSzvcmgXOd0UFdCn48DJ+aGjzPW5+FU
yfU4ynVVq67ZE9fB4qXTodW9Baiq1ZJPue7ltyK51TtpmOXz8e1YcxO+iCs8
R7lOLM1Apa6lGzVlZ8/yMq/Gtt/2fJTrfhulHSu2aMUVx/OOsVcX4IORi98o
1/WnF7UT16G+pYphk9crfLi/Wo5yXXQy14W4Dl/O3vL2m4MHZtQejaBcF2E7
ouVknAJBRlbgcNsKK+wF/ned/4SKqFlrOSRn9gXKrYtFk7bQHsp1Nr/s9bra
mmDJ57/PTFRy0Gos4gPluuHrMxnEdaD+67CunWQ59o7vvUa57sOwUQj1/7r7
YolpX6Ia8L3ykULKdUXjUs7EdbB8ME7QV70dFTQ2sinX9fIF3ySug5Mxc/KK
TvTifFU3JuW69SE6J3UOMKAza5U6cR2u23WnlHJdpP75gHULGGCwOrMUQ/tR
cPOSOMp13FCPJuI6CJNBJeI6NFq3SVKauK6j1WjBvA9V6D554vpHtz6MefnI
0NSHA5mHb+SF3P6CKy/NerZUuxkPqEhrSD0dgXcHPttPRlXCsNTzgZ413bhQ
fnzbofQhSFoRKPlvrBqumDkuX8zTj3uy5bW9WYPwlBYge2egGtQFf737u5iF
yQVvn34uY0DMR8lld3MSwWhba+SFbBoGGtb6hTzlQJHm0zWalvGQu2Nrp9Rk
C6qF/9h3rG0EBs+1/l5dEItDTANv+cYGPPnpvPx18j5P9Io8toKegG9Dh6dn
2/bi6zyDGxXtQ1DmZvc3LjwJBTY0tJ8+wkTZwvePNWsH4VDx0Mp4zSSEDQ7a
x06zcYfnIbdCRwZk9SSOijUl4PAhJRlTzhDaCvPP3rK+F47qRbZrDDbA9nQb
h6noHrxbO3uQe50DDWtc15rKNmCtumX1ppkOtE22sqwJGgLHS1LXrBiNWCLL
d2LvGzomv54MKeFhwQL2jmhJ6SbUSrCWjigfwPY4cVVJ6X4oLcP7v0wa8dSO
3myrFDY2SJXEz3XvA4EDe3dbLiXvQePxWSPBQ1grVpy481cHOC3xWib8oQrs
FQu0SV7E4YH7qLzmxzxyDbv9BcY+WB4jeUFdcuxWKq/cEnPrqahK/Mev5Efy
ggd8fzZReek+mNSZ9aMao6qqWIt4+uHUNx09Ki9Bu40n3QeqUedK6E2SF1z1
uJ1C5TUeeWabS04ifmLoxZO84ECviz+VV/GP0ABVy3gUiqhkk7xA7YWEIZXX
+7CYOcsKYkG/XyeE5AUir5+tpPKyGmEVStMTYG3TygW8tr3QJPriJpUXn1nX
56jwJCj2FGGSvMAhtjqCyoub+IvniWYSmE39MSJ5kd+rTreovFS/zTcTbEqA
rYrzV5K8oHL4yG8lkpesQilz12ADXlXtHCd5gfvRag6V17Oy1bsPyzaApYqA
/uaZDpCf12tL5RWXbnjxNKMR5gZLJpO8oOt7bACVV6V72j0J6Sb4tM7bjuQF
3R6xK6m8nGe9PT9l0gg7JpX/kbxgtqy2L5XX+WMBGyyWNoD72TgNkhc8ke53
oPKKtheYf9evE/0lvS6n9beiwOell6P3c8AgEDWURLpxrOHUuOXOLnx89HBI
TSkLPp2+caVRswOemQs9fjbWhJkl1xSqNIbAbNb90X/mTfBzl3Sj+et2/PX7
p4vUnSFwW9ZWFOfUDS3pyavD47oxbbVPXq0GC57vrZ1nfq0F7O4zuGLdNKzg
lbVkibDgpkxxiVRuDyx5VZyz6RUdV2qckE2d7AdBW5Ggk10tkKkuy1o6xkS2
U2rQA4N+yBPliG4O6wHZc90G7AImavnOrBOrpcFlSVNJw+tV0LY3cuFAAwe5
EpL7i+f1QX6E0J8WyS6QEF56u7+BjScLrzes2t9OHHGR80+vE6Jn+agu8OQg
f9ml4MA9rXCnutT3GrcC518fC0/d3ImM1qTLAruGoT12XYCMRRn4FLsf2P6j
D6MeR20zCeTAHVU1J8l5STjjWTqm/paOh674fDL+xgaJ8gg77egktPVKujIZ
NIhHpRXuT0UyIeF1tITrs0Scz7+VZ1UtByeccjUzHGhwu/yOZc93Mhd9Y+t4
TiNqfh5U9VTjgHeb836fXTWoLXNlgax0HXJj9memLB6F7haBLVeD6nB8lszs
mmvNOB0SO1L+ZxiKB21rv29vRI3rzxz4lfrwM+5NMusj781F+4pH1zahSvhG
baH+ftwl9Sg9TWgApMJvCcRWNaLV2lMLlU+xsCzo746AnXQ4v9/ezN+WDuMK
0YIV+Rk4+8o5xu75HFi3r/2hIA8db3UPlN6sp+OM+FPuRCQdhoa1ptZ0DWLj
2z+lHJEy5Gz2cNkQ1Q97Nk/5SBXSUO5birCgHhP/xm+3ucLTBx3ZaCI6m408
GPO17G4lNh3bdcKsmQ7SZ0qv5SvRMHetkZXpzkHUYK/RSrrZCfZSjm9cGtmY
q308aX13LXb5Tu61AhpcwyvPdhQxAEVe8sr8yMbkcPcjW3+z4G1Un+KaDf24
8f2KmTyrTyhS+Uz/73MWiHKddLhODMyWnzvJmFsJ61qNOJeDWGDbHdh4Qn8A
+bpai5MiCjEwtSS0xncAGstviF42ZCLX7bPYXuV6uO95hkvfNwD/Orc2HfIY
wFRIPrzYoBn0l7Wc7JLphwOe6c6iKoPYb33JI2JdO1RZRg8zFtLhVm2DQewK
Dho/H2BuuluHn//qhfR59YH3Af1of69BrFNvN9Eo6oJpF6Pp17v7gHF29iVr
8r02qLPxzfQrgZdBKflXpDgQE5er9rG9D0uTXtGP0AvB7jHjvn8zB/Zsq1/e
pNmBwieTBqLHmiB2fMXy/3vatCcy3qkbH+QM7AqL64YgZkcx1cf4qfr8Rbk9
OJNaqEZ6Bw6BzipU7+wLnBWUwnowVGp7BauACXs3nTpE9WvVVHJQln83qrao
n9xCzmGtsfPD6bw9EFkucab80SAaNv0LUnbsgaXHa3GyrgdmdHs2tUl24VNJ
A1HSO7i9TF51Nemd9FOLCp7jTShq1KhHzg1ogz+XqXPD++h2gePXWrCGxzWS
nA+wYBHdhjofQubeiz/V1YIWp70CyDkArk1iCdQ5oP5yUtvoehUqrTg6wGzg
QGz7dkuq77V3DsjN2teJdjJW9SKeHHCjXZygev3YCZ1duBUwPv1Ok/QaGr+7
nqN6rW+jfXGxRRlaFzi7kl5DgfgiNarXMxYRdkLzksDN4/4M6TUc/Fv/f69F
PNfIq0cnQfWVstuk11BhrXiP6vWFluZNl58lQrbRT0HSa3Ac2r2T6nXQtFFp
9/cefGHJmzGR00i+Z13qVK8tVs/p9N5VA59+jZ0gvQbeXPdCqteXNKwtnYLq
YJyWaEd6DQl7KqaoXq/jfGv6tr0RdtH40kmvwUaYm0z1+oDvqvcja5ugc/BA
COk1+P3ZkUT12uWq/MiLqkaI89A7TXoNV1OrVlC93hZ9Vof0Gv/cG5Wqyc8A
nh3x/VSvDYS5AqTXsPb8hQu36ukwe7K8jOq1a2dzOek1aNrJzuKKlEH2tu9n
qV4furvlx8JCGqi8tnpPeg307ac3U70O1l0lTnoNHDfV5qq7lUA/PQ5Urxv2
jTTkKZHfTUsWzJBeg+MgqyKR9HrHZhVz0muYVjl+fkt3LQwtfS5P9ZqWusWb
9BoVqqb65H5kw909zqZUr13nW4mRXkPbgonlxVafIPFWsh7Va/qmIGfSa3iY
E+dGm1uJGvp8o1SvY2d7vSe9hlGnsPWpEYXg58/xo3qtcCfwEOk1GAyG3dRR
rkeFzrx/VK8fOJ9bb+IxABV0votSBs3oVyl0m+q14vc2Juk11OeeOBG2rh0D
f32R6Se9lplU4MSsIP4aGfix5W4d5HqN2lG9XlZhovjQaxDEUz3/qhd1IZbw
qb0hvdbUKgwnvcaKbTIh6X4leP+WUS7Va+HkTUVrxRkodnabquTLcgxQm564
MMaCjF2qf0m/QPlS3nolxx40/bJO+yfpl0Gs8KZOktNwuLJRoi8Tn2xLzI5M
74LU23tmpWzpQ4QvL5pv1eOhxJZt1Rs5oM36l6y3qgVy8/aMqAywkCa1ckFf
WR+0CiX2t5PnlH/WKkwiz9nT2/T5MXnO3ir/qz6R7bD34pqYwcp2dMjiTfE7
xoFZJTHK0ydpkLs544Xn7QpMFOgRWiTLAZe/q99LBjKR8blzPmesHcMNhJfr
lTKg8Xdoc84/JhZdm9PmerwHPbt4HHSjaLAI75zpbunDg70RcrFCLLS1Vp7f
cKMd/t2JxVjBAXxgGLlgdBUNk1KXr/fV6QW6ZeyL6eP9WMDAik+uryCb4ZhR
EMcChb/lM2l+A1jksMQk7t17GDAe8ep8MgBnLM78fKLGwoS2rRL62lnAnjTG
zZ/7QeLQoWkPdzaeklxwhtbzCXRj4y76zWPAD5EHAodvcnDlrvaQxS4IEys/
NtX40ODbKYswkgsE7skNkHpZDh82/f1B5eJh77GmJq8H51xm32nzqQXXRhnR
HbOHYFejQNhhbxp6mr/peDrWDF/3GrcoXmDBB292WEUZHduO7M93tuqCKZmE
B7ZvmLDEd0HyEj0GdvV4PV8oRgORrG+BcTwMCOYvrU5TZ+Bz2i2dtnsMkBO1
0R3x6YN7R819O8h7V32zOj/JCxp8NNSp3G+9lyzdt6oF//xhR5F8YU6+zEYq
3y42b/KDyHa8qxloxqpsB3biwddUjmLRarNIjhjx5Kq/1+0KqH+SI0LluKnk
pAHJEcZ2JNcNjbXD3Eeqf/aSHLWrT5iSHGEwo7D05vEeuLskXo3KcaPl789d
LX2wqdE6g+QIEuYBIfUkx7K1gqtIjqAZulycu4qcg3XFr3xIjj2B3FSSI+yQ
Ebd75/oKMxsvZVI5WvydteC13wBYl+60efruPeqfuOJL5Xh4PGbBUzUWfMkQ
Ob1bOwtfPDtZTuXIm1+11NOdDWvnHhtq6fmELPX+W1SOl39eVCE5QrLIr5PC
LohWPo9YVI75j+/7Ur14cKS9oZH0YkuWx0aqF02Jj+Y+kutFtIjaDcUteP52
puxjPzas/5mydOW2PrjY/unyhtMs9OkpDK7ubofV/EeYRVos3BZx00/QvQOF
vrWaRS3vg0AxiVWO3Rx8dmVvmLLNO0z9fsW2SpsGQ6kr9V/cYePcJ7NHLqQ0
gtiuRJ9lKX2Q9E5J6MxTNhr2pw8/NGwC7uWEhq/6fdBw5s0J8nyw/7FKXNi9
A7gTXbLU86+qvhYiz4dDze4XwOYd+GrHnqCef615TRZ5PrwsfSl1PqURP7Uo
ZlPPv3m76yFm0TDo8cEH7vG1eNu07Pc9bzZciix3I+tC7vxiXj/DJnQb/Dmn
jqzrp6XNXreZCWW823u+ttCRzQopj33bDVdcHnzWv0tDnavNG2UODuCPi7dD
7iV1w3fFXT6MAQZKFVpJHJZkYGHxorQEzT7wXxIhR+4HF7nMWnI/mD2vXE7d
/5q/b4w+wIB2v2+95H7IMm50ou6vzOzJaR/vRV2bxwlZH9rwi9Ar7dIGFmRd
o/2NtxvExeFpnYo8naBnFuGyejUNCpasPxvcyUKBMsVBYdF6GPZSGc/XosOc
LEUpBfEh1Pa86JY39Q6Tz35eYjXaBzwuouPjf9g4Em2lV8b9in8be5fV9PWB
1veAn4uNmbj+tZP1kUEaGIqNFSlH9sK4q7M2+RzQ6deXmroWOkg052tRn4Nb
+jOnBv9BVNn8fW/h027gc62JXprYC5qr2u+8tBuE3yHHotfydKJuQlkatU+m
pHcO2Sc4NaHVPNF6THVeuxjJPs+rzP+zVnwIRAIXV1ZOvYPDDetFqX36iGyL
IvuE/ZVWatXcrxA5NGeqmuwz1+1ZsLQxEzKcZZ1NB2n40/05bCX7nKyiv3gr
zYRFyxvThsj5IX/s+qqysi7YKZI1t9F/EIKvJrDzn3ZjuPNxGVmyz7Elq904
r+lwJ+tBhf00E13tZhhiGZ0w8UvsSYEKG2MKF/ZMWTZA1roQWvdlGgwcsPlN
rsPUJH3TpGUDrhE8LtRDrvNcVLg+mdeHRW0CId8UGrHTLMPSwYcNZhpGzAly
nb8x0nuMXJ+UKj9NXd+1QuRXPm8fOF7fIbuskIW76w8v3E1rA6ZDvTCfMxvP
RPHPPvyhGkfaa8Ry59ChVDh8v84gG52nxVJWyKchW+Xo6+oMOmT7yi2PK2Hj
PtllWx7z5YKrFcO8NoEOCkLqI1+MWFiUnx2gpteJRpUetNCrJEcdkRGNtxzs
EKqR6zEqgA8/ah4c3UyDyDX7i+Y4s2Fo0cpzZh+qgcV1//6ZrBsjckSdrAul
5wtytsunwYz7xQRqXWGP3TpkXXjfuVj/IV8uLk24Z0+t269/7iRZF/ZKFS5X
J78j/fcGnKXWfZywSlbzLQdCz9dVtxoVYPpNiWfUuh83nvG8pEPHMAGNx1vf
E/9XmVW9ZrNg4b9lX4dILpk/n5ifI7kIXRsYECe5tJt7jL+RZqKI7MHoYT0G
GDJ0okpJvvOr5VJkVgzitR5/m4MNxVirevXMpodMSPlSO6EyPIgfox5NnlFI
wCAfH5ODqkxo1lPRSJpi4ppXJo23tBtBs8rrwvWNTJDlnOZL0BnER59uL/nH
WwmaxwNnvZvuh0Bb10TRWyy8z7u/vUS/HDXv5oXceM+AJ89Xa2kcZmOxRLnC
QHUKHvj0ZjLSjQGcfPVd7uwBjC1al/CSOFtgu113Mnlfuu7J1Q27wML9m6yD
vzh/BbVC7VjeZQyo9KVLsO05ePeyjJPqtTeYXpPj9m6SBnVamV8LtTjIeOq0
u+ZQPqx8QfP720+D4qreu2ResNEvVDVtKIacovNHqXljVz/jkHnhoLN2mLNC
Anh4hRtT82balvqQeaGlJ0r8pnYjbrRc50XNe81l904yL4znWqyd5q1ExViQ
pObdvYhzicwLM6Xityr0yyFsie1tal5rZb8tZF6YkBU697s6BWS3K36j5hUy
nxdD5oVrKa+j4/a0YgTPljkpZN7y3xOPyLzgKruwt8j5K64IN0dqXi+xjL8s
ew4slrY5bHztDYy/trr2/7wjAb/IvPCYbXCj8FA+Wj8xekbNe/3TAffWRBae
fTezkn62DuQ/pN4ce0OHpxEzNeQ6yA6dy+g9W4d07pun1PUkL4vT00xyTn7q
UGdhNdz4aPrqM38/3LZd6NnZ24/3ljjMj2d1QEoOm750Uz/cUb5ps19rAA9+
rUk7vq8JKm+5KY/F9UNF1r2j4eQ85D/XZH0xvhcKTZUvVdFosObhHnwhPojv
rcfN6re3wdcpt6LqENLHQLsXZF0IvhPPZGA1Zt6KKaLWvfiHq9MdOYha8x07
E6yrYGJy//Yjd/vBbuMed3IdLlkeyIq1rkK6WrERdR2GJ1ctJd9vQbfD/ZJi
NDShX82m/OPxSJFN9g+nV3jYx7I6cHWTzhxZsv89Wp8yyf5hVlB8m9m+JnxW
NmhE7X+5hJPla3UGnJ/5e7/1HgNtLsbdorzUdeRYK5kLwkLXRzrE92KqcMgL
aq6+JymKMeKDULdihdbX7W0oNDB7hJrLp0FbwXstE84/6VFQO9aNs17IW7XM
YkDZ8+njCwWYIG//I1poVRfe6e58GxLBANPrj9dyzZnw3PVmRAazD/PoAzeD
v/bB1og/Ry8dY4JMZf5uHos+fBGmPkg7SoP/AJP0doE=
            "]], {
          Axes -> True, BoxRatios -> {1, 1, 0.4}, 
           Method -> {"RotationControl" -> "Globe"}, 
           PlotRange -> {All, All, All}, 
           PlotRangePadding -> {Automatic, Automatic, Automatic}}]}}; 
     Typeset`initDone$$ = True),
    SynchronousInitialization->True,
    UnsavedVariables:>{Typeset`initDone$$},
    UntrackedVariables:>{Typeset`size$$}], "Manipulate",
   Deployed->True,
   StripOnInput->False],
  Manipulate`InterpretManipulate[1]]], "Output",
 CellChangeTimes->{3.518182447952094*^9},
 CellTags->"FIGURE 14.17  \[LightBulb]"],

Cell[TextData[{
 "\tWe now let ",
 Cell[BoxData[
  FormBox["\[CapitalDelta]", TraditionalForm]]],
 " be the maximum value of ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[CapitalDelta]", "\[VeryThinSpace]", 
     SubscriptBox["s", "1"]}], ",", "\[Ellipsis]", ",", 
    RowBox[{"\[CapitalDelta]", "\[VeryThinSpace]", 
     SubscriptBox["s", "n"]}]}], TraditionalForm]]],
 ". If the limit of the Riemann sums as ",
 Cell[BoxData[
  FormBox[
   RowBox[{"n", "\[Rule]", "\[Infinity]"}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[CapitalDelta]", "\[Rule]", "0"}], TraditionalForm]]],
 " exists over all partitions, the limit is called a ",
 StyleBox["line integral",
  FontSlant->"Italic"],
 ", and it gives the area of the curtain."
}], "Text",
 CellChangeTimes->{{3.4791675460852757`*^9, 3.479167655337818*^9}}],

Cell[TextData[{
 StyleBox["DEFINITION", "DefinitionFont"],
 "\t",
 StyleBox["Scalar Line Integral in the Plane, Arc Length Parameter",
  FontWeight->"Bold"],
 "\nSuppose the scalar-valued function ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " is defined on the smooth curve ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"C", ":", 
     RowBox[{
      StyleBox["r",
       FontWeight->Bold,
       FontSlant->Plain], "(", "s", ")"}]}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"x", "(", "s", ")"}], ",", 
      RowBox[{"y", "(", "s", ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", parameterized by the arc length ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 ". The ",
 StyleBox["line integral of ",
  FontWeight->"Bold"],
 Cell[BoxData[
  FormBox["f", TraditionalForm]],
  FontWeight->"Bold"],
 StyleBox[" over ",
  FontWeight->"Bold"],
 Cell[BoxData[
  FormBox["C", TraditionalForm]],
  FontWeight->"Bold"],
 " is \n",
 Cell[BoxData[
  FormBox[GridBox[{
     {GridBox[{
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["\[Integral]", "C"], 
           RowBox[{
            RowBox[{"f", "(", 
             RowBox[{
              RowBox[{"x", "(", "s", ")"}], ",", 
              RowBox[{"y", "(", "s", ")"}]}], ")"}], "d", "\[VeryThinSpace]", 
            "s"}]}], "=", 
          FormBox[
           RowBox[{
            RowBox[{
             UnderscriptBox["lim", 
              RowBox[{"\[CapitalDelta]", "\[Rule]", "0"}]], 
             RowBox[{
              UnderoverscriptBox["\[Sum]", 
               RowBox[{"k", "=", "1"}], "n"], 
              RowBox[{
               RowBox[{"f", "(", 
                RowBox[{
                 RowBox[{"x", "(", 
                  SubscriptBox[
                   OverscriptBox["s", "_"], "k"], ")"}], ",", 
                 RowBox[{"y", "(", 
                  SubscriptBox[
                   OverscriptBox["s", "_"], "k"], ")"}]}], ")"}], 
               "\[CapitalDelta]", "\[VeryThinSpace]", 
               SubscriptBox["s", "k"]}]}]}], ","}],
           TraditionalForm]}]}
       },
       GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
    },
    GridBoxItemSize->{"Columns" -> {{
         Scaled[0.96]}}}], TraditionalForm]]],
 "\nprovided this limit exists over all partitions of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". When the limit exists, ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " is said to be ",
 StyleBox["integrable",
  FontWeight->"Bold"],
 " on ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 "."
}], "Definition",
 CellChangeTimes->{{3.479167666978815*^9, 3.4791677585803022`*^9}, {
   3.4791678383933687`*^9, 3.479167868643135*^9}, 3.479851145768944*^9},
 CellTags->
  "DEFINITION Scalar Line Integral in the Plane, Arc Length Parameter"],

Cell[TextData[{
 "\tThe more compact notation ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Integral]", "C"], 
    RowBox[{
     RowBox[{"f", "(", 
      RowBox[{
       StyleBox["r",
        FontWeight->"Bold",
        FontSlant->"Plain"], "(", "s", ")"}], ")"}], "d", "\[VeryThinSpace]", 
     "s"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Integral]", "C"], 
    RowBox[{
     RowBox[{"f", "(", 
      RowBox[{"x", ",", "y"}], ")"}], "d", "\[VeryThinSpace]", "s"}]}], 
   TraditionalForm]]],
 ", or ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Integral]", "C"], 
    RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}], TraditionalForm]]],
 " is often used for the line integral of ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " over ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". It can be shown that if ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " is continuous on ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", then the line integral of ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " over ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " exists. If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"f", "(", 
     RowBox[{"x", ",", "y"}], ")"}], "=", "1"}], TraditionalForm]]],
 ", the line integral ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Integral]", "C"], 
    RowBox[{"d", "\[VeryThinSpace]", "s"}]}], TraditionalForm]]],
 " gives the length of the curve, just as the ordinary integral ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubsuperscriptBox["\[Integral]", 
     RowBox[{" ", "a"}], 
     RowBox[{" ", "b"}]], 
    RowBox[{"d", "\[VeryThinSpace]", "x"}]}], TraditionalForm]]],
 " gives the length of the interval ",
 Cell[BoxData[
  FormBox[
   RowBox[{"[", 
    RowBox[{"a", ",", "b"}], "]"}], TraditionalForm]]],
 ", which is ",
 Cell[BoxData[
  FormBox[
   RowBox[{"b", "-", "a"}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.479167917341848*^9, 3.479168119645726*^9}, 
   3.479168173703837*^9}],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["EXAMPLE 1\t", "ExampleFont"],
 "Average temperature on a circle"
}], "Subsubsubsection",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.479168186834433*^9, 3.4791681948450603`*^9}},
 CellTags->"EXAMPLE 1 Average temperature on a circle"],

Cell[TextData[{
 "The temperature of the circular plate ",
 Cell[BoxData[
  FormBox[
   RowBox[{"R", "=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"x", ",", "y"}], ")"}], ":", 
      RowBox[{
       RowBox[{
        SuperscriptBox["x", "2"], "+", 
        SuperscriptBox["y", "2"]}], "\[LessEqual]", "1"}]}], "}"}]}], 
   TraditionalForm]]],
 " is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"T", "(", 
     RowBox[{"x", ",", "y"}], ")"}], "=", 
    RowBox[{"100", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["x", "2"], "+", 
       RowBox[{"2", 
        SuperscriptBox["y", "2"]}]}], ")"}]}]}], TraditionalForm]]],
 ". Find the average temperature along the edge of the plate."
}], "Text",
 CellChangeTimes->{{3.479168199093853*^9, 3.479168245901264*^9}}],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   PaneBox[
    StyleBox[
     GraphicsBox[
      {RGBColor[0.976, 0.855, 0.514], EdgeForm[RGBColor[0.68, 0.57, 0.34]], 
       PolygonBox[{{0.2, 0}, {0, -0.3}, {0.9, 0}, {0, 0.3}, {0.2, 0}}]},
      ImageSize->15,
      PlotRange->{-0.5, 0.3}], "Note",
     Editable->False,
     Selectable->False],
    Alignment->{Center, Top},
    ImageSize->{18, 13}],
   Background->GrayLevel[1],
   FrameMargins->{{5, 5}, {-1, -0.1}},
   FrameStyle->RGBColor[0.976, 0.855, 0.514],
   RoundingRadius->2]]],
 "  ",
 StyleBox["Note", "CalloutIconFont"]
}], "CalloutIcon",
 WholeCellGroupOpener->True],

Cell["\<\
When we compute the average value by an ordinary integral, we divide by the \
length of the interval of integration. Analogously, when we compute the \
average value by a line integral, we divide by the length of the curve.\
\>", "Callout",
 CellChangeTimes->{{3.479168278611835*^9, 3.4791683321761637`*^9}}]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["SOLUTION", "SolutionFont"]], "Subsubsubsubsection",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "Calculating the average value requires integrating the temperature function \
over the boundary circle ",
 Cell[BoxData[
  FormBox[
   RowBox[{"C", "=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x", ",", "y"}], ")"}], ":", 
       RowBox[{
        SuperscriptBox["x", "2"], "+", 
        SuperscriptBox["y", "2"]}]}], "=", "1"}], "}"}]}], TraditionalForm]]],
 " and dividing by the length (circumference) of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". The first step is to find a parametric description for ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". Recall from Section 11.9 that a parametric description of a unit circle \
using arc length as the parameter is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    StyleBox["r",
     FontWeight->Bold,
     FontSlant->Plain], "=", 
    RowBox[{
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{"x", ",", "y"}], "\[RightAngleBracket]"}], "=", 
     RowBox[{"\[LeftAngleBracket]", 
      RowBox[{
       RowBox[{"cos", " ", "s"}], ",", 
       RowBox[{"sin", " ", "s"}]}], "\[RightAngleBracket]"}]}]}], 
   TraditionalForm]]],
 ", for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "\[LessEqual]", "s", "\[LessEqual]", 
    RowBox[{"2", "\[Pi]"}]}], TraditionalForm]]],
 ". We substitute ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"cos", " ", "s"}]}], TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", "=", 
    RowBox[{"sin", " ", "s"}]}], TraditionalForm]]],
 " into the temperature function and express the line integral as an ordinary \
integral:"
}], "Text",
 CellChangeTimes->{{3.479168374424218*^9, 3.4791685147217484`*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Integral]", "C"], 
          RowBox[{
           RowBox[{"T", "(", 
            RowBox[{"x", ",", "y"}], ")"}], "d", "\[VeryThinSpace]", "s"}]}], 
         "=", 
         FormBox[
          RowBox[{
           SubsuperscriptBox["\[Integral]", 
            RowBox[{" ", "0"}], 
            RowBox[{" ", 
             RowBox[{"2", "\[Pi]"}]}]], 
           RowBox[{
            FormBox[
             FormBox[
              UnderscriptBox[
               UnderscriptBox[
                RowBox[{"100", "[", 
                 RowBox[{
                  SuperscriptBox[
                   RowBox[{"x", "(", "s", ")"}], "2"], "+", 
                  RowBox[{"2", 
                   SuperscriptBox[
                    RowBox[{"y", "(", "s", ")"}], "2"]}]}], "]"}], 
                StyleBox["\[UnderBrace]", "TypesetAnnotationFont"]], 
               StyleBox[
                RowBox[{"T", "(", "s", ")"}], "TypesetAnnotationFont"]],
              TraditionalForm],
             TraditionalForm], " ", "d", "\[VeryThinSpace]", "s"}]}],
          TraditionalForm]}], 
        StyleBox[
         RowBox[{
         "Write", " ", "the", " ", "line", " ", "integral", " ", "with", " ", 
          "respect", " ", "to", " ", 
          RowBox[{"s", "."}]}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{"100", 
          FormBox[
           RowBox[{
            SubsuperscriptBox["\[Integral]", 
             RowBox[{" ", "0"}], 
             RowBox[{" ", 
              RowBox[{"2", "\[Pi]"}]}]], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                SuperscriptBox["cos", "2"], "s"}], "+", 
               RowBox[{"2", 
                SuperscriptBox["sin", "2"], "s"}]}], ")"}], " ", "d", 
             "\[VeryThinSpace]", "s"}]}],
           TraditionalForm]}]}], 
        StyleBox[
         RowBox[{"Substitute", " ", "for", " ", "x", " ", "and", " ", 
          RowBox[{"y", "."}]}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{"100", 
          FormBox[
           FormBox[
            FormBox[
             UnderscriptBox[
              UnderscriptBox[
               RowBox[{
                SubsuperscriptBox["\[Integral]", 
                 RowBox[{" ", "0"}], 
                 RowBox[{" ", 
                  RowBox[{"2", "\[Pi]"}]}]], 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{"1", "+", 
                   RowBox[{
                    SuperscriptBox["sin", "2"], "s"}]}], ")"}], " ", "d", 
                 "\[VeryThinSpace]", "s"}]}], 
               StyleBox["\[UnderBrace]", "TypesetAnnotationFont"]], 
              StyleBox[
               RowBox[{"3", "\[Pi]"}], "TypesetAnnotationFont"]],
             TraditionalForm],
            TraditionalForm],
           TraditionalForm]}]}], 
        StyleBox[
         RowBox[{
          RowBox[{
           RowBox[{
            SuperscriptBox["cos", "2"], "s"}], "+", 
           RowBox[{
            SuperscriptBox["sin", "2"], "s"}]}], "=", "1"}], 
         "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{"300", "\[Pi]"}]}], 
        StyleBox[
         RowBox[{
          RowBox[{"Use", " ", 
           SuperscriptBox["sin", "2"], "s"}], "=", 
          RowBox[{
           FractionBox[
            RowBox[{"1", "-", 
             RowBox[{"cos", " ", "2", "s"}]}], "2"], " ", "and", " ", 
           RowBox[{"integrate", "."}]}]}], "TypesetAnnotationFont"]}
      },
      GridBoxAlignment->{"Columns" -> {{"=", Left}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.4791685217573*^9, 3.479168538857205*^9}, {
  3.479168584141116*^9, 3.479168769457736*^9}, {3.5177503682967367`*^9, 
  3.51775038769256*^9}}],

Cell[TextData[{
 "\tThe geometry of this line integral is shown in ",
 StyleBox["Figure 14.18", "FigureFontText"],
 ". The temperature function on the boundary of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " is a function of ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 ". The line integral is an ordinary integral with respect to ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 " over the interval ",
 Cell[BoxData[
  FormBox[
   RowBox[{"[", 
    RowBox[{"0", ",", 
     RowBox[{"2", "\[Pi]"}]}], "]"}], TraditionalForm]]],
 ". To find the average value we divide the line integral of the temperature \
by the length of the curve, which is ",
 Cell[BoxData[
  FormBox[
   RowBox[{"2", "\[Pi]"}], TraditionalForm]]],
 ". Therefore, the average temperature on the boundary of the plate is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"300", 
     RowBox[{"\[Pi]", "/", 
      RowBox[{"(", 
       RowBox[{"2", "\[Pi]"}], ")"}]}]}], "=", "150"}], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.4791688570562696`*^9, 3.479168996993842*^9}}],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   PaneBox[
    StyleBox[
     GraphicsBox[
      {RGBColor[0.976, 0.855, 0.514], EdgeForm[RGBColor[0.68, 0.57, 0.34]], 
       PolygonBox[{{0.2, 0}, {0, -0.3}, {0.9, 0}, {0, 0.3}, {0.2, 0}}]},
      ImageSize->15,
      PlotRange->{-0.5, 0.3}], "Note",
     Editable->False,
     Selectable->False],
    Alignment->{Center, Top},
    ImageSize->{18, 13}],
   Background->GrayLevel[1],
   FrameMargins->{{5, 5}, {-1, -0.1}},
   FrameStyle->RGBColor[0.976, 0.855, 0.514],
   RoundingRadius->2]]],
 "  ",
 StyleBox["Note", "CalloutIconFont"]
}], "CalloutIcon",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "The line integral in Example 1 also gives the area of the cylindrical \
curtain that hangs between the surface and ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " in figure 14.18."
}], "Callout",
 CellChangeTimes->{{3.479169053052137*^9, 3.479169076758348*^9}}]
}, Open  ]],

Cell[TextData[{
 "Related Exercises 11\[LongDash]14 ",
 Cell[BoxData[
  GraphicsBox[
   {RGBColor[0.796, 0, 0.42], EdgeForm[RGBColor[0.5306666666666667, 0, 0.28]],
     PolygonBox[{{0, 0}, {0.9, -0.3}, {0.7, 0}, {0.9, 0.3}, {0, 0}}]},
   ImageSize->15,
   PlotRange->{-0.5, 0.3}]],
  CellChangeTimes->{3.443312659351997*^9}]
}], "RelatedExercises",
 CellChangeTimes->{{3.479169002513977*^9, 3.4791690042071047`*^9}}]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Parameters Other than Arc Length", "Subsubsection",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.479169123384035*^9, 3.479169130579371*^9}}],

Cell[TextData[{
 "The line integral in Example 1 is straightforward because a circle is \
easily parameterized in terms of the arc length. Suppose we have a \
parameterized curve with a parameter ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " that is ",
 StyleBox["not",
  FontSlant->"Italic"],
 " the arc length. The key is a change of variables. Assume the curve ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " is described by ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     StyleBox["r",
      FontWeight->Bold,
      FontSlant->Plain], "(", "t", ")"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"x", "(", "t", ")"}], ",", 
      RowBox[{"y", "(", "t", ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[LessEqual]", "t", "\[LessEqual]", "b"}], TraditionalForm]]],
 ". Recall from Section 11.9 that the length of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " over the interval ",
 Cell[BoxData[
  FormBox[
   RowBox[{"[", 
    RowBox[{"a", ",", "t"}], "]"}], TraditionalForm]]],
 " is"
}], "Text",
 CellChangeTimes->{{3.479169136061545*^9, 3.479169263009581*^9}, {
  3.4841314750209427`*^9, 3.484131475335025*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"s", "(", "t", ")"}], "=", 
         RowBox[{
          FormBox[
           RowBox[{
            SubsuperscriptBox["\[Integral]", 
             RowBox[{" ", "a"}], 
             RowBox[{" ", "t"}]], 
            RowBox[{
             RowBox[{"\[LeftBracketingBar]", 
              RowBox[{
               RowBox[{
                StyleBox["r",
                 FontWeight->Bold,
                 FontSlant->Plain], "'"}], 
               RowBox[{"(", "u", ")"}]}], "\[RightBracketingBar]"}], " ", "d",
              "\[VeryThinSpace]", "u"}]}],
           TraditionalForm], "."}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.4791692693514*^9, 3.479169302582632*^9}}],

Cell[TextData[{
 "Differentiating both sides of this equation and using the Fundamental \
Theorem of Calculus yields ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"s", "'"}], 
     RowBox[{"(", "t", ")"}]}], "=", 
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      RowBox[{
       StyleBox["r",
        FontWeight->Bold,
        FontSlant->Plain], "'"}], 
      RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ". We now make a standard change of variables using the relationship"
}], "Text",
 CellChangeTimes->{{3.4791693512086277`*^9, 3.479169408381856*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"d", "\[VeryThinSpace]", "s"}], "=", 
         RowBox[{
          RowBox[{
           RowBox[{"s", "'"}], 
           RowBox[{"(", "t", ")"}], "d", "\[VeryThinSpace]", "t"}], "=", 
          RowBox[{
           RowBox[{"\[LeftBracketingBar]", 
            RowBox[{
             RowBox[{
              StyleBox["r",
               FontWeight->Bold,
               FontSlant->Plain], "'"}], 
             RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "d", 
           "\[VeryThinSpace]", 
           RowBox[{"t", "."}]}]}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479169423391366*^9, 3.479169444090025*^9}}],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   PaneBox[
    StyleBox[
     GraphicsBox[
      {RGBColor[0.976, 0.855, 0.514], EdgeForm[RGBColor[0.68, 0.57, 0.34]], 
       PolygonBox[{{0.2, 0}, {0, -0.3}, {0.9, 0}, {0, 0.3}, {0.2, 0}}]},
      ImageSize->15,
      PlotRange->{-0.5, 0.3}], "Note",
     Editable->False,
     Selectable->False],
    Alignment->{Center, Top},
    ImageSize->{18, 13}],
   Background->GrayLevel[1],
   FrameMargins->{{5, 5}, {-1, -0.1}},
   FrameStyle->RGBColor[0.976, 0.855, 0.514],
   RoundingRadius->2]]],
 "  ",
 StyleBox["Note", "CalloutIconFont"]
}], "CalloutIcon",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " represents time, then the relationship ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"d", "\[VeryThinSpace]", "s"}], "=", 
    RowBox[{
     RowBox[{"\[LeftBracketingBar]", 
      RowBox[{
       RowBox[{
        StyleBox["r",
         FontWeight->Bold,
         FontSlant->Plain], "'"}], 
       RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "d", 
     "\[VeryThinSpace]", "t"}]}], TraditionalForm]]],
 " is a generalization of the familiar formula ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    StyleBox["distance",
     FontSlant->"Italic"], "=", 
    RowBox[{
     RowBox[{"(", 
      StyleBox["speed",
       FontSlant->"Italic"], ")"}], 
     RowBox[{"(", 
      StyleBox["time",
       FontSlant->"Italic"], ")"}]}]}], TraditionalForm]]],
 "."
}], "Callout",
 CellChangeTimes->{{3.479169605503294*^9, 3.4791696407365503`*^9}}]
}, Open  ]],

Cell[TextData[{
 "The original line integral with respect to ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 " is now converted into an ordinary integral with respect to ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.479169447353115*^9, 3.4791694678543997`*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Integral]", "C"], 
          RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}], "=", 
         FormBox[
          RowBox[{
           SubsuperscriptBox["\[Integral]", 
            RowBox[{" ", "a"}], 
            RowBox[{" ", "b"}]], 
           RowBox[{
            RowBox[{"f", "(", 
             RowBox[{
              RowBox[{"x", "(", "t", ")"}], ",", 
              RowBox[{"y", "(", "t", ")"}]}], ")"}], 
            FormBox[
             UnderscriptBox[
              UnderscriptBox[
               RowBox[{
                RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                  RowBox[{
                   StyleBox["r",
                    FontWeight->Bold,
                    FontSlant->Plain], "'"}], 
                  RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], " ", 
                "d", "\[VeryThinSpace]", "t"}], 
               StyleBox["\[UnderBrace]", "TypesetAnnotationFont"]], 
              StyleBox[
               RowBox[{"d", "\[VeryThinSpace]", "s"}], 
               "TypesetAnnotationFont"]],
             TraditionalForm]}]}],
          TraditionalForm]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479169506134885*^9, 3.4791695519629374`*^9}, {
  3.509881940072853*^9, 3.509881945836278*^9}}],

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   StyleBox[
    RowBox[{"QUICK", " ", "CHECK", " ", "1"}], "QuickCheckFont"],
   Alignment->Center,
   Background->RGBColor[0.976, 0.855, 0.514],
   FrameMargins->{{5, 5}, {2, 4}},
   FrameStyle->GrayLevel[0.5]]]],
 "  Explain mathematically why differentiating the arc length integral leads \
to ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"s", "'"}], 
     RowBox[{"(", "t", ")"}]}], "=", 
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      RowBox[{
       StyleBox["r",
        FontWeight->Bold,
        FontSlant->Plain], "'"}], 
      RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}]}], 
   TraditionalForm]]],
 ". ",
 Cell[BoxData[
  FormBox[
   GraphicsBox[
    {RGBColor[0.976, 0.855, 0.514], EdgeForm[RGBColor[0.68, 0.57, 0.34]], 
     PolygonBox[{{0, 0}, {0.9, -0.3}, {0.7, 0}, {0.9, 0.3}, {0, 0}}]},
    ImageSize->15,
    PlotRange->{-0.5, 0.3}], TraditionalForm]]]
}], "QuickCheck",
 CellChangeTimes->{{3.479169690834426*^9, 3.479169720505286*^9}},
 CellTags->"QUICK CHECK 1"],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FrameBox[
  StyleBox["\[Checkmark]", "SR",
   FontSize->13,
   FontWeight->"Bold"],
  Alignment->Center,
  Background->RGBColor[0.976, 0.855, 0.514],
  FrameMargins->{{4, 6}, {1, 5}},
  FrameStyle->GrayLevel[0.5],
  RoundingRadius->5]]]], "QuickCheckAnswerIcon",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "The Fundamental Theorem of Calculus says that ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     FractionBox["d", 
      RowBox[{"d", "\[VeryThinSpace]", "t"}]], 
     RowBox[{
      SubsuperscriptBox["\[Integral]", 
       RowBox[{" ", "a"}], 
       RowBox[{" ", "t"}]], 
      RowBox[{
       RowBox[{"f", "(", "u", ")"}], " ", "d", "\[VeryThinSpace]", "u"}]}]}], 
    "=", 
    RowBox[{"f", "(", "t", ")"}]}], TraditionalForm]]],
 ", which applies to differentiating the arc length integral."
}], "QuickCheckAnswer",
 CellChangeTimes->{3.521840622454032*^9}]
}, Closed]],

Cell[TextData[{
 StyleBox["THEOREM 14.1", "TheoremFont"],
 "\t",
 StyleBox["Evaluating Scalar Line Integrals in ",
  FontWeight->"Bold"],
 Cell[BoxData[
  FormBox[
   StyleBox[
    SuperscriptBox["\[DoubleStruckCapitalR]", "2"],
    FontWeight->"Bold"], TraditionalForm]]],
 "\nLet ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " be continuous on a region containing a smooth curve ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     StyleBox["r",
      FontWeight->Bold,
      FontSlant->Plain], "(", "t", ")"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"x", "(", "t", ")"}], ",", 
      RowBox[{"y", "(", "t", ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[LessEqual]", "t", "\[LessEqual]", "b"}], TraditionalForm]]],
 ". Then ",
 "\n",
 Cell[BoxData[
  FormBox[GridBox[{
     {GridBox[{
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["\[Integral]", "C"], 
           RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}], "=", 
          RowBox[{
           SubsuperscriptBox["\[Integral]", 
            RowBox[{" ", "a"}], 
            RowBox[{" ", "b"}]], 
           RowBox[{
            RowBox[{"f", "(", 
             RowBox[{
              RowBox[{"x", "(", "t", ")"}], ",", 
              RowBox[{"y", "(", "t", ")"}]}], ")"}], 
            FormBox[
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                RowBox[{
                 StyleBox["r",
                  FontWeight->Bold,
                  FontSlant->Plain], "'"}], 
                RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], " ", 
              "d", "\[VeryThinSpace]", "t"}],
             TraditionalForm]}]}]}]},
        {
         RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
          RowBox[{
           SubsuperscriptBox["\[Integral]", 
            RowBox[{" ", "a"}], 
            RowBox[{" ", "b"}]], 
           RowBox[{
            RowBox[{"f", "(", 
             RowBox[{
              RowBox[{"x", "(", "t", ")"}], ",", 
              RowBox[{"y", "(", "t", ")"}]}], ")"}], 
            RowBox[{
             FormBox[
              RowBox[{
               SqrtBox[
                RowBox[{
                 RowBox[{
                  RowBox[{"x", "'"}], 
                  SuperscriptBox[
                   RowBox[{"(", "t", ")"}], "2"]}], "+", 
                 RowBox[{
                  RowBox[{"y", "'"}], 
                  SuperscriptBox[
                   RowBox[{"(", "t", ")"}], "2"]}]}]], "d", 
               "\[VeryThinSpace]", "t"}],
              TraditionalForm], "."}]}]}]}]}
       },
       GridBoxAlignment->{"Columns" -> {{"="}}}]}
    },
    GridBoxItemSize->{"Columns" -> {{
         Scaled[0.96]}}}], TraditionalForm]]]
}], "Theorem",
 CellChangeTimes->{{3.4791697641911507`*^9, 3.479169847616704*^9}, 
   3.479170007235312*^9, {3.4841315074909897`*^9, 3.4841315154093933`*^9}},
 CellTags->"THEOREM 14.1 Evaluating Scalar Line Integrals in R2"],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   PaneBox[
    StyleBox[
     GraphicsBox[
      {RGBColor[0.976, 0.855, 0.514], EdgeForm[RGBColor[0.68, 0.57, 0.34]], 
       PolygonBox[{{0.2, 0}, {0, -0.3}, {0.9, 0}, {0, 0.3}, {0.2, 0}}]},
      ImageSize->15,
      PlotRange->{-0.5, 0.3}], "Note",
     Editable->False,
     Selectable->False],
    Alignment->{Center, Top},
    ImageSize->{18, 13}],
   Background->GrayLevel[1],
   FrameMargins->{{5, 5}, {-1, -0.1}},
   FrameStyle->RGBColor[0.976, 0.855, 0.514],
   RoundingRadius->2]]],
 "  ",
 StyleBox["Note", "CalloutIconFont"]
}], "CalloutIcon",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "The value of a line integral of a scalar-valued function is independent of \
the parameterization of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " and independent of the direction in which ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " is traversed (Exercises 54-55)."
}], "Callout",
 CellChangeTimes->{{3.479170040149514*^9, 3.4791700807954082`*^9}}]
}, Closed]],

Cell[TextData[{
 "\tIf ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " represents time and ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " is the path of a moving object, then ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftBracketingBar]", 
    RowBox[{
     RowBox[{
      StyleBox["r",
       FontWeight->Bold,
       FontSlant->Plain], "'"}], 
     RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], TraditionalForm]]],
 " is the speed of the object. The ",
 StyleBox["speed factor",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftBracketingBar]", 
    RowBox[{
     RowBox[{
      StyleBox["r",
       FontWeight->Bold,
       FontSlant->Plain], "'"}], 
     RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], TraditionalForm]]],
 " that appears in the integral relates distance traveled along the curve as \
measured by ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 " to the elapsed time as measured by the parameter ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.479170110505776*^9, 3.47917019671902*^9}, 
   3.4797775374308777`*^9}],

Cell[TextData[{
 "\tNotice that if ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " is the arc length ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 ", then ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      RowBox[{
       StyleBox["r",
        FontWeight->Bold,
        FontSlant->Plain], "'"}], 
      RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "=", "1"}], 
   TraditionalForm]]],
 " and we recover the line integral with respect to the arc length ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 ":"
}], "Text",
 CellChangeTimes->{{3.47917021804816*^9, 3.479170247676504*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{
          UnderscriptBox["\[Integral]", "C"], 
          RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}], "=", 
         RowBox[{
          SubsuperscriptBox["\[Integral]", 
           RowBox[{" ", "a"}], 
           RowBox[{" ", "b"}]], 
          RowBox[{
           RowBox[{"f", "(", 
            RowBox[{
             RowBox[{"x", "(", "s", ")"}], ",", 
             RowBox[{"y", "(", "s", ")"}]}], ")"}], 
           FormBox[
            RowBox[{"d", "\[VeryThinSpace]", "s"}],
            TraditionalForm]}]}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.4791702752302237`*^9, 3.4791702873022623`*^9}}],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"f", "(", 
     RowBox[{"x", ",", "y"}], ")"}], "=", "1"}], TraditionalForm]]],
 ", then the line integral is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubsuperscriptBox["\[Integral]", 
     RowBox[{" ", "a"}], 
     RowBox[{" ", "b"}]], 
    RowBox[{
     SqrtBox[
      RowBox[{
       RowBox[{
        RowBox[{"x", "'"}], 
        SuperscriptBox[
         RowBox[{"(", "t", ")"}], "2"]}], "+", 
       RowBox[{
        RowBox[{"y", "'"}], 
        SuperscriptBox[
         RowBox[{"(", "t", ")"}], "2"]}]}]], "d", "\[VeryThinSpace]", "t"}]}],
    TraditionalForm]]],
 ", which is the arc length formula for ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ". Theorem 14.1 leads to the following procedure for evaluating line \
integrals."
}], "Text",
 CellChangeTimes->{{3.4791702949533567`*^9, 3.4791703589727793`*^9}}],

Cell[TextData[{
 StyleBox["PROCEDURE", "ProcedureFont"],
 "\t",
 StyleBox["Evaluating the Line Integral ",
  FontWeight->"Bold"],
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["\[Integral]", "C"], 
     RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}],
    FontWeight->"Bold"], TraditionalForm]]],
 "\n",
 StyleBox["1.",
  FontWeight->"Bold"],
 "  Find a parametric description of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " in the form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     StyleBox["r",
      FontWeight->Bold,
      FontSlant->Plain], "(", "t", ")"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"x", "(", "t", ")"}], ",", 
      RowBox[{"y", "(", "t", ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[LessEqual]", "t", "\[LessEqual]", "b"}], TraditionalForm]]],
 ".\n",
 StyleBox["2.",
  FontWeight->"Bold"],
 "  Compute ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      RowBox[{
       StyleBox["r",
        FontWeight->Bold,
        FontSlant->Plain], "'"}], 
      RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "=", 
    SqrtBox[
     RowBox[{
      RowBox[{
       RowBox[{"x", "'"}], 
       SuperscriptBox[
        RowBox[{"(", "t", ")"}], "2"]}], "+", 
      RowBox[{
       RowBox[{"y", "'"}], 
       SuperscriptBox[
        RowBox[{"(", "t", ")"}], "2"]}]}]]}], TraditionalForm]]],
 ".\n",
 StyleBox["3.",
  FontWeight->"Bold"],
 "  Make substitutions for ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 " and ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 " in the integrand and evaluate an ordinary integral: ",
 "\n",
 Cell[BoxData[
  FormBox[GridBox[{
     {GridBox[{
        {
         RowBox[{
          RowBox[{
           UnderscriptBox["\[Integral]", "C"], 
           RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}], "=", 
          RowBox[{
           SubsuperscriptBox["\[Integral]", 
            RowBox[{" ", "a"}], 
            RowBox[{" ", "b"}]], 
           RowBox[{"f", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"x", 
               RowBox[{"(", "t", ")"}]}], ",", 
              RowBox[{"y", 
               RowBox[{"(", "t", ")"}]}]}], ")"}], 
            FormBox[
             RowBox[{
              RowBox[{"\[LeftBracketingBar]", 
               RowBox[{
                RowBox[{
                 StyleBox["r",
                  FontWeight->Bold,
                  FontSlant->Plain], "'"}], 
                RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], " ", 
              "d", "\[VeryThinSpace]", "t"}],
             TraditionalForm]}]}]}]}
       },
       GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
    },
    GridBoxItemSize->{"Columns" -> {{
         Scaled[0.96]}}}], TraditionalForm]]]
}], "Procedure",
 CellChangeTimes->{{3.479170383225875*^9, 3.479170519913622*^9}, 
   3.479851166558856*^9},
 CellTags->"PROCEDURE Evaluating the Line Integral"],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["EXAMPLE 2\t", "ExampleFont"],
 "Average temperature on a circle"
}], "Subsubsubsection",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.479170588535314*^9, 3.47917060285668*^9}},
 CellTags->"EXAMPLE 2 Average temperature on a circle"],

Cell[TextData[{
 "The temperature of the circular plate ",
 Cell[BoxData[
  FormBox[
   RowBox[{"R", "=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"x", ",", "y"}], ")"}], ":", 
      RowBox[{
       RowBox[{
        SuperscriptBox["x", "2"], "+", 
        SuperscriptBox["y", "2"]}], "\[LessEqual]", "1"}]}], "}"}]}], 
   TraditionalForm]]],
 " is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"T", "(", 
     RowBox[{"x", ",", "y"}], ")"}], "=", 
    RowBox[{"100", 
     RowBox[{"(", 
      RowBox[{
       SuperscriptBox["x", "2"], "+", 
       RowBox[{"2", 
        SuperscriptBox["y", "2"]}]}], ")"}]}]}], TraditionalForm]]],
 " as in Example 1. Confirm the average temperature computed in Example 1 \
when the circle has the parametric description"
}], "Text",
 CellChangeTimes->{{3.47917060784839*^9, 3.479170663715332*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{"C", "=", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{"x", ",", "y"}], ")"}], ":", "x"}], "=", 
             RowBox[{"cos", " ", 
              SuperscriptBox["t", "2"]}]}], ",", 
            RowBox[{"y", "=", 
             RowBox[{"sin", " ", 
              SuperscriptBox["t", "2"]}]}], ",", 
            RowBox[{"0", "\[LessEqual]", "t", "\[LessEqual]", 
             SqrtBox[
              RowBox[{"2", "\[Pi]"}]]}]}], "}"}], "."}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479170671952445*^9, 3.4791706971821413`*^9}}],

Cell[CellGroupData[{

Cell[TextData[StyleBox["SOLUTION", "SolutionFont"]], "Subsubsubsubsection",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "The speed factor on ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 " (using ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{
      SuperscriptBox["sin", "2"], "t"}], "+", 
     RowBox[{
      SuperscriptBox["cos", "2"], "t"}]}], "=", "1"}], TraditionalForm]]],
 ") is"
}], "Text",
 CellChangeTimes->{{3.479170732260714*^9, 3.479170752231604*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", 
          RowBox[{
           RowBox[{
            StyleBox["r",
             FontWeight->Bold,
             FontSlant->Plain], "'"}], 
           RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "=", 
         RowBox[{
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"x", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"y", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}]}]], "=", 
          RowBox[{
           SqrtBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"-", "2"}], "t", " ", "sin", " ", 
                SuperscriptBox["t", "2"]}], ")"}], "2"], "+", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"2", "t", " ", "cos", " ", 
                SuperscriptBox["t", "2"]}], ")"}], "2"]}]], "=", 
           RowBox[{"2", 
            RowBox[{"t", "."}]}]}]}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479170764090884*^9, 3.4791708096543207`*^9}}],

Cell["\<\
Making the appropriate substitutions, the value of the line integral is\
\>", "Text",
 CellChangeTimes->{{3.47917082594453*^9, 3.479170835244937*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         FormBox[
          RowBox[{
           UnderscriptBox["\[Integral]", "C"], 
           RowBox[{"T", " ", "d", "\[VeryThinSpace]", "s"}]}],
          TraditionalForm], "=", 
         FormBox[
          RowBox[{
           SubsuperscriptBox["\[Integral]", 
            RowBox[{" ", "0"}], 
            RowBox[{" ", 
             SqrtBox[
              RowBox[{"2", "\[Pi]"}]]}]], 
           RowBox[{"100", 
            RowBox[{"(", 
             RowBox[{
              SuperscriptBox[
               RowBox[{"x", "(", "t", ")"}], "2"], "+", 
              RowBox[{"2", 
               SuperscriptBox[
                RowBox[{"y", "(", "t", ")"}], "2"]}]}], ")"}], 
            RowBox[{"\[LeftBracketingBar]", 
             RowBox[{
              RowBox[{
               StyleBox["r",
                FontWeight->Bold,
                FontSlant->Plain], "'"}], 
              RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], " ", "d", 
            "\[VeryThinSpace]", "t"}]}],
          TraditionalForm]}], 
        StyleBox[
         RowBox[{
         "Write", " ", "the", " ", "line", " ", "integral", " ", "with", " ", 
          "respect", " ", "to", " ", 
          RowBox[{"t", "."}]}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         FormBox[
          RowBox[{
           SubsuperscriptBox["\[Integral]", 
            RowBox[{" ", "0"}], 
            RowBox[{" ", 
             SqrtBox[
              RowBox[{"2", "\[Pi]"}]]}]], 
           RowBox[{"100", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{
               SuperscriptBox["cos", "2"], 
               SuperscriptBox["t", "2"]}], "+", 
              RowBox[{"2", 
               SuperscriptBox["sin", "2"], 
               SuperscriptBox["t", "2"]}]}], ")"}], 
            FormBox[
             UnderscriptBox[
              RowBox[{"2", " ", "t"}], 
              StyleBox[
               UnderscriptBox["\[UnderBrace]", 
                RowBox[{"\[LeftBracketingBar]", 
                 RowBox[{
                  RowBox[{
                   StyleBox["r",
                    FontWeight->Bold,
                    FontSlant->Plain], "'"}], 
                  RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}]], 
               "TypesetAnnotationFont"]],
             TraditionalForm], " ", "d", "\[VeryThinSpace]", "t"}]}],
          TraditionalForm]}], 
        StyleBox[
         RowBox[{"Substitute", " ", "for", " ", "x", " ", "and", " ", 
          RowBox[{"y", "."}]}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{"100", 
          FormBox[
           FormBox[
            UnderscriptBox[
             UnderscriptBox[
              FormBox[
               RowBox[{
                SubsuperscriptBox["\[Integral]", 
                 RowBox[{" ", "0"}], 
                 RowBox[{" ", 
                  RowBox[{"2", "\[Pi]"}]}]], 
                RowBox[{
                 RowBox[{"(", 
                  RowBox[{
                   RowBox[{
                    SuperscriptBox["cos", "2"], "u"}], "+", 
                   RowBox[{"2", 
                    SuperscriptBox["sin", "2"], "u"}]}], ")"}], " ", "d", 
                 "\[VeryThinSpace]", "u"}]}],
               TraditionalForm], 
              StyleBox["\[UnderBrace]", "TypesetAnnotationFont"]], 
             StyleBox[
              RowBox[{"\[Pi]", "+", 
               RowBox[{"2", "\[Pi]"}]}], "TypesetAnnotationFont"]],
            TraditionalForm],
           TraditionalForm]}]}], 
        StyleBox[
         RowBox[{
          RowBox[{
           RowBox[{"Simplify", " ", "and", " ", "let", " ", "u"}], "=", 
           SuperscriptBox["t", "2"]}], ",", 
          RowBox[{
           RowBox[{"d", "\[VeryThinSpace]", "u"}], "=", 
           RowBox[{"2", "t", " ", "d", "\[VeryThinSpace]", 
            RowBox[{"t", ".", Cell[""]}]}]}]}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{"300", "\[Pi]"}]}], 
        StyleBox[
         RowBox[{"Evaluate", " ", "the", " ", 
          RowBox[{"integral", "."}]}], "TypesetAnnotationFont"]}
      },
      GridBoxAlignment->{"Columns" -> {{"=", Left}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479513336623509*^9, 3.4795133392400723`*^9}, {
  3.4795133772920732`*^9, 3.479513506468596*^9}, {3.479513537114565*^9, 
  3.479513687048731*^9}, {3.517750417192564*^9, 3.517750423644204*^9}}],

Cell[TextData[{
 "Dividing by the length of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", the average temperature on the boundary of the plate is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"300", 
     RowBox[{"\[Pi]", "/", 
      RowBox[{"(", 
       RowBox[{"2", "\[Pi]"}], ")"}]}]}], "=", "150"}], TraditionalForm]]],
 ", as found in Example 1."
}], "Text",
 CellChangeTimes->{{3.4795137606537313`*^9, 3.479513798132794*^9}}],

Cell[TextData[{
 "Related Exercises 15\[LongDash]24 ",
 Cell[BoxData[
  GraphicsBox[
   {RGBColor[0.796, 0, 0.42], EdgeForm[RGBColor[0.5306666666666667, 0, 0.28]],
     PolygonBox[{{0, 0}, {0.9, -0.3}, {0.7, 0}, {0.9, 0.3}, {0, 0}}]},
   ImageSize->15,
   PlotRange->{-0.5, 0.3}]],
  CellChangeTimes->{3.443312659351997*^9}]
}], "RelatedExercises",
 CellChangeTimes->{{3.4795138062060623`*^9, 3.479513808523369*^9}}]
}, Closed]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Line Integrals in ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[DoubleStruckCapitalR]", "3"], TraditionalForm]]]
}], "Subsection",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.4795138335518007`*^9, 3.479513840500227*^9}},
 CellTags->"Line Integrals in R3"],

Cell[TextData[{
 "The argument that leads to line integrals on plane curves extends \
immediately to three or more dimensions. Here is the corresponding evaluation \
theorem for line integrals in ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[DoubleStruckCapitalR]", "3"], TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.4795138442525177`*^9, 3.479513874716885*^9}, {
  3.48413154360314*^9, 3.484131552834229*^9}}],

Cell[TextData[{
 StyleBox["THEOREM 14.2", "TheoremFont"],
 "\t",
 StyleBox["Evaluating Scalar Line Integrals in ",
  FontWeight->"Bold"],
 Cell[BoxData[
  FormBox[
   StyleBox[
    SuperscriptBox["\[DoubleStruckCapitalR]", "3"],
    FontWeight->"Bold"], TraditionalForm]]],
 "\nLet ",
 Cell[BoxData[
  FormBox["f", TraditionalForm]]],
 " be continuous on a smooth curve ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ": ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     StyleBox["r",
      FontWeight->Bold,
      FontSlant->Plain], "(", "t", ")"}], "=", 
    RowBox[{"\[LeftAngleBracket]", 
     RowBox[{
      RowBox[{"x", "(", "t", ")"}], ",", 
      RowBox[{"y", "(", "t", ")"}], ",", 
      RowBox[{"z", "(", "t", ")"}]}], "\[RightAngleBracket]"}]}], 
   TraditionalForm]]],
 ", for ",
 Cell[BoxData[
  FormBox[
   RowBox[{"a", "\[LessEqual]", "t", "\[LessEqual]", "b"}], TraditionalForm]]],
 ". Then ",
 "\n",
 Cell[BoxData[
  FormBox[GridBox[{
     {GridBox[{
        {
         RowBox[{
          FormBox[
           RowBox[{
            UnderscriptBox["\[Integral]", "C"], 
            RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}],
           TraditionalForm], "=", 
          FormBox[
           RowBox[{
            SubsuperscriptBox["\[Integral]", 
             RowBox[{" ", "a"}], 
             RowBox[{" ", "b"}]], 
            RowBox[{
             RowBox[{"f", "(", 
              RowBox[{
               RowBox[{"x", "(", "t", ")"}], ",", 
               RowBox[{"y", "(", "t", ")"}], ",", 
               RowBox[{"z", "(", "t", ")"}]}], ")"}], 
             RowBox[{"\[LeftBracketingBar]", 
              RowBox[{
               RowBox[{
                StyleBox["r",
                 FontWeight->Bold,
                 FontSlant->Plain], "'"}], 
               RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], " ", "d",
              "\[VeryThinSpace]", "t"}]}],
           TraditionalForm]}]},
        {
         RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
          RowBox[{
           FormBox[
            RowBox[{
             SubsuperscriptBox["\[Integral]", 
              RowBox[{" ", "a"}], 
              RowBox[{" ", "b"}]], 
             RowBox[{
              RowBox[{"f", "(", 
               RowBox[{
                RowBox[{"x", "(", "t", ")"}], ",", 
                RowBox[{"y", "(", "t", ")"}], ",", 
                RowBox[{"z", "(", "t", ")"}]}], ")"}], 
              SqrtBox[
               RowBox[{
                RowBox[{
                 RowBox[{"x", "'"}], 
                 SuperscriptBox[
                  RowBox[{"(", "t", ")"}], "2"]}], "+", 
                RowBox[{
                 RowBox[{"y", "'"}], 
                 SuperscriptBox[
                  RowBox[{"(", "t", ")"}], "2"]}], "+", 
                RowBox[{
                 RowBox[{"z", "'"}], 
                 SuperscriptBox[
                  RowBox[{"(", "t", ")"}], "2"]}]}]], " ", "d", 
              "\[VeryThinSpace]", "t"}]}],
            TraditionalForm], "."}]}]}
       },
       GridBoxAlignment->{"Columns" -> {{"="}}}]}
    },
    GridBoxItemSize->{"Columns" -> {{
         Scaled[0.96]}}}], TraditionalForm]]]
}], "Theorem",
 CellChangeTimes->{{3.479513908599587*^9, 3.479513975999812*^9}, 
   3.479514089547267*^9, {3.484131566431756*^9, 3.484131567394678*^9}},
 CellTags->"THEOREM 14.2 Evaluating Scalar Line Integrals in R3"],

Cell[TextData[{
 "\tAs before, if ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " is the arc length ",
 Cell[BoxData[
  FormBox["s", TraditionalForm]]],
 ", then ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"\[LeftBracketingBar]", 
     RowBox[{
      RowBox[{
       StyleBox["r",
        FontWeight->Bold,
        FontSlant->Plain], "'"}], 
      RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "=", "1"}], 
   TraditionalForm]]],
 " and"
}], "Text",
 CellChangeTimes->{{3.479514120632792*^9, 3.4795141410513763`*^9}, {
  3.484131578163426*^9, 3.4841315784965343`*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         FormBox[
          RowBox[{
           UnderscriptBox["\[Integral]", "C"], 
           RowBox[{"f", " ", "d", "\[VeryThinSpace]", "s"}]}],
          TraditionalForm], "=", 
         RowBox[{
          FormBox[
           RowBox[{
            SubsuperscriptBox["\[Integral]", 
             RowBox[{" ", "a"}], 
             RowBox[{" ", "b"}]], 
            RowBox[{
             RowBox[{"f", "(", 
              RowBox[{
               RowBox[{"x", "(", "s", ")"}], ",", 
               RowBox[{"y", "(", "s", ")"}], ",", 
               RowBox[{"z", "(", "s", ")"}]}], ")"}], " ", "d", 
             "\[VeryThinSpace]", "s"}]}],
           TraditionalForm], "."}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.4795141456072083`*^9, 3.479514180056793*^9}}],

Cell[TextData[{
 "If ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"f", "(", 
     RowBox[{"x", ",", "y", ",", "z"}], ")"}], "=", "1"}], TraditionalForm]]],
 ", then the line integral gives the length of ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 "."
}], "Text",
 CellChangeTimes->{{3.479514194254436*^9, 3.479514207043784*^9}}],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["EXAMPLE 3\t", "ExampleFont"],
 "Line integrals in ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[DoubleStruckCapitalR]", "3"], TraditionalForm]]]
}], "Subsubsubsection",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.479514221419921*^9, 3.479514230029784*^9}},
 CellTags->"EXAMPLE 3 Line integrals in R3"],

Cell[TextData[{
 "Evaluate ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    UnderscriptBox["\[Integral]", "C"], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"x", " ", "y"}], "+", 
       RowBox[{"2", "z"}]}], ")"}], " ", "d", "\[VeryThinSpace]", "s"}]}], 
   TraditionalForm]]],
 " on the following lines."
}], "Text",
 CellChangeTimes->{{3.4795142349981823`*^9, 3.479514258339776*^9}}],

Cell[TextData[{
 StyleBox["a.",
  FontWeight->"Bold"],
 "\tThe line from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "(", 
    RowBox[{"1", ",", "0", ",", "0"}], ")"}], TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "(", 
    RowBox[{"0", ",", "1", ",", "1"}], ")"}], TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.479514262069805*^9, 3.4795142766747437`*^9}}],

Cell[TextData[{
 StyleBox["b.",
  FontWeight->"Bold"],
 "\tThe line from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "(", 
    RowBox[{"0", ",", "1", ",", "1"}], ")"}], TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "(", 
    RowBox[{"1", ",", "0", ",", "0"}], ")"}], TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.479514262069805*^9, 3.479514298837207*^9}}],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   PaneBox[
    StyleBox[
     GraphicsBox[
      {RGBColor[0.976, 0.855, 0.514], EdgeForm[RGBColor[0.68, 0.57, 0.34]], 
       PolygonBox[{{0.2, 0}, {0, -0.3}, {0.9, 0}, {0, 0.3}, {0.2, 0}}]},
      ImageSize->15,
      PlotRange->{-0.5, 0.3}], "Note",
     Editable->False,
     Selectable->False],
    Alignment->{Center, Top},
    ImageSize->{18, 13}],
   Background->GrayLevel[1],
   FrameMargins->{{5, 5}, {-1, -0.1}},
   FrameStyle->RGBColor[0.976, 0.855, 0.514],
   RoundingRadius->2]]],
 "  ",
 StyleBox["Note", "CalloutIconFont"]
}], "CalloutIcon",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "A parametric equation of a line is \n",
 Cell[BoxData[
  FormBox[GridBox[{
     {GridBox[{
        {
         RowBox[{" ", 
          RowBox[{
           FormBox[
            RowBox[{
             RowBox[{
              StyleBox["r",
               FontWeight->Bold,
               FontSlant->Plain], "(", "t", ")"}], "=", 
             RowBox[{
              RowBox[{"\[LeftAngleBracket]", 
               RowBox[{
                SubscriptBox["x", "0"], ",", 
                SubscriptBox["y", "0"], ",", 
                SubscriptBox["z", "0"]}], "\[RightAngleBracket]"}], "+", 
              RowBox[{"t", 
               RowBox[{"\[LeftAngleBracket]", 
                RowBox[{"a", ",", "b", ",", "c"}], 
                "\[RightAngleBracket]"}]}]}]}],
            TraditionalForm], ","}]}]}
       },
       GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
    },
    GridBoxItemSize->{"Columns" -> {{
         Scaled[0.96]}}}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{
     SubscriptBox["x", "0"], ",", 
     SubscriptBox["y", "0"], ",", 
     SubscriptBox["z", "0"]}], "\[RightAngleBracket]"}], TraditionalForm]]],
 " is a position vector associated with a fixed point on the line and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"\[LeftAngleBracket]", 
    RowBox[{"a", ",", "b", ",", "c"}], "\[RightAngleBracket]"}], 
   TraditionalForm]]],
 " is a vector parallel to the line."
}], "Callout",
 CellChangeTimes->{{3.479514366701434*^9, 3.4795144289809647`*^9}, {
  3.479514786327159*^9, 3.479514804285552*^9}}]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["SOLUTION", "SolutionFont"]], "Subsubsubsubsection",
 WholeCellGroupOpener->True],

Cell[TextData[{
 StyleBox["a.",
  FontWeight->"Bold"],
 "\tA parametric description of the line from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "(", 
    RowBox[{"1", ",", "0", ",", "0"}], ")"}], TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "(", 
    RowBox[{"0", ",", "1", ",", "1"}], ")"}], TraditionalForm]]],
 " is"
}], "Text",
 CellChangeTimes->{{3.479514481516829*^9, 3.4795145968913107`*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{
          RowBox[{
           StyleBox["r",
            FontWeight->Bold,
            FontSlant->Plain], 
           RowBox[{"(", "t", ")"}]}], "=", 
          RowBox[{
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"1", ",", "0", ",", "0"}], "\[RightAngleBracket]"}], "+", 
            RowBox[{"t", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{
               RowBox[{"-", "1"}], ",", "1", ",", "1"}], 
              "\[RightAngleBracket]"}]}]}], "=", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{"1", "-", "t"}], ",", "t", ",", "t"}], 
            "\[RightAngleBracket]"}]}]}], ",", "    ", 
         RowBox[{
          RowBox[{"for", " ", "0"}], "\[LessEqual]", "t", "\[LessEqual]", 
          "1."}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.4795145898863077`*^9, 3.4795145946303864`*^9}}],

Cell["The speed factor is", "Text",
 CellChangeTimes->{{3.4795146055046787`*^9, 3.479514608165422*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", 
          RowBox[{
           RowBox[{
            StyleBox["r",
             FontWeight->Bold,
             FontSlant->Plain], "'"}], 
           RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "=", 
         RowBox[{
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"x", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"y", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"z", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}]}]], "=", 
          RowBox[{
           SqrtBox[
            RowBox[{
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"-", "1"}], ")"}], "2"], "+", 
             SuperscriptBox["1", "2"], "+", 
             SuperscriptBox["1", "2"]}]], "=", 
           RowBox[{
            SqrtBox["3"], "."}]}]}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479514613830915*^9, 3.4795146214254017`*^9}, {
  3.479514660755823*^9, 3.479514698049995*^9}}],

Cell[TextData[{
 "Substituting ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", 
    RowBox[{"1", "-", "t"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", "=", "t"}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", "=", "t"}], TraditionalForm]]],
 ", the value of the line integral is"
}], "Text",
 CellChangeTimes->{{3.479514871103249*^9, 3.47951489026593*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        FormBox[
         RowBox[{
          RowBox[{
           UnderscriptBox["\[Integral]", "C"], 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"x", " ", "y"}], "+", 
              RowBox[{"2", "z"}]}], ")"}], " ", "d", "\[VeryThinSpace]", 
            "s"}]}], "=", 
          FormBox[
           RowBox[{
            SubsuperscriptBox["\[Integral]", 
             RowBox[{" ", "0"}], 
             RowBox[{" ", "1"}]], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{
                FormBox[
                 UnderscriptBox[
                  RowBox[{"(", 
                   RowBox[{"1", "-", "t"}], ")"}], 
                  StyleBox[
                   UnderscriptBox["\[UnderBrace]", "x"], 
                   "TypesetAnnotationFont"]],
                 TraditionalForm], 
                FormBox[
                 UnderscriptBox[
                  RowBox[{"(", "t", ")"}], 
                  StyleBox[
                   UnderscriptBox["\[UnderBrace]", "y"], 
                   "TypesetAnnotationFont"]],
                 TraditionalForm]}], "+", 
               RowBox[{"2", 
                FormBox[
                 UnderscriptBox[
                  RowBox[{"(", "t", ")"}], 
                  StyleBox[
                   UnderscriptBox["\[UnderBrace]", "z"], 
                   "TypesetAnnotationFont"]],
                 TraditionalForm]}]}], ")"}], 
             SqrtBox["3"], " ", "d", "\[VeryThinSpace]", "t"}]}],
           TraditionalForm]}],
         TraditionalForm], 
        StyleBox[
         RowBox[{
          RowBox[{"Substitute", " ", "for", " ", "x"}], ",", "y", ",", 
          RowBox[{"z", "."}]}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{
          SqrtBox["3"], 
          FormBox[
           RowBox[{
            SubsuperscriptBox["\[Integral]", 
             RowBox[{" ", "0"}], 
             RowBox[{" ", "1"}]], 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"3", "t"}], "-", 
               SuperscriptBox["t", "2"]}], ")"}], " ", "d", 
             "\[VeryThinSpace]", "t"}]}],
           TraditionalForm]}]}], 
        StyleBox[
         RowBox[{"Simplify", "."}], "TypesetAnnotationFont"]},
       {
        SubsuperscriptBox[
         RowBox[{
          RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
           RowBox[{
            SqrtBox["3"], 
            RowBox[{"(", 
             RowBox[{
              FractionBox[
               RowBox[{"3", 
                SuperscriptBox["t", "2"]}], "2"], "-", 
              FractionBox[
               SuperscriptBox["t", "3"], "3"]}], ")"}]}]}], 
          "\[RightBracketingBar]"}], "0", "1"], 
        StyleBox[
         RowBox[{"Integrate", "."}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{
          FractionBox[
           RowBox[{"7", 
            SqrtBox["3"]}], "6"], "."}]}], 
        StyleBox[
         RowBox[{"Evaluate", "."}], "TypesetAnnotationFont"]}
      },
      GridBoxAlignment->{"Columns" -> {{"=", Left}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479514894732431*^9, 3.479515019476945*^9}, {
  3.47951506614743*^9, 3.479515092602057*^9}}],

Cell[TextData[{
 StyleBox["b.",
  FontWeight->"Bold"],
 "\tThe line from ",
 Cell[BoxData[
  FormBox[
   RowBox[{"Q", "(", 
    RowBox[{"0", ",", "1", ",", "1"}], ")"}], TraditionalForm]]],
 " to ",
 Cell[BoxData[
  FormBox[
   RowBox[{"P", "(", 
    RowBox[{"1", ",", "0", ",", "0"}], ")"}], TraditionalForm]]],
 " may be described parametrically by"
}], "Text",
 CellChangeTimes->{{3.479515121857172*^9, 3.479515150515635*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{
          RowBox[{
           StyleBox["r",
            FontWeight->Bold,
            FontSlant->Plain], 
           RowBox[{"(", "t", ")"}]}], "=", 
          RowBox[{
           RowBox[{
            RowBox[{"\[LeftAngleBracket]", 
             RowBox[{"0", ",", "1", ",", "1"}], "\[RightAngleBracket]"}], "+", 
            RowBox[{"t", 
             RowBox[{"\[LeftAngleBracket]", 
              RowBox[{"1", ",", 
               RowBox[{"-", "1"}], ",", 
               RowBox[{"-", "1"}]}], "\[RightAngleBracket]"}]}]}], "=", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{"t", ",", 
             RowBox[{"1", "-", "t"}], ",", 
             RowBox[{"1", "-", "t"}]}], "\[RightAngleBracket]"}]}]}], ",", 
         "    ", 
         RowBox[{
          RowBox[{"for", " ", "0"}], "\[LessEqual]", "t", "\[LessEqual]", 
          "1."}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.4795145898863077`*^9, 3.4795145946303864`*^9}, {
  3.479515161842922*^9, 3.4795151795593843`*^9}}],

Cell["The speed factor is", "Text",
 CellChangeTimes->{{3.479515185147195*^9, 3.479515187485271*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", 
          RowBox[{
           RowBox[{
            StyleBox["r",
             FontWeight->Bold,
             FontSlant->Plain], "'"}], 
           RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "=", 
         RowBox[{
          SqrtBox[
           RowBox[{
            RowBox[{
             RowBox[{"x", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"y", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"z", "'"}], 
             SuperscriptBox[
              RowBox[{"(", "t", ")"}], "2"]}]}]], "=", 
          RowBox[{
           SqrtBox[
            RowBox[{
             SuperscriptBox["1", "2"], "+", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"-", "1"}], ")"}], "2"], "+", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{"-", "1"}], ")"}], "2"]}]], "=", 
           RowBox[{
            SqrtBox["3"], "."}]}]}]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479514613830915*^9, 3.4795146214254017`*^9}, {
  3.479514660755823*^9, 3.479514698049995*^9}, {3.479515199356248*^9, 
  3.4795152085226707`*^9}}],

Cell[TextData[{
 "We substitute ",
 Cell[BoxData[
  FormBox[
   RowBox[{"x", "=", "t"}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{"y", "=", 
    RowBox[{"1", "-", "t"}]}], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   RowBox[{"z", "=", 
    RowBox[{"1", "-", "t"}]}], TraditionalForm]]],
 " and do a calculation similar to that in part (a). The value of the line \
integral is again ",
 Cell[BoxData[
  FormBox[
   FractionBox[
    RowBox[{"7", 
     SqrtBox["3"]}], "6"], TraditionalForm]]],
 ", emphasizing the fact that a scalar line integral is independent of the \
orientation and parameterization of the curve."
}], "Text",
 CellChangeTimes->{{3.479515211870021*^9, 3.479515271145701*^9}, {
  3.484131604225374*^9, 3.484131605007647*^9}}],

Cell[TextData[{
 "Related Exercises 25\[LongDash]30 ",
 Cell[BoxData[
  GraphicsBox[
   {RGBColor[0.796, 0, 0.42], EdgeForm[RGBColor[0.5306666666666667, 0, 0.28]],
     PolygonBox[{{0, 0}, {0.9, -0.3}, {0.7, 0}, {0.9, 0.3}, {0, 0}}]},
   ImageSize->15,
   PlotRange->{-0.5, 0.3}]],
  CellChangeTimes->{3.443312659351997*^9}]
}], "RelatedExercises",
 CellChangeTimes->{{3.4795152763476677`*^9, 3.4795152784167423`*^9}}]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["EXAMPLE 4\t", "ExampleFont"],
 "Flight of an eagle"
}], "Subsubsubsection",
 WholeCellGroupOpener->True,
 CellChangeTimes->{{3.4795153298172407`*^9, 3.479515334573287*^9}},
 CellTags->"EXAMPLE 4 Flight of an eagle"],

Cell["An eagle soars on the ascending spiral path", "Text",
 CellChangeTimes->{{3.479515336108615*^9, 3.479515345066699*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{
          RowBox[{"C", ":", 
           RowBox[{
            StyleBox["r",
             FontWeight->Bold,
             FontSlant->Plain], "(", "t", ")"}]}], "=", 
          RowBox[{
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{"x", "(", "t", ")"}], ",", 
             RowBox[{"y", "(", "t", ")"}], ",", 
             RowBox[{"z", "(", "t", ")"}]}], "\[RightAngleBracket]"}], "=", 
           RowBox[{"\[LeftAngleBracket]", 
            RowBox[{
             RowBox[{"2400", "cos", " ", 
              FractionBox["t", "2"]}], ",", 
             RowBox[{"2400", "sin", " ", 
              FractionBox["t", "2"]}], ",", 
             RowBox[{"500", "t"}]}], "\[RightAngleBracket]"}]}]}], ","}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.47951538549752*^9, 3.479515415199027*^9}}],

Cell[TextData[{
 "where ",
 Cell[BoxData[
  FormBox["x", TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox["y", TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox["z", TraditionalForm]]],
 " are measured in feet and ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " is measured in minutes. How far does the eagle fly over the time interval ",
 Cell[BoxData[
  FormBox[
   RowBox[{"0", "\[LessEqual]", "t", "\[LessEqual]", "10"}], 
   TraditionalForm]]],
 "?"
}], "Text",
 CellChangeTimes->{{3.479515417701277*^9, 3.4795154452809353`*^9}}],

Cell[CellGroupData[{

Cell[TextData[StyleBox["SOLUTION", "SolutionFont"]], "Subsubsubsubsection",
 WholeCellGroupOpener->True],

Cell[CellGroupData[{

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   PaneBox[
    StyleBox[
     GraphicsBox[
      {RGBColor[0.976, 0.855, 0.514], EdgeForm[RGBColor[0.68, 0.57, 0.34]], 
       PolygonBox[{{0.2, 0}, {0, -0.3}, {0.9, 0}, {0, 0.3}, {0.2, 0}}]},
      ImageSize->15,
      PlotRange->{-0.5, 0.3}], "Note",
     Editable->False,
     Selectable->False],
    Alignment->{Center, Top},
    ImageSize->{18, 13}],
   Background->GrayLevel[1],
   FrameMargins->{{5, 5}, {-1, -0.1}},
   FrameStyle->RGBColor[0.976, 0.855, 0.514],
   RoundingRadius->2]]],
 "  ",
 StyleBox["Note", "CalloutIconFont"]
}], "CalloutIcon",
 WholeCellGroupOpener->True],

Cell[TextData[{
 "Because we are finding the length of a curve, the integrand in this line \
integral is ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{"f", "(", 
     RowBox[{"x", ",", "y", ",", "z"}], ")"}], "=", "1"}], TraditionalForm]]],
 "."
}], "Callout",
 CellChangeTimes->{{3.479515723390368*^9, 3.479515748345335*^9}}]
}, Closed]],

Cell[TextData[{
 "The distance traveled is found by integrating the element of arc length ",
 Cell[BoxData[
  FormBox[
   RowBox[{"d", "\[VeryThinSpace]", "s"}], TraditionalForm]]],
 " along ",
 Cell[BoxData[
  FormBox["C", TraditionalForm]]],
 ", that is, ",
 Cell[BoxData[
  FormBox[
   RowBox[{"L", "=", 
    RowBox[{
     UnderscriptBox[
      RowBox[{"\[Integral]", " "}], "C"], "d", "\[VeryThinSpace]", "s"}]}], 
   TraditionalForm]]],
 ". ",
 "We now make a change of variables to the parameter ",
 Cell[BoxData[
  FormBox["t", TraditionalForm]]],
 " using"
}], "Text",
 CellChangeTimes->{{3.479515472857222*^9, 3.4795155279178762`*^9}, 
   3.479516036682613*^9, 3.4795161204171133`*^9}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"\[LeftBracketingBar]", 
          RowBox[{
           RowBox[{
            StyleBox["r",
             FontWeight->Bold,
             FontSlant->Plain], "'"}], 
           RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], "=", 
         SqrtBox[
          RowBox[{
           RowBox[{
            RowBox[{"x", "'"}], 
            SuperscriptBox[
             RowBox[{"(", "t", ")"}], "2"]}], "+", 
           RowBox[{
            RowBox[{"y", "'"}], 
            SuperscriptBox[
             RowBox[{"(", "t", ")"}], "2"]}], "+", 
           RowBox[{
            RowBox[{"z", "'"}], 
            SuperscriptBox[
             RowBox[{"(", "t", ")"}], "2"]}]}]]}], " "},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         SqrtBox[
          RowBox[{
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{
              RowBox[{"-", "1200"}], "sin", " ", 
              FractionBox["t", "2"]}], ")"}], "2"], "+", 
           SuperscriptBox[
            RowBox[{"(", 
             RowBox[{"1200", "cos", " ", 
              FractionBox["t", "2"]}], ")"}], "2"], "+", 
           SuperscriptBox["500", "2"]}]]}], 
        StyleBox[
         RowBox[{"Substitute", " ", 
          RowBox[{"derivatives", "."}]}], "TypesetAnnotationFont"]},
       {
        RowBox[{"\[InvisiblePostfixScriptBase]", "=", 
         RowBox[{
          SqrtBox[
           RowBox[{
            SuperscriptBox["1200", "2"], "+", 
            SuperscriptBox["500", "2"]}]], "=", "1300"}]}], 
        StyleBox[
         RowBox[{
          RowBox[{
           RowBox[{
            SuperscriptBox["sin", "2"], 
            FractionBox["t", "2"]}], "+", 
           RowBox[{
            SuperscriptBox["cos", "2"], 
            FractionBox["t", "2"]}]}], "=", "1"}], "TypesetAnnotationFont"]}
      },
      GridBoxAlignment->{"Columns" -> {{"=", Left}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479515533565786*^9, 3.479515623198*^9}}],

Cell["It follows that the distance traveled is", "Text",
 CellChangeTimes->{{3.479515783327368*^9, 3.479515790169135*^9}}],

Cell[TextData[Cell[BoxData[
 FormBox[GridBox[{
    {GridBox[{
       {
        RowBox[{"L", "=", 
         FormBox[
          RowBox[{
           RowBox[{
            UnderscriptBox["\[Integral]", "C"], " ", 
            RowBox[{"d", "\[VeryThinSpace]", "s"}]}], "=", 
           FormBox[
            RowBox[{
             RowBox[{
              SubsuperscriptBox["\[Integral]", 
               RowBox[{" ", "0"}], 
               RowBox[{" ", "10"}]], 
              RowBox[{
               RowBox[{"\[LeftBracketingBar]", 
                RowBox[{
                 RowBox[{
                  StyleBox["r",
                   FontWeight->Bold,
                   FontSlant->Plain], "'"}], 
                 RowBox[{"(", "t", ")"}]}], "\[RightBracketingBar]"}], " ", 
               "d", "\[VeryThinSpace]", "t"}]}], "=", 
             RowBox[{
              FormBox[
               RowBox[{
                SubsuperscriptBox["\[Integral]", 
                 RowBox[{" ", "0"}], 
                 RowBox[{" ", "10"}]], 
                RowBox[{"1300", " ", "d", "\[VeryThinSpace]", "t"}]}],
               TraditionalForm], "=", 
              RowBox[{"13000", " ", 
               RowBox[{"ft", "."}]}]}]}],
            TraditionalForm]}],
          TraditionalForm]}]}
      },
      GridBoxAlignment->{"Columns" -> {{"Center"}}}]}
   },
   GridBoxItemSize->{"Columns" -> {{
        Scaled[0.96]}}}], TraditionalForm]]]], "Text",
 CellChangeTimes->{{3.479515795786582*^9, 3.479515839973564*^9}}],

Cell[TextData[{
 "Related Exercises 31\[LongDash]32 ",
 Cell[BoxData[
  GraphicsBox[
   {RGBColor[0.796, 0, 0.42], EdgeForm[RGBColor[0.5306666666666667, 0, 0.28]],
     PolygonBox[{{0, 0}, {0.9, -0.3}, {0.7, 0}, {0.9, 0.3}, {0, 0}}]},
   ImageSize->15,
   PlotRange->{-0.5, 0.3}]],
  CellChangeTimes->{3.443312659351997*^9}]
}], "RelatedExercises",
 CellChangeTimes->{{3.479515845850905*^9, 3.47951584736758*^9}}]
}, Closed]]
}, Open  ]],

Cell[TextData[{
 Cell[BoxData[
  FrameBox[
   StyleBox[
    RowBox[{"QUICK", " ", "CHECK", " ", "2"}], "QuickCheckFont"],
   Alignment->Center,
   Background->RGBColor[0.976, 0.855, 0.514],
   FrameMargins->{{5, 5}, {2, 4}},
   FrameStyle->GrayLevel[0.5]]]],
 "  What is the speed of the eagle in Example 4?"
}], "QuickCheck",
 CellChangeTimes->{{3.479515876962641*^9, 3.479515885482938*^9}},
 CellTags->"QUICK CHECK 2"],

Cell[CellGroupData[{

Cell[TextData[Cell[BoxData[
 FrameBox[
  StyleBox["\[Checkmark]", "SR",
   FontSize->13,
   FontWeight->"Bold"],
  Alignment->Center,
  Background->RGBColor[0.976, 0.855, 0.514],
  FrameMargins->{{4, 6}, {1, 5}},
  FrameStyle->GrayLevel[0.5],
  RoundingRadius->5]]]], "QuickCheckAnswerIcon",
 WholeCellGroupOpener->True],

Cell[TextData[Cell[BoxData[
 FormBox[
  RowBox[{"1300", " ", 
   RowBox[{"ft", "/", "min"}]}], TraditionalForm]]]], "QuickCheckAnswer",
 CellChangeTimes->{3.521840628616984*^9}]
}, Closed]]
}, Open  ]]
}, Open  ]]
},
WindowSize->{707, 867},
WindowMargins->{{Automatic, 52}, {Automatic, 69}},
FrontEndVersion->"8.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (November 6, \
2010)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[StyleDefinitions -> "Default.nb"]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Section"], CellFrame -> {{0, 0}, {2, 0}}, ShowGroupOpener -> 
       False, CellMargins -> {{10, 0}, {3, 10}}, CellFrameColor -> 
       GrayLevel[0.5], FontFamily -> "Times", FontSize -> 20, FontWeight -> 
       "Bold", FontSlant -> "Plain", FontTracking -> "Wide"], 
      Cell[
       StyleData["Section", "Printout"], CellFrame -> {{0, 0}, {2, 0}}, 
       ShowGroupOpener -> False, CellMargins -> {{0, Inherited}, {3, 10}}, 
       CellFrameColor -> GrayLevel[0.5], FontFamily -> "Times", FontSize -> 
       20, FontWeight -> "Bold", FontSlant -> "Plain", FontTracking -> 
       "Wide"]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Subsection"], CellFrame -> {{3, 0}, {0, 0.5}}, CellDingbat -> 
       None, ShowGroupOpener -> True, CellMargins -> {{15, 0}, {2, 5}}, 
       CellFrameColor -> GrayLevel[0.85], FontFamily -> "Times", FontSize -> 
       18, FontWeight -> "Bold", FontSlant -> "Plain"], 
      Cell[
       StyleData["Subsection", "Printout"], CellFrame -> {{3, 0}, {0, 0.5}}, 
       CellDingbat -> None, ShowGroupOpener -> True, 
       CellMargins -> {{0, 0}, {2, 10}}, CellFrameColor -> GrayLevel[0.85], 
       FontFamily -> "Times", FontSize -> 18, FontWeight -> "Bold", FontSlant -> 
       "Plain"]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Subsubsection"], CellDingbat -> None, ShowGroupOpener -> 
       True, CellMargins -> {{30, Inherited}, {2, 10}}, FontFamily -> 
       "Helvetica", FontSize -> 14, FontWeight -> "Bold"], 
      Cell[
       StyleData["Subsubsection", "Printout"], CellDingbat -> None, 
       ShowGroupOpener -> True, CellMargins -> {{0, Inherited}, {2, 10}}, 
       FontSize -> 14]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Subsubsubsection"], CellDingbat -> None, ShowGroupOpener -> 
       True, CellMargins -> {{30, Inherited}, {2, 10}}, FontFamily -> "Times",
        FontSize -> 14, FontWeight -> "Bold", FontSlant -> "Plain"], 
      Cell[
       StyleData["Subsubsubsection", "Printout"], CellDingbat -> None, 
       ShowGroupOpener -> True, CellMargins -> {{0, Inherited}, {2, 10}}, 
       FontFamily -> "Times", FontSize -> 14, FontSlant -> "Plain"]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Subsubsubsubsection"], CellDingbat -> None, ShowGroupOpener -> 
       True, CellMargins -> {{30, Inherited}, {2, 10}}, FontSize -> 14], 
      Cell[
       StyleData["Subsubsubsubsection", "Printout"], CellDingbat -> None, 
       ShowGroupOpener -> True, CellMargins -> {{0, Inherited}, {2, 10}}, 
       FontSize -> 14]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["FirstCell"], ShowCellBracket -> False, 
       CellMargins -> {{30, 10}, {7, 7}}, CellOpen -> False, TabSpacings -> 3,
        FontSize -> 14], 
      Cell[
       StyleData["FirstCell", "Printout"], CellMargins -> {{30, 10}, {7, 7}}, 
       CellOpen -> False, FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Text"], CellMargins -> {{30, 10}, {7, 7}}, TabSpacings -> 3,
        FontSize -> 14], 
      Cell[
       StyleData["Text", "Printout"], CellMargins -> {{0, 0}, {7, 7}}, 
       TabSpacings -> 3, FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Item"], CellDingbat -> 
       Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
       CellMargins -> {{50, 10}, {4, 4}}, 
       CellGroupingRules -> {"GroupTogetherNestedGrouping", 15000}, 
       CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", TabSpacings -> 
       3, CounterIncrements -> "Item", 
       CounterAssignments -> {{"Subitem", 0}, {"Subsubitem", 0}}, FontSize -> 
       14, Notebook$$20`MenuPosition -> 
       1600, $CellContext`ReturnCreatesNewCell -> True], 
      Cell[
       StyleData["Item", "Printout"], CellDingbat -> 
       Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
       CellMargins -> {{20, 0}, {2, 2}}, CellFrameLabelMargins -> 4, 
       TabSpacings -> 3, FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Subitem"], CellDingbat -> 
       Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
       CellMargins -> {{70, 12}, {4, 4}}, 
       CellGroupingRules -> {"GroupTogetherNestedGrouping", 15100}, 
       CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Item", TabSpacings -> 
       3, CounterIncrements -> "Subitem", 
       CounterAssignments -> {{"Subsubitem", 0}}, FontSize -> 
       14, $CellContext`MenuPosition -> 
       1610, $CellContext`ReturnCreatesNewCell -> True], 
      Cell[
       StyleData["Subitem", "Printout"], CellDingbat -> 
       Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
       CellMargins -> {{40, 2}, {1, 1}}, CellFrameLabelMargins -> 4, 
       TabSpacings -> 3, FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Subsubitem"], CellDingbat -> 
       Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
       CellMargins -> {{90, 12}, {4, 4}}, 
       CellGroupingRules -> {"GroupTogetherNestedGrouping", 15200}, 
       CellFrameLabelMargins -> 4, DefaultNewCellStyle -> "Subsubitem", 
       TabSpacings -> 3, CounterIncrements -> "Subsubitem", FontSize -> 
       14, $CellContext`ReturnCreatesNewCell -> True], 
      Cell[
       StyleData["Subsubitem", "Printout"], CellDingbat -> 
       Cell["\[FilledSmallCircle]", FontWeight -> "Bold"], 
       CellMargins -> {{60, 2}, {1, 1}}, CellFrameLabelMargins -> 4, 
       TabSpacings -> 3, FontSize -> 13]}, Open]], 
   Cell[
    StyleData["QuickCheckFont"], FontFamily -> "Helvetica", FontSize -> 11, 
    FontWeight -> "Bold", FontSlant -> "Plain", FontColor -> GrayLevel[0]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["QuickCheck"], CellMargins -> {{30, 10}, {0, 7}}, 
       CellGroupingRules -> {"GroupTogetherGrouping", 51}, 
       LineSpacing -> {1, 3}, TabSpacings -> 3, FontSize -> 14], 
      Cell[
       StyleData["QuickCheck", "Printout"], CellMargins -> {{0, 30}, {0, 7}}, 
       PageBreakWithin -> False, LineSpacing -> {1, 3}, TabSpacings -> 3, 
       FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["QuickCheckAnswerIcon"], CellMargins -> {{40, 108}, {0, 0}}, 
       CellGroupingRules -> {"GroupTogetherGrouping", 100000}, CellFrameColor -> 
       RGBColor[0.976, 0.855, 0.514], TextAlignment -> Left, FontSize -> 13], 
      Cell[
       StyleData["QuickCheckAnswerIcon", "Printout"], 
       CellMargins -> {{10, 288}, {0, 0}}, PageBreakAbove -> False, 
       PageBreakBelow -> False, CellFrameColor -> 
       RGBColor[0.976, 0.855, 0.514], FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["QuickCheckAnswer"], CellFrame -> {{3, 0}, {0, 0}}, 
       CellMargins -> {{60, 60}, {7, 0}}, 
       CellGroupingRules -> {"GroupTogetherGrouping", 100000}, CellFrameColor -> 
       RGBColor[0.976, 0.855, 0.514], LineSpacing -> {1, 0}, 
       ParagraphSpacing -> {0, 6}, FontSize -> 14], 
      Cell[
       StyleData["QuickCheckAnswer", "Printout"], 
       CellFrame -> {{3, 0}, {0, 0}}, CellMargins -> {{30, 60}, {7, 0}}, 
       PageBreakWithin -> False, CellFrameColor -> 
       RGBColor[0.976, 0.855, 0.514], LineSpacing -> {1, 0}, 
       ParagraphSpacing -> {0, 6}, FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["DefinitionFont"], FontFamily -> "Helvetica", FontSize -> 14,
        FontWeight -> "Bold", FontSlant -> "Plain", FontColor -> 
       RGBColor[0.6, 0.408, 0.122]], 
      Cell[
       StyleData["DefinitionFont", "Printout"], FontFamily -> "Helvetica", 
       FontSize -> 13, FontWeight -> "Bold", FontSlant -> "Plain", FontColor -> 
       RGBColor[0.6, 0.408, 0.122]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Definition"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, CellFrameColor -> 
       RGBColor[0.6, 0.408, 0.122], LineSpacing -> {1, 3}, 
       ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 14], 
      Cell[
       StyleData["Definition", "Printout"], CellFrame -> {{0.5, 3}, {3, 0.5}},
        CellMargins -> {{30, 30}, {7, 7}}, PageBreakWithin -> False, 
       CellFrameColor -> RGBColor[0.6, 0.408, 0.122], LineSpacing -> {1, 3}, 
       ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["TheoremFont"], FontFamily -> "Helvetica", FontSize -> 14, 
       FontWeight -> "Bold", FontSlant -> "Plain", FontColor -> 
       RGBColor[0, 0.451, 0.592]], 
      Cell[
       StyleData["TheoremFont", "Printout"], FontFamily -> "Helvetica", 
       FontSize -> 13, FontWeight -> "Bold", FontSlant -> "Plain", FontColor -> 
       RGBColor[0, 0.451, 0.592]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Theorem"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, CellFrameColor -> 
       RGBColor[0, 0.451, 0.592], LineSpacing -> {1, 3}, 
       ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 14], 
      Cell[
       StyleData["Theorem", "Printout"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, PageBreakWithin -> False, 
       CellFrameColor -> RGBColor[0, 0.451, 0.592], LineSpacing -> {1, 3}, 
       ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["ProofFont"], FontSize -> 14, FontWeight -> "Bold", 
       FontColor -> RGBColor[0., 0.576, 0.816]], 
      Cell[
       StyleData["ProofFont", "Printout"], FontSize -> 13, FontWeight -> 
       "Bold", FontColor -> RGBColor[0., 0.576, 0.816]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["SummaryFont"], FontFamily -> "Helvetica", FontSize -> 14, 
       FontWeight -> "Bold", FontSlant -> "Plain", FontColor -> 
       RGBColor[0.204, 0.114, 0.396]], 
      Cell[
       StyleData["SummaryFont", "Printout"], FontFamily -> "Helvetica", 
       FontSize -> 13, FontWeight -> "Bold", FontSlant -> "Plain", FontColor -> 
       RGBColor[0.204, 0.114, 0.396]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Summary"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, CellFrameColor -> 
       RGBColor[0.204, 0.114, 0.396], LineSpacing -> {1, 3}, 
       ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 14], 
      Cell[
       StyleData["Summary", "Printout"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, PageBreakWithin -> False, 
       CellFrameColor -> RGBColor[0.204, 0.114, 0.396], LineSpacing -> {1, 3},
        ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 13]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["ProcedureFont"], FontFamily -> "Helvetica", FontSize -> 14, 
       FontWeight -> "Bold", FontSlant -> "Plain", FontTracking -> 
       "Condensed", FontColor -> RGBColor[0.204, 0.114, 0.396]], 
      Cell[
       StyleData["ProcedureFont", "Printout"], FontFamily -> "Helvetica", 
       FontSize -> 13, FontWeight -> "Bold", FontSlant -> "Plain", 
       FontTracking -> "Condensed", FontColor -> 
       RGBColor[0.204, 0.114, 0.396]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Procedure"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, CellFrameColor -> 
       RGBColor[0.204, 0.114, 0.396], LineSpacing -> {1, 3}, 
       ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 14], 
      Cell[
       StyleData["Procedure", "Printout"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, PageBreakWithin -> False, 
       CellFrameColor -> RGBColor[0.204, 0.114, 0.396], LineSpacing -> {1, 3},
        ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 13]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Important"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, CellFrameColor -> 
       RGBColor[0.204, 0.114, 0.396], LineSpacing -> {1, 3}, 
       ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 14], 
      Cell[
       StyleData["Important", "Printout"], CellFrame -> {{0.5, 3}, {3, 0.5}}, 
       CellMargins -> {{30, 30}, {7, 7}}, PageBreakWithin -> False, 
       CellFrameColor -> RGBColor[0.204, 0.114, 0.396], LineSpacing -> {1, 3},
        ParagraphSpacing -> {0, 6}, TabSpacings -> 3, FontSize -> 13]}, 
     Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["RelatedExercises"], TextAlignment -> Right, FontSize -> 14, 
       FontSlant -> "Italic", FontColor -> RGBColor[0.796, 0, 0.42]], 
      Cell[
       StyleData["RelatedExercises", "Printout"], TextAlignment -> Right, 
       FontSize -> 13, FontSlant -> "Italic", FontColor -> 
       RGBColor[0.796, 0, 0.42]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["FigureCaption"], CellMargins -> {{30, 30}, {7, 7}}, 
       LineSpacing -> {1, 3}, FontSize -> 14], 
      Cell[
       StyleData["FigureCaption", "Printout"], 
       CellMargins -> {{30, 30}, {7, 7}}, LineSpacing -> {1, 3}, FontSize -> 
       13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["TableFont"], FontFamily -> "Helvetica", FontSize -> 13, 
       FontWeight -> "Bold", FontColor -> RGBColor[0, 0.576, 0.816]], 
      Cell[
       StyleData["TableFont", "Printout"], FontFamily -> "Helvetica", 
       FontSize -> 13, FontWeight -> "Bold", FontColor -> 
       RGBColor[0, 0.576, 0.816]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["TableCell"], CellMargins -> {{30, 10}, {5, 5}}], 
      Cell[
       StyleData["TableCell", "Printout"], CellMargins -> {{0, 0}, {5, 5}}]}, 
     Open]], 
   Cell[
    StyleData["Note"], FontFamily -> "Verdana", FontSize -> 10, FontColor -> 
    RGBColor[0.6, 0.4, 0.4]], 
   Cell[
    StyleData["SectionTitleFont"], FontWeight -> "Bold", FontTracking -> 
    "Wide", FontColor -> RGBColor[0.8, 0, 0.4]], 
   Cell[
    StyleData["TypesetAnnotationFont"], FontColor -> 
    RGBColor[0, 0.576, 0.816]], 
   Cell[
    StyleData["ExampleFont"], FontFamily -> "Helvetica", FontColor -> 
    RGBColor[0, 0.576, 0.816]], 
   Cell[
    StyleData["SolutionFont"], FontFamily -> "Helvetica", FontColor -> 
    RGBColor[0.796, 0, 0.42]], 
   Cell[
    StyleData["ExerciseFont"], FontFamily -> "Helvetica", FontColor -> 
    RGBColor[0.8, 0, 0.4]], 
   Cell[
    StyleData["ExerciseGroupFont"], FontFamily -> "Times", FontColor -> 
    RGBColor[0, 0.576, 0.816]], 
   Cell[
    StyleData["FigureFont"], FontFamily -> "Helvetica", FontSize -> 13, 
    FontWeight -> "Bold", FontColor -> RGBColor[0., 0.576, 0.816]], 
   Cell[
    StyleData["FigureFontText"], FontColor -> RGBColor[0., 0.576, 0.816]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["ExerciseDirectionsCell"], CellMargins -> {{30, 10}, {7, 7}},
        LineSpacing -> {1, 3}, TabSpacings -> 2, FontFamily -> "Times", 
       FontSize -> 14, FontWeight -> "Plain", FontSlant -> "Italic"], 
      Cell[
       StyleData["ExerciseDirectionsCell", "Printout"], 
       CellMargins -> {{30, 10}, {7, 7}}, LineSpacing -> {1, 3}, TabSpacings -> 
       2, FontFamily -> "Times", FontSize -> 13, FontWeight -> "Plain", 
       FontSlant -> "Italic"]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["TExerciseDirectionsCell"], 
       CellMargins -> {{10, 10}, {7, 7}}, LineSpacing -> {1, 3}, 
       ParagraphIndent -> -15, TabSpacings -> 2, FontFamily -> "Times", 
       FontSize -> 14, FontWeight -> "Plain", FontSlant -> "Italic"], 
      Cell[
       StyleData["TExerciseDirectionsCell", "Printout"], 
       CellMargins -> {{10, 10}, {7, 7}}, LineSpacing -> {1, 3}, 
       ParagraphIndent -> -15, TabSpacings -> 2, FontFamily -> "Times", 
       FontSize -> 13, FontWeight -> "Plain", FontSlant -> "Italic"]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["SubExerciseDirectionsCell"], 
       CellMargins -> {{58, 10}, {2, 2}}, LineSpacing -> {1, 3}, 
       ParagraphIndent -> -22, TabSpacings -> 2, FontFamily -> "Times", 
       FontSize -> 14, FontWeight -> "Plain", FontSlant -> "Italic"], 
      Cell[
       StyleData["SubExerciseDirectionsCell", "Printout"], 
       CellMargins -> {{58, 10}, {2, 2}}, LineSpacing -> {1, 3}, 
       ParagraphIndent -> -16, TabSpacings -> 2, FontFamily -> "Times", 
       FontSize -> 13, FontWeight -> "Plain", FontSlant -> "Italic"]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Problem"], CellMargins -> {{30, 10}, {7, 7}}, 
       LineSpacing -> {1, 3}, ParagraphIndent -> -30, TabSpacings -> 2.5, 
       FontSize -> 14], 
      Cell[
       StyleData["Problem", "Printout"], CellMargins -> {{30, 10}, {7, 7}}, 
       LineSpacing -> {1, 3}, ParagraphIndent -> -21, TabSpacings -> 2.5, 
       FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["TProblem"], CellMargins -> {{10, 10}, {7, 7}}, 
       LineSpacing -> {1, 3}, ParagraphIndent -> -47, TabSpacings -> 2, 
       FontSize -> 14], 
      Cell[
       StyleData["TProblem", "Printout"], CellMargins -> {{10, 10}, {7, 7}}, 
       LineSpacing -> {1, 3}, ParagraphIndent -> -36, TabSpacings -> 2, 
       FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["SubProblem"], CellMargins -> {{58, 10}, {2, 2}}, 
       LineSpacing -> {1, 3}, ParagraphIndent -> -22, TabSpacings -> 2, 
       FontSize -> 14], 
      Cell[
       StyleData["SubProblem", "Printout"], CellMargins -> {{58, 10}, {2, 2}},
        LineSpacing -> {1, 3}, ParagraphIndent -> -16, TabSpacings -> 2, 
       FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Comment"], CellFrame -> {{3, 0}, {0, 0}}, 
       CellMargins -> {{30, 10}, {7, 7}}, CellFrameColor -> RGBColor[0, 0, 1],
        LineSpacing -> {1, 3}, FontSize -> 14, Background -> 
       RGBColor[0.87, 0.94, 1]], 
      Cell[
       StyleData["Comment", "Printout"], CellFrame -> {{3, 0}, {0, 0}}, 
       CellMargins -> {{30, 30}, {7, 7}}, LineSpacing -> {1, 3}, FontSize -> 
       13, Background -> RGBColor[0.87, 0.94, 1]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["CalloutIcon"], CellMargins -> {{30, 108}, {0, 0}}, 
       CellGroupingRules -> {"GroupTogetherGrouping", 100000}, CellFrameColor -> 
       RGBColor[0.976, 0.855, 0.514], TextAlignment -> Left, 
       LineSpacing -> {1, 2}, FontSize -> 13], 
      Cell[
       StyleData["CalloutIcon", "Printout"], 
       CellMargins -> {{30, 288}, {0, 2}}, PageBreakBelow -> False, 
       CellFrameColor -> RGBColor[0.976, 0.855, 0.514], LineSpacing -> {1, 2},
        FontSize -> 13]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["CalloutIconFont"], FontFamily -> "Verdana", FontSize -> 12, 
       FontColor -> RGBColor[0.68, 0.57, 0.34]], 
      Cell[
       StyleData["CalloutIconFont", "Printout"], FontFamily -> "Verdana", 
       FontSize -> 12, FontColor -> RGBColor[0.68, 0.57, 0.34]]}, Open]], 
   Cell[
    CellGroupData[{
      Cell[
       StyleData["Callout"], CellFrame -> 1, 
       CellMargins -> {{30, 108}, {2, 0}}, 
       CellGroupingRules -> {"GroupTogetherGrouping", 100000}, CellFrameColor -> 
       RGBColor[0.976, 0.855, 0.514], LineSpacing -> {1, 2}, 
       ParagraphSpacing -> {0, 6}, FontSize -> 13], 
      Cell[
       StyleData["Callout", "Printout"], CellFrame -> 1, 
       CellMargins -> {{30, 180}, {2, 0}}, PageBreakWithin -> False, 
       CellFrameColor -> RGBColor[0.976, 0.855, 0.514], LineSpacing -> {1, 2},
        ParagraphSpacing -> {0, 6}, FontSize -> 13]}, Open]], 
   Cell[
    StyleData["InlineCell"], ScriptLevel -> 0, NumberSeparator -> ","], 
   Cell[
    StyleData["Output"], NumberSeparator -> ","]}, Visible -> False, 
  FrontEndVersion -> 
  "8.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (November 6, 2010)", 
  StyleDefinitions -> "PrivateStylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "14.2 Line Integrals"->{
  Cell[579, 22, 229, 5, 39, "Section",
   CellTags->"14.2 Line Integrals"]},
 "Scalar Line Integrals in the Plane"->{
  Cell[2526, 72, 200, 3, 33, "Subsection",
   CellTags->"Scalar Line Integrals in the Plane"]},
 "FIGURE 14.16  \[LightBulb]"->{
  Cell[4167, 126, 75936, 1345, 449, "Output",
   CellTags->"FIGURE 14.16  \[LightBulb]"]},
 "FIGURE 14.17  \[LightBulb]"->{
  Cell[85501, 1663, 87089, 1564, 490, "Output",
   CellTags->"FIGURE 14.17  \[LightBulb]"]},
 "DEFINITION Scalar Line Integral in the Plane, Arc Length Parameter"->{
  Cell[173442, 3256, 2841, 92, 159, "Definition",
   CellTags->
    "DEFINITION Scalar Line Integral in the Plane, Arc Length Parameter"]},
 "EXAMPLE 1 Average temperature on a circle"->{
  Cell[178406, 3432, 268, 6, 29, "Subsubsubsection",
   CellTags->"EXAMPLE 1 Average temperature on a circle"]},
 "QUICK CHECK 1"->{
  Cell[196131, 4004, 1066, 35, 31, "QuickCheck",
   CellTags->"QUICK CHECK 1"]},
 "THEOREM 14.1 Evaluating Scalar Line Integrals in R2"->{
  Cell[198148, 4076, 3112, 97, 145, "Theorem",
   CellTags->"THEOREM 14.1 Evaluating Scalar Line Integrals in R2"]},
 "PROCEDURE Evaluating the Line Integral"->{
  Cell[205894, 4342, 3061, 108, 198, "Procedure",
   CellTags->"PROCEDURE Evaluating the Line Integral"]},
 "EXAMPLE 2 Average temperature on a circle"->{
  Cell[208980, 4454, 265, 6, 29, "Subsubsubsection",
   CellTags->"EXAMPLE 2 Average temperature on a circle"]},
 "Line Integrals in R3"->{
  Cell[218680, 4752, 284, 8, 37, "Subsection",
   CellTags->"Line Integrals in R3"]},
 "THEOREM 14.2 Evaluating Scalar Line Integrals in R3"->{
  Cell[219398, 4774, 3389, 104, 148, "Theorem",
   CellTags->"THEOREM 14.2 Evaluating Scalar Line Integrals in R3"]},
 "EXAMPLE 3 Line integrals in R3"->{
  Cell[224753, 4951, 339, 9, 29, "Subsubsubsection",
   CellTags->"EXAMPLE 3 Line integrals in R3"]},
 "EXAMPLE 4 Flight of an eagle"->{
  Cell[240238, 5458, 242, 6, 29, "Subsubsubsection",
   CellTags->"EXAMPLE 4 Flight of an eagle"]},
 "QUICK CHECK 2"->{
  Cell[248263, 5717, 420, 12, 31, "QuickCheck",
   CellTags->"QUICK CHECK 2"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"14.2 Line Integrals", 269803, 6214},
 {"Scalar Line Integrals in the Plane", 269921, 6217},
 {"FIGURE 14.16  \[LightBulb]", 270050, 6220},
 {"FIGURE 14.17  \[LightBulb]", 270174, 6223},
 {"DEFINITION Scalar Line Integral in the Plane, Arc Length Parameter", \
270340, 6226},
 {"EXAMPLE 1 Average temperature on a circle", 270528, 6230},
 {"QUICK CHECK 1", 270661, 6233},
 {"THEOREM 14.1 Evaluating Scalar Line Integrals in R2", 270800, 6236},
 {"PROCEDURE Evaluating the Line Integral", 270962, 6239},
 {"EXAMPLE 2 Average temperature on a circle", 271117, 6242},
 {"Line Integrals in R3", 271257, 6245},
 {"THEOREM 14.2 Evaluating Scalar Line Integrals in R3", 271401, 6248},
 {"EXAMPLE 3 Line integrals in R3", 271556, 6251},
 {"EXAMPLE 4 Flight of an eagle", 271693, 6254},
 {"QUICK CHECK 2", 271813, 6257}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 229, 5, 39, "Section",
 CellTags->"14.2 Line Integrals"],
Cell[811, 29, 881, 24, 85, "Text"],
Cell[1695, 55, 806, 13, 113, "Text"],
Cell[CellGroupData[{
Cell[2526, 72, 200, 3, 33, "Subsection",
 CellTags->"Scalar Line Integrals in the Plane"],
Cell[2729, 77, 1435, 47, 96, "Text"],
Cell[4167, 126, 75936, 1345, 449, "Output",
 CellTags->"FIGURE 14.16  \[LightBulb]"],
Cell[80106, 1473, 1098, 41, 62, "Text"],
Cell[81207, 1516, 620, 18, 32, "Text"],
Cell[81830, 1536, 2793, 98, 106, "Text"],
Cell[84626, 1636, 872, 25, 32, "Text"],
Cell[85501, 1663, 87089, 1564, 490, "Output",
 CellTags->"FIGURE 14.17  \[LightBulb]"],
Cell[172593, 3229, 846, 25, 49, "Text"],
Cell[173442, 3256, 2841, 92, 159, "Definition",
 CellTags->
  "DEFINITION Scalar Line Integral in the Plane, Arc Length Parameter"],
Cell[176286, 3350, 2095, 78, 147, "Text"],
Cell[CellGroupData[{
Cell[178406, 3432, 268, 6, 29, "Subsubsubsection",
 CellTags->"EXAMPLE 1 Average temperature on a circle"],
Cell[178677, 3440, 805, 28, 54, "Text"],
Cell[CellGroupData[{
Cell[179507, 3472, 630, 21, 19, "CalloutIcon"],
Cell[180140, 3495, 318, 5, 69, "Callout"]
}, Closed]],
Cell[CellGroupData[{
Cell[180495, 3505, 104, 1, 22, "Subsubsubsubsection"],
Cell[180602, 3508, 1701, 54, 111, "Text"],
Cell[182306, 3564, 4060, 113, 155, "Text"],
Cell[186369, 3679, 1084, 33, 87, "Text"],
Cell[CellGroupData[{
Cell[187478, 3716, 630, 21, 19, "CalloutIcon"],
Cell[188111, 3739, 287, 7, 52, "Callout"]
}, Open  ]],
Cell[188413, 3749, 416, 10, 25, "RelatedExercises"]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[188878, 3765, 152, 2, 28, "Subsubsection"],
Cell[189033, 3769, 1249, 40, 87, "Text"],
Cell[190285, 3811, 912, 27, 37, "Text"],
Cell[191200, 3840, 619, 19, 49, "Text"],
Cell[191822, 3861, 863, 25, 30, "Text"],
Cell[CellGroupData[{
Cell[192710, 3890, 630, 21, 19, "CalloutIcon"],
Cell[193343, 3913, 923, 33, 52, "Callout"]
}, Open  ]],
Cell[194281, 3949, 313, 9, 30, "Text"],
Cell[194597, 3960, 1531, 42, 51, "Text"],
Cell[196131, 4004, 1066, 35, 31, "QuickCheck",
 CellTags->"QUICK CHECK 1"],
Cell[CellGroupData[{
Cell[197222, 4043, 320, 10, 26, "QuickCheckAnswerIcon"],
Cell[197545, 4055, 588, 18, 61, "QuickCheckAnswer"]
}, Closed]],
Cell[198148, 4076, 3112, 97, 145, "Theorem",
 CellTags->"THEOREM 14.1 Evaluating Scalar Line Integrals in R2"],
Cell[CellGroupData[{
Cell[201285, 4177, 630, 21, 19, "CalloutIcon"],
Cell[201918, 4200, 388, 10, 52, "Callout"]
}, Closed]],
Cell[202321, 4213, 1141, 40, 65, "Text"],
Cell[203465, 4255, 657, 24, 49, "Text"],
Cell[204125, 4281, 858, 25, 47, "Text"],
Cell[204986, 4308, 905, 32, 70, "Text"],
Cell[205894, 4342, 3061, 108, 198, "Procedure",
 CellTags->"PROCEDURE Evaluating the Line Integral"],
Cell[CellGroupData[{
Cell[208980, 4454, 265, 6, 29, "Subsubsubsection",
 CellTags->"EXAMPLE 2 Average temperature on a circle"],
Cell[209248, 4462, 865, 29, 54, "Text"],
Cell[210116, 4493, 867, 25, 37, "Text"],
Cell[CellGroupData[{
Cell[211008, 4522, 104, 1, 28, "Subsubsubsubsection"],
Cell[211115, 4525, 390, 15, 32, "Text"],
Cell[211508, 4542, 1401, 42, 59, "Text"],
Cell[212912, 4586, 160, 3, 30, "Text"],
Cell[213075, 4591, 4661, 125, 159, "Text"],
Cell[217739, 4718, 449, 14, 49, "Text"],
Cell[218191, 4734, 416, 10, 25, "RelatedExercises"]
}, Closed]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[218680, 4752, 284, 8, 37, "Subsection",
 CellTags->"Line Integrals in R3"],
Cell[218967, 4762, 428, 10, 51, "Text"],
Cell[219398, 4774, 3389, 104, 148, "Theorem",
 CellTags->"THEOREM 14.2 Evaluating Scalar Line Integrals in R3"],
Cell[222790, 4880, 593, 22, 30, "Text"],
Cell[223386, 4904, 993, 29, 47, "Text"],
Cell[224382, 4935, 346, 12, 30, "Text"],
Cell[CellGroupData[{
Cell[224753, 4951, 339, 9, 29, "Subsubsubsection",
 CellTags->"EXAMPLE 3 Line integrals in R3"],
Cell[225095, 4962, 400, 14, 59, "Text"],
Cell[225498, 4978, 391, 14, 31, "Text"],
Cell[225892, 4994, 389, 14, 31, "Text"],
Cell[CellGroupData[{
Cell[226306, 5012, 630, 21, 19, "CalloutIcon"],
Cell[226939, 5035, 1597, 47, 96, "Callout"]
}, Closed]],
Cell[CellGroupData[{
Cell[228573, 5087, 104, 1, 22, "Subsubsubsubsection"],
Cell[228680, 5090, 427, 15, 31, "Text"],
Cell[229110, 5107, 1145, 32, 30, "Text"],
Cell[230258, 5141, 103, 1, 30, "Text"],
Cell[230364, 5144, 1408, 43, 41, "Text"],
Cell[231775, 5189, 414, 16, 30, "Text"],
Cell[232192, 5207, 3482, 101, 159, "Text"],
Cell[235677, 5310, 429, 15, 31, "Text"],
Cell[236109, 5327, 1243, 34, 30, "Text"],
Cell[237355, 5363, 101, 1, 30, "Text"],
Cell[237459, 5366, 1525, 46, 41, "Text"],
Cell[238987, 5414, 781, 26, 92, "Text"],
Cell[239771, 5442, 418, 10, 25, "RelatedExercises"]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[240238, 5458, 242, 6, 29, "Subsubsubsection",
 CellTags->"EXAMPLE 4 Flight of an eagle"],
Cell[240483, 5466, 125, 1, 30, "Text"],
Cell[240611, 5469, 1064, 29, 41, "Text"],
Cell[241678, 5500, 551, 20, 49, "Text"],
Cell[CellGroupData[{
Cell[242254, 5524, 104, 1, 28, "Subsubsubsubsection"],
Cell[CellGroupData[{
Cell[242383, 5529, 630, 21, 19, "CalloutIcon"],
Cell[243016, 5552, 333, 10, 52, "Callout"]
}, Closed]],
Cell[243364, 5565, 694, 23, 75, "Text"],
Cell[244061, 5590, 2134, 64, 125, "Text"],
Cell[246198, 5656, 122, 1, 30, "Text"],
Cell[246323, 5659, 1497, 42, 47, "Text"],
Cell[247823, 5703, 413, 10, 25, "RelatedExercises"]
}, Closed]]
}, Open  ]],
Cell[248263, 5717, 420, 12, 31, "QuickCheck",
 CellTags->"QUICK CHECK 2"],
Cell[CellGroupData[{
Cell[248708, 5733, 320, 10, 26, "QuickCheckAnswerIcon"],
Cell[249031, 5745, 177, 4, 23, "QuickCheckAnswer"]
}, Closed]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
